package demo.barcharts.category ;

import java.awt.Color ;
import java.awt.Font ;
import java.awt.GridLayout ;

import javax.swing.JPanel ;

import com.ijchart.xychart.axis.category.CategoryAxis ;
import com.ijchart.xychart.axis.value.NumberValueAxis ;
import com.ijchart.xychart.client.ChartFactory ;
import com.ijchart.xychart.client.ChartPanel ;
import com.ijchart.xychart.client.IJChart ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.HorizontalAlignment ;
import com.ijchart.xychart.constant.IAngle ;
import com.ijchart.xychart.constant.TextDirection ;
import com.ijchart.xychart.constant.VerticalAlignment ;
import com.ijchart.xychart.data.NumberChartData ;
import com.ijchart.xychart.dataset.DefaultMatrixDataset ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.label.DefaultLabelGenerator ;
import com.ijchart.xychart.marker.IntervalValueMarker ;
import com.ijchart.xychart.plot.CategoryPlot ;
import com.ijchart.xychart.render.category.Bar2DRender ;
import com.ijchart.xychart.tickunit.LongTickUnit ;
import com.ijchart.xychart.ui.Layer ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class BarChartDemo7 extends JPanel {
  private ChartOrientation orientation = ChartOrientation.VERTICAL ;

  public BarChartDemo7() {
    setLayout(new GridLayout(1, 1)) ;
    addPanel() ;
  }

  private void addPanel() {
    final IMatrixDataset dataset = createDataset() ;
    final IJChart chart = ChartFactory.createCategoryBar2DChart(
      "Bar Chart Demo 7", "Category", "Value", dataset, orientation, false, true) ;

    CategoryPlot plot = (CategoryPlot) chart.getPlot() ;
    IntervalValueMarker intervalmarker = new IntervalValueMarker(4.5D, 7.5D) ;
    intervalmarker.setText("Target Range") ;
    intervalmarker.setTextFont(new Font("SansSerif", 2, 11)) ;
    intervalmarker.setTextHorizontalAlign(HorizontalAlignment.LEFT) ;
    intervalmarker.setTextVerticalAlign(VerticalAlignment.CENTER) ;
    intervalmarker.setFillPaint(new Color(222, 222, 255, 128)) ;

    CategoryAxis categoryaxis = plot.getCategoryAxis() ;
    categoryaxis.setTickLabelAngle(IAngle.ANGLE_UP_90) ;

    NumberValueAxis numberaxis = (NumberValueAxis) plot.getValueAxis() ;
    numberaxis.setTickUnit(new LongTickUnit()) ;
    numberaxis.addIntervalValueMarker(Layer.BACKGROUND, intervalmarker) ;

    Bar2DRender render = (Bar2DRender) plot.getRender() ;
    render.setItemMargin(0.10000000000000001D) ;
    render.setBaseItemLabelVisible(true) ;
    render.setBaseItemLabelDirection(TextDirection.VERTICAL) ;
    render.setBaseItemLabelVerticalAlignment(VerticalAlignment.BOTTOM) ;
    render.setLabelGenerator(new DefaultLabelGenerator() {
      public String generateLabel(IMatrixDataset dataset, int row, int column) {
        return dataset.getColKey(column).toString() ;
      }
    }) ;

    ChartPanel chartpanel = new ChartPanel(chart) ;
    add(chartpanel) ;
  }

  private static IMatrixDataset createDataset() {
    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addColKey("First") ;
    dataset.addColKey("Second") ;
    dataset.addColKey("Third") ;

    for (int i = 1 ; i <= 5 ; i++) {
      dataset.addRowKey("Category " + i) ;
    }

    double data[][] = { {1.0D, 4D, 3D, 5D, 5D}, {5D, 7D, 6D, 8D, 4D}, {4D, 3D, 0.0D, 3D, 6D}
    } ;
    dataset.addRowData(5) ;

    for (int i = 0 ; i < 5 ; i++) {
      for (int j = 0 ; j < data.length ; j++) {
        dataset.addData(i, new NumberChartData(data[j][i])) ;
      }
    }

    return dataset ;
  }

}
