package demo.others.axis;

import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Dimension ;
import java.awt.Graphics ;
import java.awt.Graphics2D ;
import java.awt.GridLayout ;
import java.awt.Paint ;
import java.awt.RenderingHints ;
import java.awt.Stroke ;
import java.awt.geom.Rectangle2D ;
import java.util.HashMap ;
import java.util.Map ;

import javax.swing.JPanel ;

import com.ijchart.xychart.axis.category.CategoryAxis3D ;
import com.ijchart.xychart.chartinfo.AxisShapeCollection ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.HorizontalAlignment ;
import com.ijchart.xychart.constant.VerticalAlignment ;
import com.ijchart.xychart.data.NumberChartData ;
import com.ijchart.xychart.dataset.DefaultMatrixDataset ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.plot.Category3DPlot ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.ui.RectangleInsets ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class CategoryAxis3DPanel extends JPanel {
  private Map hashMap = new HashMap() ;

  public static final Stroke DEFAULT_GRIDLINE_STROKE =
    new BasicStroke(0.5f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 0.0f,
                    new float[] {2.0f, 2.0f}, 0.0f) ;

  public static final Paint DEFAULT_GRIDLINE_PAINT = Color.red ;

  public CategoryAxis3DPanel() {
    init() ;
  }

  private void init() {
    hashMap.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON) ;
    setLayout(new GridLayout(2, 2)) ;
    addPanel1() ;
    addPanel2() ;
    addPanel3() ;
    addPanel4() ;
  }

  private void addPanel1() {
    JPanel panel = new JPanel() {

      public void paint(Graphics g) {
        super.paintComponent(g) ;
        Graphics2D g2 = (Graphics2D) g ;
        g2.addRenderingHints(hashMap) ;
        setBackground(Color.white) ;
        g2.setPaint(Color.red) ;
        int width = getWidth() ;
        int height = getHeight() ;
        g2.drawRect(1, 1, getWidth() - 2, getHeight() - 2) ;

        final double angle = 0.25 * Math.PI ;
        CategoryAxis3D axis = new CategoryAxis3D(
          "labellabellabellabellabellabellabellabellabellabellabel", createDataset(),
          new Category3DPlot()) ;
        axis.setTickMarkVisible(true) ;
        axis.setTickMarkInsideLength(3) ;
        axis.setTickMarkOutsideLength(8) ;
//        axis.setTickLabelAngle(angle) ;
        axis.setMaximumCategoryLabelWidthRatio(0.3F) ;
//        axis.setTickLabelHorizontalAlign(HorizontalAlignment.RIGHT);
        axis.setTickLabelVerticalAlign(VerticalAlignment.CENTER) ;
        axis.setTickLabelInsets(RectangleInsets.INSETS_ZERO) ;

        axis.setAxisLabelHorizontalAlign(HorizontalAlignment.CENTER) ;
//        axis.setAxisLabelAngle(angle) ;

        Rectangle2D axisArea = new Rectangle2D.Double(width * 0.1, height * 0.1, width * 0.8,
                                                      height * 0.3) ;
        Rectangle2D dataArea = new Rectangle2D.Double(width * 0.1, height * 0.45, width * 0.8,
                                                      height * 0.4) ;
        g2.setPaint(Color.blue) ;
        g2.draw(axisArea) ;
        g2.setPaint(Color.GREEN) ;
        g2.draw(dataArea) ;

        AxisShapeCollection shapes = new AxisShapeCollection() ;
        shapes.setDrawArea(axisArea) ;
        double axisHeight = axis.draw(g2, new Rectangle2D.Double(0, 0, width, height),
                                      axisArea, RectangleEdge.TOP, shapes, true) ;
        axisHeight = axis.calculateAxisSpace(g2, new Rectangle2D.Double(0, 0, width, height),
                                             axisArea, RectangleEdge.TOP) ;

        RenderShapeCollection renderShapes = new RenderShapeCollection() ;
        renderShapes.setDrawArea(dataArea) ;
        axis.drawGridLine(g2, dataArea, RectangleEdge.TOP, renderShapes,
                          DEFAULT_GRIDLINE_STROKE, DEFAULT_GRIDLINE_PAINT) ;

        shapes.draw(g2) ;
        renderShapes.draw(g2) ;
      }

    } ;

    panel.setPreferredSize(new Dimension(300, 300)) ;
    add(panel) ;
  }

  private void addPanel2() {
    JPanel panel = new JPanel() {

      public void paint(Graphics g) {
        super.paintComponent(g) ;
        Graphics2D g2 = (Graphics2D) g ;
        g2.addRenderingHints(hashMap) ;
        setBackground(Color.white) ;
        g2.setPaint(Color.red) ;
        int width = getWidth() ;
        int height = getHeight() ;
        g2.drawRect(1, 1, getWidth() - 2, getHeight() - 2) ;

        final double angle = 0.25 * Math.PI ;
        CategoryAxis3D axis = new CategoryAxis3D("label", createDataset(), new Category3DPlot()) ;
        axis.setTickMarkVisible(true) ;
        axis.setTickMarkInsideLength(3) ;
        axis.setTickMarkOutsideLength(8) ;
        axis.setTickLabelAngle(angle) ;
        axis.setTickLabelHorizontalAlign(HorizontalAlignment.RIGHT) ;
        axis.setTickLabelVerticalAlign(VerticalAlignment.CENTER) ;
        axis.setAxisLabelHorizontalAlign(HorizontalAlignment.CENTER) ;
        axis.setAxisLabelAngle(angle) ;

        Rectangle2D axisArea = new Rectangle2D.Double(width * 0.1, height * 0.45, width * 0.8,
                                                      height * 0.3) ;
        Rectangle2D dataArea = new Rectangle2D.Double(width * 0.1, height * 0.1, width * 0.8,
                                                      height * 0.3) ;
        g2.setPaint(Color.blue) ;
        g2.draw(axisArea) ;
        g2.setPaint(Color.GREEN) ;
        g2.draw(dataArea) ;

        AxisShapeCollection shapes = new AxisShapeCollection() ;
        shapes.setDrawArea(axisArea) ;
        double axisHeight = axis.draw(g2, new Rectangle2D.Double(0, 0, width, height),
                                      axisArea, RectangleEdge.BOTTOM, shapes, true) ;

        axisHeight = axis.calculateAxisSpace(g2, new Rectangle2D.Double(0, 0, width, height),
                                             axisArea, RectangleEdge.BOTTOM) ;

        RenderShapeCollection renderShapes = new RenderShapeCollection() ;
        renderShapes.setDrawArea(dataArea) ;
        axis.drawGridLine(g2, dataArea, RectangleEdge.BOTTOM, renderShapes,
                          DEFAULT_GRIDLINE_STROKE, DEFAULT_GRIDLINE_PAINT) ;

        shapes.draw(g2) ;
        renderShapes.draw(g2) ;

      }

    } ;

    panel.setPreferredSize(new Dimension(300, 300)) ;
    add(panel) ;
  }

  private void addPanel3() {
    JPanel panel = new JPanel() {

      public void paint(Graphics g) {
        super.paintComponent(g) ;
        Graphics2D g2 = (Graphics2D) g ;
        g2.addRenderingHints(hashMap) ;
        setBackground(Color.white) ;
        g2.setPaint(Color.red) ;
        int width = getWidth() ;
        int height = getHeight() ;
        g2.drawRect(1, 1, getWidth() - 2, getHeight() - 2) ;

        final double angle = 0.25 * Math.PI ;
        CategoryAxis3D axis = new CategoryAxis3D("label", createDataset(), new Category3DPlot()) ;
        axis.setTickMarkVisible(true) ;
        axis.setTickMarkInsideLength(3) ;
        axis.setTickMarkOutsideLength(8) ;
        axis.setTickLabelAngle(Math.PI * 0.2) ;
        axis.setTickLabelHorizontalAlign(HorizontalAlignment.RIGHT) ;
        axis.setTickLabelVerticalAlign(VerticalAlignment.CENTER) ;
        axis.setAxisLabelHorizontalAlign(HorizontalAlignment.CENTER) ;
        axis.setAxisLabelAngle(angle) ;

        Rectangle2D axisArea = new Rectangle2D.Double(width * 0.1, height * 0.1, width * 0.3,
                                                      height * 0.8) ;
        Rectangle2D dataArea = new Rectangle2D.Double(width * 0.45, height * 0.1, width * 0.3,
                                                      height * 0.8) ;
        g2.setPaint(Color.blue) ;
        g2.draw(axisArea) ;
        g2.setPaint(Color.GREEN) ;
        g2.draw(dataArea) ;

        AxisShapeCollection shapes = new AxisShapeCollection() ;
        shapes.setDrawArea(axisArea) ;
        double axisHeight = axis.draw(g2, new Rectangle2D.Double(0, 0, width, height),
                                      axisArea, RectangleEdge.LEFT, shapes, true) ;

        axisHeight = axis.calculateAxisSpace(g2, new Rectangle2D.Double(0, 0, width, height),
                                             axisArea, RectangleEdge.LEFT) ;

        RenderShapeCollection renderShapes = new RenderShapeCollection() ;
        renderShapes.setDrawArea(dataArea) ;
        axis.drawGridLine(g2, dataArea, RectangleEdge.LEFT, renderShapes,
                          DEFAULT_GRIDLINE_STROKE, DEFAULT_GRIDLINE_PAINT) ;

        shapes.draw(g2) ;
        renderShapes.draw(g2) ;
      }

    } ;

    panel.setPreferredSize(new Dimension(300, 300)) ;
    add(panel) ;
  }

  private void addPanel4() {
    JPanel panel = new JPanel() {

      public void paint(Graphics g) {
        super.paintComponent(g) ;
        Graphics2D g2 = (Graphics2D) g ;
        g2.addRenderingHints(hashMap) ;
        setBackground(Color.white) ;
        g2.setPaint(Color.red) ;
        int width = getWidth() ;
        int height = getHeight() ;
        g2.drawRect(1, 1, getWidth() - 2, getHeight() - 2) ;

        final double angle = 0.25 * Math.PI ;
        CategoryAxis3D axis = new CategoryAxis3D("label", createDataset(), new Category3DPlot()) ;
        axis.setTickMarkVisible(true) ;
        axis.setTickMarkInsideLength(3) ;
        axis.setTickMarkOutsideLength(8) ;
        axis.setTickLabelAngle(Math.PI * -0.2) ;
        axis.setAxisLabelHorizontalAlign(HorizontalAlignment.CENTER) ;
        axis.setAxisLabelAngle(angle) ;

        Rectangle2D axisArea = new Rectangle2D.Double(width * 0.45, height * 0.1, width * 0.3,
                                                      height * 0.8) ;
        Rectangle2D dataArea = new Rectangle2D.Double(width * 0.1, height * 0.1, width * 0.3,
                                                      height * 0.8) ;
        g2.setPaint(Color.blue) ;
        g2.draw(axisArea) ;
        g2.setPaint(Color.GREEN) ;
        g2.draw(dataArea) ;

        AxisShapeCollection shapes = new AxisShapeCollection() ;
        shapes.setDrawArea(axisArea) ;
        double axisHeight = axis.draw(g2, new Rectangle2D.Double(0, 0, width, height),
                                      axisArea, RectangleEdge.RIGHT, shapes, true) ;

        axisHeight = axis.calculateAxisSpace(g2, new Rectangle2D.Double(0, 0, width, height),
                                             axisArea, RectangleEdge.RIGHT) ;

        RenderShapeCollection renderShapes = new RenderShapeCollection() ;
        renderShapes.setDrawArea(dataArea) ;
        axis.drawGridLine(g2, dataArea, RectangleEdge.RIGHT, renderShapes,
                          DEFAULT_GRIDLINE_STROKE, DEFAULT_GRIDLINE_PAINT) ;

        shapes.draw(g2) ;
        renderShapes.draw(g2) ;

      }

    } ;

    panel.setPreferredSize(new Dimension(300, 300)) ;
    add(panel) ;
  }

  private IMatrixDataset createDataset() {
    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addRowKey("123456789011") ;
    dataset.addRowKey("ab\ncd\nef") ;
    dataset.addRowKey("abcde") ;
    dataset.addRowKey("12345678") ;
//    dataset.addRowKey("1234567890") ;
//    dataset.addRowKey("1234567890") ;
//    dataset.addRowKey("1234567890") ;
//    dataset.addRowKey("1234567890") ;
//    dataset.addRowKey("1234567890") ;
//    dataset.addRowKey("1234567890") ;

    for (int i = 0 ; i < 4 ; i++) {
      dataset.addRowData() ;
      for (int j = 0 ; j < 5 ; j++) {
        dataset.addData(i, new NumberChartData((i + 1) * j + 1)) ;
      }
    }

    return dataset ;
  }
}
