package demo.stackedbarcharts ;

import java.awt.GridLayout ;
import java.text.NumberFormat ;

import javax.swing.JPanel ;

import com.ijchart.xychart.axis.value.Stacked3DNumberValueAxis ;
import com.ijchart.xychart.client.ChartFactory ;
import com.ijchart.xychart.client.ChartPanel ;
import com.ijchart.xychart.client.IJChart ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.data.NumberChartData ;
import com.ijchart.xychart.dataset.DefaultMatrixDataset ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.format.AppendFormatter ;
import com.ijchart.xychart.label.DefaultLabelGenerator ;
import com.ijchart.xychart.plot.CategoryPlot ;
import com.ijchart.xychart.render.category.StackedBar3DRender ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class StackedBarChart3DDemo2 extends JPanel {
  private ChartOrientation orientation = ChartOrientation.VERTICAL ;

  public StackedBarChart3DDemo2() {
    setLayout(new GridLayout(1, 1)) ;
    addPanel1() ;
  }

  private void addPanel1() {
    IMatrixDataset dataset = createDataset1() ;
    IJChart chart = ChartFactory.createStackedBar3DChart("Stacked Bar Chart 3D Demo 2", "Category",
                                                         "Value", dataset, orientation, true, true) ;
    CategoryPlot plot = (CategoryPlot) chart.getPlot() ;
    Stacked3DNumberValueAxis valueAxis = (Stacked3DNumberValueAxis) plot.getValueAxis() ;
    valueAxis.setTickValueFormatter(new AppendFormatter(null, "%")) ;

    StackedBar3DRender render = (StackedBar3DRender) plot.getRender() ;
    render.setBaseItemLabelVisible(true) ;
    render.setRenderAsPercentages(true) ;
    render.setLabelGenerator(new DefaultLabelGenerator() {
      private NumberFormat format = NumberFormat.getPercentInstance() ;

      public String generateLabel(IMatrixDataset dataset, int row, int column) {
        format.setMaximumFractionDigits(1) ;
        NumberChartData data = (NumberChartData) dataset.getData(row, column) ;
        return format.format(data.getValue() / calculateRowTotal(dataset, row)) ;
      }

      private double calculateRowTotal(IMatrixDataset dataset, int row) {
        double rowTotal = 0 ;

        int currRowCols = dataset.getColumnCount(row) ;
        for (int j = 0 ; j < currRowCols ; j++) {
          NumberChartData numberData = (NumberChartData) dataset.getData(row, j) ;
          rowTotal += Math.abs(numberData.getValue()) ;
        }

        return rowTotal ;

      }
    }) ;

    ChartPanel chartpanel = new ChartPanel(chart) ;
    add(chartpanel) ;
  }

  private static IMatrixDataset createDataset1() {
    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addColKey("Series 1") ;
    dataset.addColKey("Series 2") ;
    dataset.addColKey("Series 3") ;
    dataset.addColKey("Series 4") ;

    for (int i = 1 ; i < 10 ; i++) {
      dataset.addRowKey("C" + i) ;
    }

    double data1[] = {10D, 5D, 6D, 7D, 8D, 9D, 10D, 11D, 3D} ;
    double data2[] = {4D, 7D, 17D, 15D, 6D, 8D, 9D, 13D, 7D} ;
    double data3[] = {15D, 14D, 12D, 11D, 10D, 0.0D, 7D, 9D, 11D} ;
    double data4[] = {14D, 3D, 7D, 0.0D, 9D, 6D, 7D, 9D, 10D} ;

    dataset.addRowData(data1.length) ;
    for (int i = 0 ; i < data1.length ; i++) {
      dataset.addData(i, new NumberChartData(data1[i])) ;
      dataset.addData(i, new NumberChartData(data2[i])) ;
      dataset.addData(i, new NumberChartData(data3[i])) ;
      dataset.addData(i, new NumberChartData(data4[i])) ;
    }

    return dataset ;
  }

}
