package demo.timeseriescharts ;

import java.awt.GridLayout ;
import java.util.ArrayList ;
import java.util.Calendar ;
import java.util.List ;

import javax.swing.JPanel ;

import com.ijchart.xychart.axis.value.XYNumberValueAxis ;
import com.ijchart.xychart.client.ChartFactory ;
import com.ijchart.xychart.client.ChartPanel ;
import com.ijchart.xychart.client.IJChart ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.DateTickUnitType ;
import com.ijchart.xychart.data.XYNumberChartData ;
import com.ijchart.xychart.dataset.DefaultMatrixDataset ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.plot.XYPlot ;
import com.ijchart.xychart.render.xy.XYLineAndShapeRender ;
import com.ijchart.xychart.tickunit.DateTickUnit ;
import com.ijchart.xychart.util.DateUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class TimeSeriesDemo10 extends JPanel {

  public TimeSeriesDemo10() {
    setLayout(new GridLayout(1, 1)) ;
    addPanel1() ;
  }

  private void addPanel1() {
    IMatrixDataset dataset = createDataset() ;
    IJChart chart = ChartFactory.createXYLineAndShapeChart("Time Series Demo 10", "Time",
                                                           "Value", dataset,
                                                           ChartOrientation.VERTICAL,
                                                           true, true, true, false) ;
    XYPlot xyplot = (XYPlot) chart.getPlot() ;

    XYNumberValueAxis xAxis = (XYNumberValueAxis) xyplot.getDomainAxis() ;
    xAxis.setTickUnit(new DateTickUnit(DateTickUnitType.DATE_UNIT_MINUTE)) ;

    XYNumberValueAxis yAxis = (XYNumberValueAxis) xyplot.getValueAxis() ;
    yAxis.setAxisRangeIncludeZero(false) ;

    XYLineAndShapeRender render = (XYLineAndShapeRender) xyplot.getRender() ;
    render.setBaseItemShapeVisible(false) ;

    ChartPanel chartpanel = new ChartPanel(chart) ;
    add(chartpanel) ;
  }

  private static IMatrixDataset createDataset() {
    Calendar cal = Calendar.getInstance() ;
    int year = cal.get(Calendar.YEAR) ;
    int month = cal.get(Calendar.MONTH) + 1 ;
    int day = cal.get(Calendar.DAY_OF_MONTH) ;
    int hour = cal.get(Calendar.HOUR_OF_DAY) ;

    DefaultMatrixDataset dataset = new DefaultMatrixDataset() ;
    dataset.addRowKey("Per Minute Data") ;
    dataset.addRowData() ;

    dataset.addData(0, createData(1, hour, day, month, year, 10.199999999999999D)) ;
    dataset.addData(0, createData(3, hour, day, month, year, 17.300000000000001D)) ;
    dataset.addData(0, createData(9, hour, day, month, year, 14.6D)) ;
    dataset.addData(0, createData(11, hour, day, month, year, 11.9D)) ;
    dataset.addData(0, createData(15, hour, day, month, year, 13.5D)) ;
    dataset.addData(0, createData(19, hour, day, month, year, 10.9D)) ;

    int cols = dataset.getColumnCount() ;
    List datas = new ArrayList() ;
    for (int i = 0 ; i < cols ; i++) {
      datas.add(dataset.getData(0, i)) ;
    }

    return dataset ;
  }

  private static XYNumberChartData createData(int minute, int hour, int day,
                                              int month, int year, double value) {
    return new XYNumberChartData(DateUtil.dateToTimeInMillis(year, month, day, hour, minute), value) ;
  }

}
