package com.ijchart.xychart.annotation ;

import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.AnnotationShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.shape.RectangleTextChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;

/**
 * <p>Title: IJChart</p>
 * <zh>{@link com.ijchart.xychart.plot.XYPlot}ıע(annotation).</zh>
 * <en>A text annotation that can be placed on a {@link com.ijchart.xychart.plot.XYPlot}.</en>
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class XYTextAnnotation extends AbstractTextAnnotation implements IXYAnnotation,
  Cloneable, Serializable {

  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 8904611945496504238L ;

  /**
   * <zh>xֵ.</zh>
   * <en>The x value.</en>
   */
  private double x ;

  /**
   * <zh>yֵ.</zh>
   * <en>The y value.</en>
   */
  private double y ;

  /**
   * <zh>xyֵȷʾעλ.</zh>
   * <en>Creates a new annotation to be displayed at the given location.</en>
   *
   * @param text String <zh>ע (Ϊnull).</zh>
   * <en>The text (null not permitted).</en>
   * @param x double <zh>xֵ.</zh>
   * <en>The x value.</en>
   * @param y double <zh>yֵ.</zh>
   * <en>The y value.</en>
   */
  public XYTextAnnotation(String text, double x, double y) {
    super(text) ;

    this.x = x ;
    this.y = y ;
  }

  /**
   * <zh>xֵ.</zh>
   * <en>Returns the x value.</en>
   *
   * @return double <zh>xֵ.</zh>
   * <en>The x value.</en>
   */
  public double getX() {
    return this.x ;
  }

  /**
   * <zh>xֵ,м
   * {@link com.ijchart.xychart.event.AnnotationChangeEvent}¼.</zh>
   * <en>Sets the x value and sends an
   * {@link com.ijchart.xychart.event.AnnotationChangeEvent} to all registered listeners.</en>
   *
   * @param x double <zh>xֵ.</zh>
   * <en>The x value.</en>
   */
  public void setX(double x) {
    if (this.x != x) {
      this.x = x ;
      fireAnnotationChanged() ;
    }
  }

  /**
   * <zh>yֵ.</zh>
   * <en>Returns the y value.</en>
   *
   * @return double <zh>yֵ.</zh>
   * <en>The y value.</en>
   */
  public double getY() {
    return this.y ;
  }

  /**
   * <zh>yֵ,м
   * {@link com.ijchart.xychart.event.AnnotationChangeEvent}¼.</zh>
   * <en>Sets the y value and sends an
   * {@link com.ijchart.xychart.event.AnnotationChangeEvent} to all registered listeners.</en>
   *
   * @param y double <zh>yֵ.</zh>
   * <en>The y value.</en>
   */
  public void setY(double y) {
    if (this.y != y) {
      this.y = y ;
      fireAnnotationChanged() ;
    }
  }

  /**
   * <zh>ע(annotation)ͼ.</zh>
   * <en>Draws the annotation.</en>
   *
   * @param xAxis AbstractValueAxis <zh>x (Ϊnull).</zh>
   * <en>The x-axis (null not permitted).</en>
   * @param xAxisEdge RectangleEdge <zh>xλ (Ϊnull).</zh>
   * <en>The x-axis edge (null not permitted).</en>
   * @param yAxis AbstractValueAxis <zh>y (Ϊnull).</zh>
   * <en>The y-axis (null not permitted).</en>
   * @param yAxisEdge RectangleEdge <zh>yλ (Ϊnull).</zh>
   * <en>The y-axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param annotationShapes AnnotationShapeCollection <zh>ע(annotation)ͼμ (Ϊnull).</zh>
   * <en>Collects information about the annotation (null not permitted).</en>
   */
  public void drawAnnotation(AbstractValueAxis xAxis, RectangleEdge xAxisEdge,
                             AbstractValueAxis yAxis, RectangleEdge yAxisEdge, Rectangle2D dataArea,
                             ChartOrientation orientation,
                             AnnotationShapeCollection annotationShapes) {
    if (xAxis == null) {
      throw new IllegalArgumentException("Null 'xAxis' argument.") ;
    }
    if (xAxisEdge == null) {
      throw new IllegalArgumentException("Null 'xAxisEdge' argument.") ;
    }
    if (yAxis == null) {
      throw new IllegalArgumentException("Null 'yAxis' argument.") ;
    }
    if (yAxisEdge == null) {
      throw new IllegalArgumentException("Null 'yAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (annotationShapes == null) {
      throw new IllegalArgumentException("Null 'annotationShapes' argument.") ;
    }

    double xPos = xAxis.valueToJava2D(dataArea, xAxisEdge, this.x, false) ;
    if (xPos == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
      return ;
    }

    double yPos = yAxis.valueToJava2D(dataArea, yAxisEdge, this.y, false) ;
    if (yPos == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
      return ;
    }

    Rectangle2D rect = new Rectangle2D.Double() ;
    if (orientation.equals(ChartOrientation.VERTICAL)) {
      rect = new Rectangle2D.Double(xPos, yPos, 1, 1) ;
    } else if (orientation.equals(ChartOrientation.HORIZONTAL)) {
      rect = new Rectangle2D.Double(yPos, xPos, 1, 1) ;
    }

    RectangleTextChartShape annotation = new RectangleTextChartShape(rect, getText(), getTextFont()) ;
    annotation.setDrawBorder(false) ;
    annotation.setTextPaint(getTextPaint()) ;
    annotation.setFillShape(false) ;
    annotation.setRotate(getRotationAngle()) ;
    annotation.setForceShowAllText(true) ;
    annotation.setForceInsideTextNoWrap(true);
    annotation.setTooltipText(getTooltipText()) ;

    annotation.setTextHorizontalAlign(getTextHorizontalAlign()) ;
    annotation.setTextVerticalAlign(getTextVerticalAlign()) ;
    annotation.setTextPosition(getTextPosition()) ;
    annotation.setTextDirection(getTextDirection()) ;

    annotationShapes.addAnnotationShape(annotation) ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof XYTextAnnotation)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    XYTextAnnotation that = (XYTextAnnotation) obj ;
    if (this.x != that.x) {
      return false ;
    }
    if (this.y != that.y) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double hashcode = super.hashCode() ;
    hashcode = 37 * hashcode + this.x ;
    hashcode = 37 * hashcode + this.y ;

    return (int) hashcode ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[x=" + this.x
      + ",y=" + this.y
      + "]" ;
  }

}
