package com.ijchart.xychart.axis.value ;

import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Graphics2D ;
import java.awt.Paint ;
import java.awt.Stroke ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Collection ;
import java.util.Collections ;
import java.util.List ;

import com.ijchart.xychart.axis.AbstractAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.event.MarkerChangeEvent ;
import com.ijchart.xychart.format.IFormatter ;
import com.ijchart.xychart.listener.IMarkerChangeListener ;
import com.ijchart.xychart.marker.AbstractMarker ;
import com.ijchart.xychart.marker.IntervalValueMarker ;
import com.ijchart.xychart.marker.ValueMarker ;
import com.ijchart.xychart.range.NumberRange ;
import com.ijchart.xychart.render.IRenderWithAxis ;
import com.ijchart.xychart.ui.Layer ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ʾֵ.</zh>
 * <en>An axis that displays values (Y-Axis).</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public abstract class AbstractValueAxis extends AbstractAxis implements IMarkerChangeListener,
  Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 450837188238014211L ;

  /**
   * <zh>axisRangeIncludeZeroȱʡֵ.</zh>
   * <en>The default value for the axisRangeIncludeZero flag.</en>
   */
  public static final boolean DEFAULT_AXIS_TICK_INCLUDE_ZERO = true ;

  /**
   * <zh>Ƿתȱʡֵ.</zh>
   * <en>The default inverted flag setting. </en>
   */
  public static final boolean DEFAULT_INVERTED = false ;

  /**
   * <zh>ȱʡֵǷʾ</zh>
   * <en>The default crosshair visibility.</en>
   */
  public static final boolean DEFAULT_VALUE_LINE_VISIBLE = false ;

  /**
   * <zh>ȱʡֵstroke.</zh>
   * <en>The default value line stroke.</en>
   */
  public static final Stroke DEFAULT_CROSSHAIR_STROKE =
    new BasicStroke(0.5f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 0.0f,
                    new float[] {2.0f, 2.0f}, 0.0f) ;

  /**
   * <zh>ȱʡʮpaint.</zh>
   * <en>The default crosshair line paint.</en>
   */
  public static final Paint DEFAULT_CROSSHAIR_PAINT = Color.blue ;

  /**
   * <zh>ȱʡֵpaint.</zh>
   * <en>The default value line paint.</en>
   */
  public static final Paint DEFAULT_VALUE_PAINT = new Color(182, 182, 182) ;

  /**
   * <zh>ĳ̶ֵֵ֮,ֵûжӦĵ.</zh>
   * <en>If the value not between the minimum and maximum of axis tick mark,this axis in the
   * coordinates did not correspond to a point.</en>
   */
  public static final double INVALID_COORDINATE_VALUE = Double.MIN_VALUE ;

  /**
   * <zh>ȱʡĻPaint.</zh>
   * <en>The default grid band paint.</en>
   */
  public static final Paint DEFAULT_GRID_BAND_PAINT = new Color(232, 234, 232, 128) ;

  /**
   * <zh>ȱʡĻPaint.</zh>
   * <en>The default paint for alternate grid bands.</en>
   */
  public static final Paint DEFAULT_GRID_BAND_ALTERNATE_PAINT = new Color(0, 0, 0, 0) ; // transparent

  /**
   * <zh>ȱʡǷʾı־.</zh>
   * <en>The default flag that indicates whether or not grid bands are visible.</en>
   */
  public static final boolean DEFAULT_GRID_BAND_VISIBLE = false ;

  /**
   * <zh>Ŀ̶ʱǷ0.</zh>
   * <en>Whether or not include zero when calculate the axis data range.</en>
   */
  private boolean axisRangeIncludeZero ;

  /**
   * <zh>Ƿתֵ.</zh>
   * <en>A flag that affects the orientation of the values on the axis.</en>
   */
  private boolean inverted ;

  /**
   * <zh>ݼΪʱݷΧ.</zh>
   * <en>The default range is used when the dataset is empty and the axis needs
   * to determine the auto range.</en>
   */
  private NumberRange emptyDatasetRange ;

  /**
   * <zh>ڸʽǩֵ.</zh>
   * <en>The tick value formatter.</en>
   */
  private IFormatter tickValueFormatter ;

  /**
   * <zh>Ƿvalueֵ.</zh>
   * <en>A flag that controls whether or not a value line is drawn.</en>
   */
  private boolean valueLineVisible ;

  /**
   * <zh>valueֵ᷽ߵֵ.</zh>
   * <en>The value line value.</en>
   */
  private double valueLineValue ;

  /**
   * <zh>ֵߵstroke.</zh>
   * <en>The pen/brush used to draw the value line (if any).</en>
   */
  private transient Stroke valueLineStroke ;

  /**
   * <zh>ֵߵpaint.</zh>
   * <en>The color used to draw the value line (if any).</en>
   */
  private transient Paint valueLinePaint ;

  /**
   * <zh>value᷽ǰǼ.</zh>
   * <en>A list containing lists of markers for the value axes.</en>
   */
  private List foregroundValueMarkers ;

  /**
   * <zh>value᷽ıǼ</zh>
   * <en>A list containing lists of markers for the value axes.</en>
   */
  private List backgroundValueMarkers ;

  /**
   * <zh>value᷽ǰintervalǼ.</zh>
   * <en>A list containing lists of interval markers for the value axes.</en>
   */
  private List foregroundIntervalValueMarkers ;

  /**
   * <zh>value᷽ıǼ</zh>
   * <en>A list containing lists of markers for the value axes.</en>
   */
  private List backgroundIntervalValueMarkers ;

  /**
   * <zh>Ƿʾı־.</zh>
   * <en>Flag that indicates whether or not grid bands are visible.</en>
   */
  private boolean gridBandVisible ;

  /**
   * <zh>Paint.</zh>
   * <en>The paint used to color the grid bands (if the bands are visible).</en>
   */
  private transient Paint gridBandPaint ;

  /**
   * <zh>Paint.</zh>
   * <en>The paint used to fill the alternate grid bands.</en>
   */
  private transient Paint gridBandAlternatePaint ;

  /**
   * <zh>IRender.</zh>
   * <en>The render object.</en>
   */
  private IRenderWithAxis render ;

  /**
   * <zh>ݼ.</zh>
   * <en>Constructs an axis using a dataset.</en>
   *
   * @param dataset IMatrixDataset <zh>άݼ (Ϊnull).</zh>
   * <en>The two dimension matrix dataset (null not permitted).</en>
   */
  protected AbstractValueAxis(IMatrixDataset dataset) {
    this(null, dataset) ;
  }

  /**
   * <zh>ǩݼ.</zh>
   * <en>Constructs an axis using label and a dataset.</en>
   *
   * @param label <zh>ǩ.</zh>
   * <en>The axis label (null permitted).</en>
   * @param dataset IMatrixDataset <zh>άݼ (Ϊnull).</zh>
   * <en>The two dimension matrix dataset (null not permitted).</en>
   */
  protected AbstractValueAxis(String label, IMatrixDataset dataset) {
    super(label, dataset) ;

    this.axisRangeIncludeZero = DEFAULT_AXIS_TICK_INCLUDE_ZERO ;
    this.inverted = DEFAULT_INVERTED ;
    this.emptyDatasetRange = new NumberRange(0, 1) ;
    this.tickValueFormatter = null ;

    this.valueLineVisible = DEFAULT_VALUE_LINE_VISIBLE ;
    this.valueLineValue = Double.NaN ;
    this.valueLineStroke = DEFAULT_CROSSHAIR_STROKE ;
    this.valueLinePaint = DEFAULT_VALUE_PAINT ;

    this.gridBandVisible = DEFAULT_GRID_BAND_VISIBLE ;
    this.gridBandPaint = DEFAULT_GRID_BAND_PAINT ;
    this.gridBandAlternatePaint = DEFAULT_GRID_BAND_ALTERNATE_PAINT ;

    this.foregroundValueMarkers = new ArrayList() ;
    this.backgroundValueMarkers = new ArrayList() ;

    this.foregroundIntervalValueMarkers = new ArrayList() ;
    this.backgroundIntervalValueMarkers = new ArrayList() ;

    setLowerMargin(0.0) ;
  }

  /**
   * <zh>ؼĿ̶ʱǷ0.</zh>
   * <en>Returns the flag whether or not include zero when calculate the axis data range.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en> The flag.</en>
   */
  public boolean isAxisRangeIncludeZero() {
    return this.axisRangeIncludeZero ;
  }

  /**
   * <zh>üĿ̶ʱǷ0,м
   * {@link com.ijchart.xychart.event.AxisChangeEvent}¼.</zh>
   * <en>Sets the flag whether or not include zero when calculate the axis ticks,and
   * notifies registered listeners that the axis has changed.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>The new value of the flag.</en>
   */
  public void setAxisRangeIncludeZero(boolean flag) {
    if (this.axisRangeIncludeZero != flag) {
      this.axisRangeIncludeZero = flag ;
      fireAxisChanged() ;
    }
  }

  /**
   * <zh>Ƿתֵ.</zh>
   * <en>Returns a flag that controls the direction of values on the axis.
   * <P>
   * For a regular axis, values increase from left to right (for a horizontal
   * axis) and bottom to top (for a vertical axis).  When the axis is
   * 'inverted', the values increase in the opposite direction.</en>
   *
   * @return <zh>ֵ.</zh>
   * <en> The flag.</en>
   */
  public boolean isInverted() {
    return this.inverted ;
  }

  /**
   * <zh>Ƿתֵ,м{@link com.ijchart.xychart.event.AxisChangeEvent}¼.</zh>
   * <en>Sets a flag that controls the direction of values on the axis,and
   * notifies registered listeners that the axis has changed.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en> The flag.</en>
   */
  public void setInverted(boolean flag) {
    if (this.inverted != flag) {
      this.inverted = flag ;
      fireAxisChanged() ;
    }
  }

  /**
   * <zh>ݼΪʱݷΧ.</zh>
   * <en>Returns the default range is used when the dataset is empty and the axis needs
   * to determine the auto range.</en>
   *
   * @return NumberRange <zh>᷶Χ.</zh>
   * <en>The range.</en>
   */
  public NumberRange getEmptyDatasetRange() {
    return this.emptyDatasetRange ;
  }

  /**
   * <zh>ݼΪʱݷΧ,м{@link com.ijchart.xychart.event.AxisChangeEvent}
   * ¼.</zh>
   * <en>Sets the default range is used when the dataset is empty and the axis needs
   * to determine the auto range,and notifies registered listeners that the axis has changed.</en>
   *
   * @param range NumberRange <zh>(Ϊnull).</zh>
   * <en>(null not permitted).</en>
   */
  public void setEmptyDatasetRange(NumberRange range) {
    if (range == null) {
      throw new IllegalArgumentException("Null 'range' argument.") ;
    }

    if (!this.emptyDatasetRange.equals(range)) {
      this.emptyDatasetRange = range ;
      fireAxisChanged() ;
    }
  }

  /**
   * <zh>ڸʽǩֵĸʽ.</zh>
   * <en>Returns the tick value formatter.</en>
   *
   * @return IFormatter
   */
  public IFormatter getTickValueFormatter() {
    return this.tickValueFormatter ;
  }

  /**
   * <zh>ڸʽǩֵĸʽ,м
   * {@link com.ijchart.xychart.event.AxisChangeEvent}¼.</zh>
   * <en>Sets the tick value formatter,and notifies registered listeners
   * that the axis has changed.</en>
   *
   * @param tickValueFormatter IFormatter <zh>ʽǩֵĸʽ.</zh>
   * <en>The tick value formatter.</en>
   */
  public void setTickValueFormatter(IFormatter tickValueFormatter) {
    if (!(this.tickValueFormatter == null && tickValueFormatter == null)) {
      this.tickValueFormatter = tickValueFormatter ;

      fireAxisChanged() ;
    }
  }

  /**
   * <zh>Ƿvalueֵ.</zh>
   * <en>Returns the flag that controls whether or not a value line is drawn.</en>
   *
   * @return boolean <zh>booleanֵ.</zh>
   * <en>The flag.</en>
   */
  public boolean isValueLineVisible() {
    return this.valueLineVisible ;
  }

  /**
   * <zh>Ƿvalueֵ,м
   * {@link com.ijchart.xychart.event.AxisChangeEvent}¼.</zh>
   * <en>Sets the flag indicating whether or not the value line is visible,and notifies registered
   *  listeners that the axis has changed.</en>
   *
   * @param flag boolean <zh>booleanֵ.</zh>
   * <en>The new value of the flag.</en>
   */
  public void setValueLineVisible(boolean flag) {
    if (this.valueLineVisible != flag) {
      this.valueLineVisible = flag ;
      fireAxisChanged() ;
    }
  }

  /**
   * <zh>valueӦֵߵֵ.</zh>
   * <en>Returns the value line value.</en>
   *
   * @return boolean <zh>booleanֵ.</zh>
   * <en>The value.</en>
   */
  public double getValueLineValue() {
    return this.valueLineValue ;
  }

  /**
   * <zh>valueֵ᷽ߵֵ,м{@link com.ijchart.xychart.event.AxisChangeEvent}
   * ¼.</zh>
   * <en>Sets the value line value and, if requested, sends an
   * {@link com.ijchart.xychart.event.AxisChangeEvent} to
   * all registered listeners (but only if the value line is visible).</en>
   *
   * @param value double <zh>ֵ.</zh>
   * <en>The new value.</en>
   */
  public void setValueLineValue(double value) {
    if (this.valueLineValue != value) {
      this.valueLineValue = value ;
      if (isValueLineVisible()) {
        fireAxisChanged() ;
      }
    }
  }

  /**
   * <zh>ػֵߵstroke.</zh>
   * <en>Returns the pen-style (Stroke) used to draw the value line (if visible).</en>
   *
   * @return Stroke <zh>stroke.</zh>
   * <en>The value line stroke (never null).</en>
   */
  public Stroke getValueLineStroke() {
    return this.valueLineStroke ;
  }

  /**
   * <zh>ûֵߵstroke,м{@link com.ijchart.xychart.event.AxisChangeEvent}
   * ¼.</zh>
   * <en>Sets the pen-style (Stroke) used to draw the value crosshair (if visible),
   * and sends an {@link com.ijchart.xychart.event.AxisChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>ֵߵstroke (Ϊnull).</zh>
   * <en>The new value line stroke (null not permitted).</en>
   */
  public void setValueLineStroke(Stroke stroke) {
    if (stroke == null) {
      throw new IllegalArgumentException("Null 'stroke' argument.") ;
    }

    this.valueLineStroke = stroke ;
    fireAxisChanged() ;
  }

  /**
   * <zh>ػֵߵpaint.</zh>
   * <en>Returns the paint used to draw the value line.</en>
   *
   * @return Paint <zh>paint.</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getValueLinePaint() {
    return this.valueLinePaint ;
  }

  /**
   * <zh>ûֵߵpaint,м{@link com.ijchart.xychart.event.AxisChangeEvent}
   * ¼.</zh>
   * <en>Sets the paint used to draw the value line (if visible) and
   * sends an {@link com.ijchart.xychart.event.AxisChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setValueLinePaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.valueLinePaint = paint ;
    fireAxisChanged() ;
  }

  /**
   * <zh>ؿǷʾı־.</zh>
   * <en>Returns the flag that indicates whether or not grid bands are visible.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>The boolean.</en>
   */
  public boolean isGridBandVisible() {
    return this.gridBandVisible ;
  }

  /**
   * <zh>ÿǷʾı־,м
   * {@link com.ijchart.xychart.event.AxisChangeEvent}¼.</zh>
   * <en>Sets the flag that indicates whether or not grid bands are visible and notifies registered
   *  listeners that the axis has changed.</en>
   *
   * @param visible boolean <zh>ֵ.</zh>
   * <en>The boolean.</en>
   */
  public void setGridBandVisible(boolean visible) {
    if (this.gridBandVisible != visible) {
      this.gridBandVisible = visible ;
      fireAxisChanged() ;
    }
  }

  /**
   * <zh>ػPaint.</zh>
   * <en>Returns the paint used to color the grid bands.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getGridBandPaint() {
    return this.gridBandPaint ;
  }

  /**
   * <zh>ûPaint,м
   * {@link com.ijchart.xychart.event.AxisChangeEvent}¼.</zh>
   * <en>Sets the paint used to color the grid bands and notifies registered listeners that the
   *  axis has changed.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint  (null not permitted).</en>
   */
  public void setGridBandPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.gridBandPaint = paint ;
    fireAxisChanged() ;
  }

  /**
   * <zh>ػPaint.</zh>
   * <en>Returns the paint used to fill the alternate grid bands.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getGridBandAlternatePaint() {
    return this.gridBandAlternatePaint ;
  }

  /**
   * <zh>ûPaint,м
   * {@link com.ijchart.xychart.event.AxisChangeEvent}¼.</zh>
   * <en>Sets the paint used to fill the alternate grid bands and notifies registered listeners
   *  that the axis has changed.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint  (null not permitted).</en>
   */
  public void setGridBandAlternatePaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.gridBandAlternatePaint = paint ;
    fireAxisChanged() ;
  }

  /**
   * <zh>IRender.</zh>
   * <en>Returns the render object.</en>
   *
   * @return IRenderWithAxis <zh>IRender (Ϊnull).</zh>
   * <en>The render object (maybe null).</en>
   */
  public IRenderWithAxis getRender() {
    return this.render ;
  }

  /**
   * <zh>IRender.</zh>
   * <en>The render object.</en>
   *
   * @param render IRenderWithAxis <zh>IRender (Ϊnull).</zh>
   * <en>The render object (null permitted).</en>
   */
  public void setRender(IRenderWithAxis render) {
    if (!(this.render == null && render == null)) {
      this.render = render ;
      fireAxisChanged() ;
    }
  }

  /**
   * <zh>ֵ,IAxis.draw()֮.</zh>
   * <en>Draw the value line,only call after IAxis.draw().</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param dataArea Rectangle2D <zh>ͼdata() (Ϊnull).</zh>
   * <en>The area within which the data should be drawn (null not permitted).</en>
   * @param edge RectangleEdge <zh>λ (Ϊnull).</zh>
   * <en>The axis position (null not permitted).</en>
   * @param shapeCollection RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   */
  public abstract void drawValueLine(Graphics2D g2, Rectangle2D dataArea, RectangleEdge edge,
                                     RenderShapeCollection shapeCollection) ;

  /**
   * <zh>ĳֵӦλ,IAxis.draw()֮.</zh>
   * <en>Converts a data value to a coordinate in Java2D space, assuming that the
   * axis runs along one edge of the specified dataArea.</en>
   *
   * @param dataArea Rectangle2D <zh>ͼdata() (Ϊnull).</zh>
   * <en>The area within which the data should be drawn (null not permitted).</en>
   * @param edge RectangleEdge <zh>λ (Ϊnull).</zh>
   * <en>The axis position (null not permitted).</en>
   * @param value double <zh>ֵ.</zh>
   * <en>The data value.</en>
   * @param forceInDataArea boolean <zh>ǷǿdataArea֮.</zh>
   * <en>Whether or not the return value force in the region dataArea.</en>
   *
   * @return double <zh>λ.</zh>
   * <en>The Java2D coordinate.</en>
   */
  public abstract double valueToJava2D(Rectangle2D dataArea, RectangleEdge edge, double value,
                                       boolean forceInDataArea) ;

  /**
   * <zh>ĳֵӦλ,IAxis.draw()֮.</zh>
   * <en>Converts a data value to a coordinate in Java2D space, assuming that the
   * axis runs along one edge of the specified dataArea.</en>
   *
   * @param dataArea Rectangle2D <zh>ͼdata() (Ϊnull).</zh>
   * <en>The area within which the data should be drawn (null not permitted).</en>
   * @param edge RectangleEdge <zh>λ (Ϊnull).</zh>
   * <en>The axis position (null not permitted).</en>
   * @param value double <zh>ֵ.</zh>
   * <en>The data value.</en>
   * @param forceInDataArea boolean <zh>ǷǿdataArea֮.</zh>
   * <en>Whether or not the return value force in the region dataArea.</en>
   * @param expandAxisRange boolean <zh>ǷչķΧ.</zh>
   * <en>Whether or not expand the range of axis.</en>
   *
   * @return double <zh>λ.</zh>
   * <en>The Java2D coordinate.</en>
   */
  public abstract double valueToJava2D(Rectangle2D dataArea, RectangleEdge edge, double value,
                                       boolean forceInDataArea, boolean expandAxisRange) ;

  /**
   * <zh>õݼݷΧ.</zh>
   * <en>Returns the value range of a dataset.</en>
   *
   * @return NumberRange <zh>ݷΧ (Ϊnull).</zh>
   * <en>The value range of a dataset (never null).</en>
   */
  public abstract NumberRange getDatasetRange() ;

  /**
   * <zh>һvalue marker,м{@link com.ijchart.xychart.event.AxisChangeEvent}
   * ¼.</zh>
   * <en>Adds a value marker for display by a particular render.</en>
   *
   * @param layer Layer <zh>Ƶ˳(ǰ򱳾) (Ϊnull).</zh>
   * <en>The layer (foreground or background) (null not permitted).</en>
   * @param marker ValueMarker <zh>ValueMarker (Ϊnull).</zh>
   * <en>The marker (null not permitted).</en>
   */
  public void addValueMarker(Layer layer, ValueMarker marker) {
    if (layer == null) {
      throw new IllegalArgumentException("Null 'layer' not permitted.") ;
    }
    if (marker == null) {
      throw new IllegalArgumentException("Null 'marker' not permitted.") ;
    }

    if (layer.equals(Layer.FOREGROUND)) {
      this.foregroundValueMarkers.add(marker) ;
    } else if (layer.equals(Layer.BACKGROUND)) {
      this.backgroundValueMarkers.add(marker) ;
    }

    marker.addChangeListener(this) ;
    fireAxisChanged() ;
  }

  /**
   * <zh>Layer󷵻еvalue marker.</zh>
   * <en>Returns a collection of value markers for specified layer.</en>
   *
   * @param layer <zh>Layer (Ϊnull).</zh>
   * <en>The layer (null not permitted).</en>
   *
   * @return Collection <zh>value marker󼯺 (Ϊnull).</zh>
   * <en>A collection of markers (possibly null).</en>
   */
  public Collection getValueMarkers(Layer layer) {
    if (layer == null) {
      throw new IllegalArgumentException("Null 'layer' not permitted.") ;
    }

    if (layer.equals(Layer.FOREGROUND)) {
      return Collections.unmodifiableCollection(this.foregroundValueMarkers) ;
    } else if (layer.equals(Layer.BACKGROUND)) {
      return Collections.unmodifiableCollection(this.backgroundValueMarkers) ;
    }

    return null ;
  }

  /**
   * <zh>ָLayerɾmarker.</zh>
   * <en>Removes a marker for a specific layer and sends an
   * {@link com.ijchart.xychart.event.AxisChangeEvent} to all registered listeners.</en>
   *
   * @param layer Layer <zh>layer (Ϊnull).</zh>
   * <en>The layer (foreground or background) (null not permitted).</en>
   * @param marker ValueMarker <zh>ValueMarker (Ϊnull).</zh>
   * <en>The marker (null not permitted).</en>
   *
   * @return <zh>Ƿɾ˴˶.</zh>
   * <en>A boolean indicating whether or not the marker was actually removed.</en>
   */
  public boolean removeValueMarker(Layer layer, ValueMarker marker) {
    if (layer == null) {
      throw new IllegalArgumentException("Null 'layer' not permitted.") ;
    }
    if (marker == null) {
      throw new IllegalArgumentException("Null 'marker' not permitted.") ;
    }

    boolean removed = false ;

    if (layer.equals(Layer.FOREGROUND)) {
      removed = this.foregroundValueMarkers.remove(marker) ;
    } else if (layer.equals(Layer.BACKGROUND)) {
      removed = this.backgroundValueMarkers.remove(marker) ;
    }

    if (removed) {
      marker.removeChangeListener(this) ;
      fireAxisChanged() ;
    }

    return removed ;
  }

  /**
   * <zh>ɾеvalue marker,м{@link com.ijchart.xychart.event.AxisChangeEvent}
   * ¼.</zh>
   * <en>Clears all the value markers for the plot and sends an
   * {@link com.ijchart.xychart.event.AxisChangeEvent} to all registered listeners.</en>
   *
   * @param layer Layer <zh>Ƶ˳(ǰ򱳾) (Ϊnull).</zh>
   * <en>The layer (foreground or background) (null not permitted).</en>
   */
  public void clearValueMarkers(Layer layer) {
    if (layer == null) {
      throw new IllegalArgumentException("Null 'layer' not permitted.") ;
    }

    boolean notify = false ;

    if (layer.equals(Layer.FOREGROUND)) {
      for (int i = 0 ; i < this.foregroundValueMarkers.size() ; i++) {
        AbstractMarker marker = (AbstractMarker)this.foregroundValueMarkers.get(i) ;
        marker.removeChangeListener(this) ;
      }

      notify = this.foregroundValueMarkers.size() > 0 ;
      this.foregroundValueMarkers.clear() ;
    }

    if (layer.equals(Layer.BACKGROUND)) {
      for (int i = 0 ; i < this.backgroundValueMarkers.size() ; i++) {
        AbstractMarker marker = (AbstractMarker)this.backgroundValueMarkers.get(i) ;
        marker.removeChangeListener(this) ;
      }

      notify = this.backgroundValueMarkers.size() > 0 ;
      this.backgroundValueMarkers.clear() ;
    }

    if (notify) {
      fireAxisChanged() ;
    }
  }

  /**
   * <zh>һinterval value marker,м
   * {@link com.ijchart.xychart.event.AxisChangeEvent}¼.</zh>
   * <en>Adds a interval value marker for display by a particular render.</en>
   *
   * @param layer Layer <zh>Ƶ˳(ǰ򱳾) (Ϊnull).</zh>
   * <en>The layer (foreground or background) (null not permitted).</en>
   * @param marker IntervalValueMarker <zh>IntervalValueMarker (Ϊnull).</zh>
   * <en>The marker (null not permitted).</en>
   */
  public void addIntervalValueMarker(Layer layer, IntervalValueMarker marker) {
    if (layer == null) {
      throw new IllegalArgumentException("Null 'layer' not permitted.") ;
    }
    if (marker == null) {
      throw new IllegalArgumentException("Null 'marker' not permitted.") ;
    }

    if (layer.equals(Layer.FOREGROUND)) {
      this.foregroundIntervalValueMarkers.add(marker) ;
    } else if (layer.equals(Layer.BACKGROUND)) {
      this.backgroundIntervalValueMarkers.add(marker) ;
    }

    marker.addChangeListener(this) ;
    fireAxisChanged() ;
  }

  /**
   * <zh>Layer󷵻еinterval value marker.</zh>
   * <en>Returns a collection of interval value markers for specified layer.</en>
   *
   * @param layer <zh>Layer (Ϊnull).</zh>
   * <en>The layer (null not permitted).</en>
   *
   * @return Collection <zh>value marker󼯺 (Ϊnull).</zh>
   * <en>A collection of markers (possibly null).</en>
   */
  public Collection getIntervalValueMarkers(Layer layer) {
    if (layer == null) {
      throw new IllegalArgumentException("Null 'layer' not permitted.") ;
    }

    if (layer.equals(Layer.FOREGROUND)) {
      return Collections.unmodifiableCollection(this.foregroundIntervalValueMarkers) ;
    } else if (layer.equals(Layer.BACKGROUND)) {
      return Collections.unmodifiableCollection(this.backgroundIntervalValueMarkers) ;
    }

    return null ;
  }

  /**
   * <zh>ָLayerɾinterval marker.</zh>
   * <en>Removes a interval marker for a specific layer and sends an
   * {@link com.ijchart.xychart.event.AxisChangeEvent} to all registered listeners.</en>
   *
   * @param layer Layer <zh>layer (Ϊnull).</zh>
   * <en>The layer (foreground or background) (null not permitted).</en>
   * @param marker IntervalValueMarker <zh>IntervalValueMarker (Ϊnull).</zh>
   * <en>The marker (null not permitted).</en>
   *
   * @return <zh>Ƿɾ˴˶.</zh>
   * <en>A boolean indicating whether or not the marker was actually removed.</en>
   */
  public boolean removeValueMarker(Layer layer, IntervalValueMarker marker) {
    if (layer == null) {
      throw new IllegalArgumentException("Null 'layer' not permitted.") ;
    }
    if (marker == null) {
      throw new IllegalArgumentException("Null 'marker' not permitted.") ;
    }

    boolean removed = false ;

    if (layer.equals(Layer.FOREGROUND)) {
      removed = this.foregroundIntervalValueMarkers.remove(marker) ;
    } else if (layer.equals(Layer.BACKGROUND)) {
      removed = this.backgroundIntervalValueMarkers.remove(marker) ;
    }

    if (removed) {
      marker.removeChangeListener(this) ;
      fireAxisChanged() ;
    }

    return removed ;
  }

  /**
   * <zh>ɾеinterval value marker,м
   * {@link com.ijchart.xychart.event.AxisChangeEvent}¼.</zh>
   * <en>Clears all the interval value markers for the plot and sends an
   * {@link com.ijchart.xychart.event.AxisChangeEvent} to all registered listeners.</en>
   *
   * @param layer Layer <zh>Ƶ˳(ǰ򱳾) (Ϊnull).</zh>
   * <en>The layer (foreground or background) (null not permitted).</en>
   */
  public void clearIntervalValueMarkers(Layer layer) {
    if (layer == null) {
      throw new IllegalArgumentException("Null 'layer' not permitted.") ;
    }

    boolean notify = false ;

    if (layer.equals(Layer.FOREGROUND)) {
      for (int i = 0 ; i < this.foregroundIntervalValueMarkers.size() ; i++) {
        AbstractMarker marker = (AbstractMarker)this.foregroundIntervalValueMarkers.get(i) ;
        marker.removeChangeListener(this) ;
      }

      notify = this.foregroundIntervalValueMarkers.size() > 0 ;
      this.foregroundIntervalValueMarkers.clear() ;
    }

    if (layer.equals(Layer.BACKGROUND)) {
      for (int i = 0 ; i < this.backgroundIntervalValueMarkers.size() ; i++) {
        AbstractMarker marker = (AbstractMarker)this.backgroundIntervalValueMarkers.get(i) ;
        marker.removeChangeListener(this) ;
      }

      notify = this.backgroundIntervalValueMarkers.size() > 0 ;
      this.backgroundIntervalValueMarkers.clear() ;
    }

    if (notify) {
      fireAxisChanged() ;
    }
  }

  /**
   * <zh>յmarker仯¼.</zh>
   * <en>Receives notification of an marker change event.</en>
   *
   * @param event <zh>marker仯¼.</zh>
   * <en>The marker change event.</en>
   */
  public void markerChanged(MarkerChangeEvent event) {
    fireAxisChanged() ;
  }

  /**
   * <zh>ԱǷһ.</zh>
   * <en>Tests the axis for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return true or false.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof AbstractValueAxis)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    AbstractValueAxis that = (AbstractValueAxis) obj ;
    if (this.axisRangeIncludeZero != that.axisRangeIncludeZero) {
      return false ;
    }
    if (this.inverted != that.inverted) {
      return false ;
    }
    if (!EqualsUtil.equals(this.emptyDatasetRange, that.emptyDatasetRange)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.tickValueFormatter, that.tickValueFormatter)) {
      return false ;
    }
    if (this.valueLineVisible != that.valueLineVisible) {
      return false ;
    }
    if (this.valueLineValue != that.valueLineValue) {
      return false ;
    }
    if (!EqualsUtil.equals(this.valueLineStroke, that.valueLineStroke)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.valueLinePaint, that.valueLinePaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.foregroundValueMarkers, that.foregroundValueMarkers)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.backgroundValueMarkers, that.backgroundValueMarkers)) {
      return false ;
    }
    if (this.gridBandVisible != that.gridBandVisible) {
      return false ;
    }
    if (!EqualsUtil.equals(this.gridBandPaint, that.gridBandPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.gridBandAlternatePaint, that.gridBandAlternatePaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.foregroundIntervalValueMarkers, that.foregroundIntervalValueMarkers)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.backgroundIntervalValueMarkers, that.backgroundIntervalValueMarkers)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.render, that.render)) {
      return false ;
    }

    return true ;
  }

  /**
   * Returns a clone of the object.
   *
   * @return A clone.
   *
   * @throws CloneNotSupportedException if some component of the axis does
   *         not support cloning.
   */
  public Object clone() throws CloneNotSupportedException {
    AbstractValueAxis clone = (AbstractValueAxis)super.clone() ;
    return clone ;
  }

  /**
   * <zh>hashCode.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = result * 29 + HashcodeUtil.hashCode(this.axisRangeIncludeZero) ;
    result = result * 29 + HashcodeUtil.hashCode(this.inverted) ;
    result = result * 29 + HashcodeUtil.hashCode(this.emptyDatasetRange) ;
    result = result * 29 + HashcodeUtil.hashCode(this.tickValueFormatter) ;
    result = 29 * result + HashcodeUtil.hashCode(valueLineVisible) ;
    result = 29 * result + (int) valueLineValue ;
    result = 29 * result + HashcodeUtil.hashCode(valueLineStroke) ;
    result = 29 * result + HashcodeUtil.hashCode(valueLinePaint) ;
    result = 29 * result + HashcodeUtil.hashCode(this.foregroundValueMarkers) ;
    result = 29 * result + HashcodeUtil.hashCode(this.backgroundValueMarkers) ;
    result = 29 * result + HashcodeUtil.hashCode(this.gridBandVisible) ;
    result = 29 * result + HashcodeUtil.hashCode(this.gridBandPaint) ;
    result = 29 * result + HashcodeUtil.hashCode(this.gridBandAlternatePaint) ;
    result = 29 * result + HashcodeUtil.hashCode(this.foregroundIntervalValueMarkers) ;
    result = 29 * result + HashcodeUtil.hashCode(this.backgroundIntervalValueMarkers) ;
    result = 29 * result + HashcodeUtil.hashCode(this.render) ;

    return result ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return A string representation of the object.
   */
  public String toString() {
    return "AbstractValueAxis" ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writeStroke(stream, this.valueLineStroke) ;
    SerialUtil.writePaint(stream, this.valueLinePaint) ;
    SerialUtil.writePaint(stream, this.gridBandPaint) ;
    SerialUtil.writePaint(stream, this.gridBandAlternatePaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.valueLineStroke = SerialUtil.readStroke(stream) ;
    this.valueLinePaint = SerialUtil.readPaint(stream) ;
    this.gridBandPaint = SerialUtil.readPaint(stream) ;
    this.gridBandAlternatePaint = SerialUtil.readPaint(stream) ;
  }

}
