package com.ijchart.xychart.axis.value ;

import java.io.Serializable ;

import com.ijchart.xychart.constant.XYValueAxisType ;
import com.ijchart.xychart.data.IntervalXYNumberChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.exception.InvalidChartDataException ;
import com.ijchart.xychart.range.NumberRange ;
import com.ijchart.xychart.tickunit.DateTickUnit ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>XY plotʹõֵ,ÿһ{@link IntervalXYNumberChartData}.</zh>
 * <en>The number value axis used for XY plot,every data is {@link IntervalXYNumberChartData}.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class IntervalXYValueAxis extends AbstractNumberValueAxisWithType implements Cloneable,
  Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -7844512034828849253L ;

  /**
   * <zh>ݼ.</zh>
   * <en>Constructs an axis using a dataset.</en>
   *
   * @param dataset IMatrixDataset <zh>άݼ (Ϊnull).</zh>
   * <en>The two dimension matrix dataset (null not permitted).</en>
   * @param valueAxisType XYValueAxisType <zh>value (Ϊnull).</zh>
   * <en>The type of value axis (null not permitted).</en>
   * @param isDateAxis boolean <zh>Ƿ͵.</zh>
   * <en>Whether or not is a date axis.</en>
   */
  public IntervalXYValueAxis(IMatrixDataset dataset, XYValueAxisType valueAxisType,
                             boolean isDateAxis) {
    this(null, dataset, valueAxisType, isDateAxis) ;
  }

  /**
   * <zh>ǩݼ.</zh>
   * <en>Constructs an axis using label and a dataset.</en>
   *
   * @param label <zh>ǩ.</zh>
   * <en>The axis label (null permitted).</en>
   * @param dataset IMatrixDataset <zh>άݼ (Ϊnull).</zh>
   * <en>The two dimension matrix dataset (null not permitted).</en>
   * @param valueAxisType XYValueAxisType <zh>value (Ϊnull).</zh>
   * <en>The type of value axis (null not permitted).</en>
   * @param isDateAxis boolean <zh>Ƿ͵.</zh>
   * <en>Whether or not is a date axis.</en>
   */
  public IntervalXYValueAxis(String label, IMatrixDataset dataset, XYValueAxisType valueAxisType,
                             boolean isDateAxis) {
    super(label, dataset, valueAxisType) ;

    if (isDateAxis) {
      this.tickUnit = new DateTickUnit(null) ;
    }
  }

  /**
   * <zh>õݼݷΧ.</zh>
   * <en>Returns the value range of a dataset.</en>
   *
   * @return NumberRange <zh>ݷΧ (Ϊnull).</zh>
   * <en>The value range of a dataset (never null).</en>
   */
  public NumberRange getDatasetRange() {
    if (getUserDefineDatasetRange() != null) {
      return getUserDefineDatasetRange() ;
    }

    double minValue = Double.NaN ;
    double maxValue = Double.NaN ;
    IMatrixDataset dataset = getDataset() ;

    boolean isX = false, isY = false ; //Ƿxyֵ
    if (getValueAxisType().equals(XYValueAxisType.XY_VALUE_AXIS_X)) {
      isX = true ;
    } else if (getValueAxisType().equals(XYValueAxisType.XY_VALUE_AXIS_Y)) {
      isY = true ;
    }

    if (dataset != null) {
      int rows = dataset.getRowCount() ;
      for (int i = 0 ; i < rows ; i++) {

        int currRowCols = dataset.getColumnCount(i) ;
        for (int j = 0 ; j < currRowCols ; j++) {
          Object data = dataset.getData(i, j) ;
          if (data != null) {
            if (!(data instanceof IntervalXYNumberChartData)) {
              throw new InvalidChartDataException(
                "Invalid chart data type,musts be 'IntervalXYNumberChartData'.") ;
            } else {
              double startValue = Double.NaN ;
              double endValue = Double.NaN ;
              if (isX) { //ȡxֵ
                startValue = ((IntervalXYNumberChartData) data).getStartXValue() ;
                endValue = ((IntervalXYNumberChartData) data).getEndXValue() ;
              } else if (isY) { //ȡyֵ
                startValue = ((IntervalXYNumberChartData) data).getStartYValue() ;
                endValue = ((IntervalXYNumberChartData) data).getEndYValue() ;
              }

              if (Double.isNaN(minValue)) {
                if (Double.isNaN(startValue) && !Double.isNaN(endValue)) {
                  minValue = endValue ;
                } else if (!Double.isNaN(startValue) && Double.isNaN(endValue)) {
                  minValue = endValue ;
                } else if (!Double.isNaN(startValue) && !Double.isNaN(endValue)) {
                  minValue = Math.min(startValue, endValue) ;
                }
              } else {
                if (Double.isNaN(startValue) && !Double.isNaN(endValue)) {
                  minValue = Math.min(minValue, endValue) ;
                } else if (!Double.isNaN(startValue) && Double.isNaN(endValue)) {
                  minValue = Math.min(minValue, endValue) ;
                } else if (!Double.isNaN(startValue) && !Double.isNaN(endValue)) {
                  minValue = Math.min(minValue, Math.min(startValue, endValue)) ;
                }
              }

              if (Double.isNaN(maxValue)) {
                if (Double.isNaN(startValue) && !Double.isNaN(endValue)) {
                  maxValue = endValue ;
                } else if (!Double.isNaN(startValue) && Double.isNaN(endValue)) {
                  maxValue = endValue ;
                } else if (!Double.isNaN(startValue) && !Double.isNaN(endValue)) {
                  maxValue = Math.max(startValue, endValue) ;
                }
              } else {
                if (Double.isNaN(startValue) && !Double.isNaN(endValue)) {
                  maxValue = Math.max(maxValue, endValue) ;
                } else if (!Double.isNaN(startValue) && Double.isNaN(endValue)) {
                  maxValue = Math.max(maxValue, endValue) ;
                } else if (!Double.isNaN(startValue) && !Double.isNaN(endValue)) {
                  maxValue = Math.max(maxValue, Math.max(startValue, endValue)) ;
                }
              }

            }
          }
        }
      }
    }

    if (Double.isNaN(minValue)) {
      minValue = 0 ;
    }
    if (Double.isNaN(maxValue)) {
      maxValue = 0 ;
    }

    return new NumberRange(minValue, maxValue) ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates whether some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument;
   * false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return false ;
    }
    if (!(obj instanceof IntervalXYValueAxis)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    return super.hashCode() ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object. </en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "IntervalXYValueAxis" ;
  }

}
