package com.ijchart.xychart.axis.value ;

import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.chartinfo.AxisShapeCollection ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.render.IStackedRender ;
import com.ijchart.xychart.ui.IEffect3D ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ջ͵ͼʹõ3Dֵ.</zh>
 * <en>The 3D axis used for stack bar chart.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class Stacked3DNumberValueAxis extends StackedNumberValueAxis implements Cloneable,
  Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -1826453119551818560L ;

  /**
   * <zh>ջ״ͼƶ.</zh>
   * <en>The stacked bar render object.</en>
   */
  private IStackedRender stackedRender ;

  /**
   * <zh>3DЧƫ.</zh>
   * <en>The 3D effect.</en>
   */
  private IEffect3D effect3D ;

  /**
   * <zh>ǿ2Dʽ.</zh>
   * <en>A boolean flag controls whether or not draw as 2D axis.</en>
   */
  private boolean force2D ;

  /**
   * <zh>ݼ.</zh>
   * <en>Constructs an axis using a dataset.</en>
   *
   * @param dataset IMatrixDataset <zh>άݼ (Ϊnull).</zh>
   * <en>The two dimension matrix dataset (null not permitted).</en>
   * @param render IStackedRender <zh>ջ״ͼƶ (Ϊnull).</zh>
   * <en>The stacked bar render object (null not permitted).</en>
   * @param effect3D IEffect3D <zh>3DЧƫ.</zh>
   * <en>The 3D effect.</en>
   */
  public Stacked3DNumberValueAxis(IMatrixDataset dataset, IStackedRender render, IEffect3D effect3D) {
    this(null, dataset, render, effect3D) ;
  }

  /**
   * <zh>ǩݼ.</zh>
   * <en>Constructs an axis using label and a dataset.</en>
   *
   * @param label <zh>ǩ.</zh>
   * <en>The axis label (null permitted).</en>
   * @param dataset IMatrixDataset <zh>άݼ (Ϊnull).</zh>
   * <en>The two dimension matrix dataset (null not permitted).</en>
   * @param render IStackedRender <zh>ջ״ͼƶ (Ϊnull).</zh>
   * <en>The stacked bar render object (null not permitted).</en>
   * @param effect3D IEffect3D <zh>3DЧƫ.</zh>
   * <en>The 3D effect.</en>
   */
  public Stacked3DNumberValueAxis(String label, IMatrixDataset dataset, IStackedRender render,
                                  IEffect3D effect3D) {
    super(label, dataset, render) ;

    this.stackedRender = render ;
    this.effect3D = effect3D ;
    this.force2D = false ;
  }

  /**
   * <zh>״ͼƶ.</zh>
   * <en>Returns the stacked bar render.</en>
   *
   * @return StackedBarRender <zh>ջ״ͼƶ (Ϊnull).</zh>
   * <en>The stacked bar render object (never null).</en>
   */
  public IStackedRender getStackedBarRender() {
    return this.stackedRender ;
  }

  /**
   * <zh>״ͼƶ,м{@link com.ijchart.xychart.event.AxisChangeEvent}¼.</zh>
   * <en>Sets the stacked bar render object and sends a
   *  {@link com.ijchart.xychart.event.AxisChangeEvent} to all registered listeners.</en>
   *
   * @param render IStackedRender <zh>ջ״ͼƶ (Ϊnull).</zh>
   * <en>The stacked bar render object (null not permitted).</en>
   */
  public void setStackedBarRender(IStackedRender render) {
    if (render == null) {
      throw new IllegalArgumentException("Null 'render' argument.") ;
    }

    this.stackedRender = render ;
    fireAxisChanged() ;
  }

  /**
   * <zh>ػ3DЧʱX᷽ƫֵ.</zh>
   * <en>Returns the x-offset (in Java2D units) for the 3D effect.</en>
   *
   * @return double <zh>X᷽ƫֵ.</zh>
   * <en>The offset.</en>
   */
  public double getXOffset() {
    return this.effect3D.getXOffset() ;
  }

  /**
   * <zh>ػ3DЧʱY᷽ƫֵ.</zh>
   * <en>Returns the y-offset (in Java2D units) for the 3D effect.</en>
   *
   * @return double <zh>Y᷽ƫֵ.</zh>
   * <en>The offset.</en>
   */
  public double getYOffset() {
    return this.effect3D.getYOffset() ;
  }

  /**
   * <zh>Ƿǿ2Dʽ.</zh>
   * <en>Returns whether or not draw as 2D axis.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isForce2D() {
    return this.force2D ;
  }

  /**
   * <zh>Ƿǿ2Dʽ,˷ᴥ仯¼.</zh>
   * <en>Sets whether or not draw as 2D axis and NOT send any event.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setForce2D(boolean flag) {
    this.force2D = flag ;
  }

  /**
   * <zh></zh>
   * <en>Draws an axis line in a specified area and edge.</en>
   *
   * @param axisArea Rectangle2D <zh>ʣĿհ.</zh>
   * <en>The empty chart area.</en>
   * @param edge RectangleEdge <zh>λ.</zh>
   * <en>The axis position.</en>
   * @param shapeCollection AxisShapeCollection <zh>ͼμ.</zh>
   * <en>Collects information about the axis (null permitted).</en>
   *
   * @return double <zh>ǰλ.</zh>
   * <en>The cursor location.</en>
   */
  protected double drawAxisLine(Rectangle2D axisArea, RectangleEdge edge,
                                AxisShapeCollection shapeCollection) {
    if (this.force2D) { //ǿ2Dʽ
      return super.drawAxisLine(axisArea, edge, shapeCollection) ;
    }

    if (edge.equals(RectangleEdge.TOP) || edge.equals(RectangleEdge.RIGHT)) {
      return super.drawAxisLine(axisArea, edge, shapeCollection) ;
    }

    return super.drawAxis3DLine(this.effect3D, axisArea, edge, shapeCollection) ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates whether some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument;
   * false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return false ;
    }
    if (!(obj instanceof Stacked3DNumberValueAxis)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    Stacked3DNumberValueAxis that = (Stacked3DNumberValueAxis) obj ;
    if (!EqualsUtil.equals(this.stackedRender, that.stackedRender)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.effect3D, that.effect3D)) {
      return false ;
    }
    if (this.force2D != that.force2D) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int hashCode = super.hashCode() ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.stackedRender) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.effect3D) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.force2D) ;

    return hashCode ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object. </en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[effect3D=" + this.effect3D + ",force2D=" + force2D + "]" ;
  }

}
