package com.ijchart.xychart.client ;

import java.awt.AWTEvent ;
import java.awt.Color ;
import java.awt.Dimension ;
import java.awt.Graphics ;
import java.awt.Graphics2D ;
import java.awt.Image ;
import java.awt.Insets ;
import java.awt.Paint ;
import java.awt.Point ;
import java.awt.RenderingHints ;
import java.awt.event.MouseEvent ;
import java.awt.event.MouseListener ;
import java.awt.event.MouseMotionListener ;
import java.awt.geom.AffineTransform ;
import java.awt.geom.Line2D ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Collection ;
import java.util.Collections ;
import java.util.EventListener ;
import java.util.Iterator ;
import java.util.List ;

import javax.swing.JPanel ;
import javax.swing.ToolTipManager ;
import javax.swing.event.EventListenerList ;

import com.ijchart.xychart.chartinfo.ChartShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.ChartShapeDrawType ;
import com.ijchart.xychart.event.ChartChangeEvent ;
import com.ijchart.xychart.event.ChartMouseEvent ;
import com.ijchart.xychart.event.ChartProgressEvent ;
import com.ijchart.xychart.event.ChartShapeChangeEvent ;
import com.ijchart.xychart.listener.IChartChangeListener ;
import com.ijchart.xychart.listener.IChartMouseListener ;
import com.ijchart.xychart.listener.IChartProgressListener ;
import com.ijchart.xychart.listener.IChartShapeChangeListener ;
import com.ijchart.xychart.plot.AbstractAxisPlot ;
import com.ijchart.xychart.plot.AbstractPlot ;
import com.ijchart.xychart.plot.NoAxisPlot ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ڻ{@link IJChart}panel,˶ʵ{@link IChartChangeListener},chartκα仯
 * ˶».</zh>
 * <en>A Swing GUI component for displaying a {@link IJChart} object.
 * The panel registers with the chart to receive notification of changes to any
 * component of the chart.The chart is redrawn automatically whenever this
 * notification is received.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class ChartPanel extends JPanel implements IChartChangeListener, IChartShapeChangeListener,
  IChartProgressListener, MouseListener, MouseMotionListener, Serializable {

  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -3787661459210554930L ;

  /**
   * <zh>ȱʡpanel.</zh>
   * <en>The default panel width..</en>
   */
  public static final int DEFAULT_WIDTH = 680 ;

  /**
   * <zh>ȱʡpanel߶.</zh>
   * <en>The default panel heigh.</en>
   */
  public static final int DEFAULT_HEIGHT = 420 ;

  /**
   * <zh>ȱʡСƿ.</zh>
   * <en>The default limit below which chart scaling kicks in.</en>
   */
  public static final int DEFAULT_MINIMUM_DRAW_WIDTH = 300 ;

  /**
   * <zh>ȱʡСƸ߶.</zh>
   * <en>The default limit below which chart scaling kicks in.</en>
   */
  public static final int DEFAULT_MINIMUM_DRAW_HEIGHT = 200 ;

  /**
   * <zh>ȱʡƿ.</zh>
   * <en>The default limit below which chart scaling kicks in.</en>
   */
  public static final int DEFAULT_MAXIMUM_DRAW_WIDTH = 800 ;

  /**
   * <zh>ȱʡƸ߶.</zh>
   * <en>The default limit below which chart scaling kicks in.</en>
   */
  public static final int DEFAULT_MAXIMUM_DRAW_HEIGHT = 600 ;

  /**
   * <zh>ȱʡž.</zh>
   * <en>The minimum size required to perform a zoom on a rectangle.</en>
   */
  public static final int DEFAULT_ZOOM_TRIGGER_DISTANCE = 10 ;

  /**
   * <zh>ȱʡķŴ.</zh>
   * <en>The default factor used to zoom in on an axis range.</en>
   */
  public static final double DEFAULT_ZOOM_IN_FACTOR = 0.5 ;

  /**
   * <zh>ȱʡС.</zh>
   * <en>The default factor used to zoom out on an axis range.</en>
   */
  public static final double DEFAULT_ZOOM_OUT_FACTOR = 2.0 ;

  /**
   * <zh>ȱʡǷٴֱ.</zh>
   * <en>A default flag that controls whether or not vertical tracing is enabled.</en>
   */
  public static final boolean DEFAULT_VERTICAL_AXIS_TRACE = false ;

  /**
   * <zh>ȱʡǷˮƽ.</zh>
   * <en>A default flag that controls whether or not horizontal tracing is enabled.</en>
   */
  public static final boolean DEFAULT_HORIZONTAL_AXIS_TRACE = false ;

  /**
   * <zh>ȱʡǷž.</zh>
   * <en>The default flag controls if the zoom rectangle is drawn as a border or filled.</en>
   */
  public static final boolean DEFAULT_FILL_ZOOM_RECTANGLE = true ;

  /**
   * <zh>ȱʡĻſı߿paint.</zh>
   * <en>The default paint used to draw the zoom rectangle outline.</en>
   */
  public static final Paint DEFAULT_ZOOM_BORDER_PAINT = Color.blue ;

  /**
   * <zh>ȱʡſpaint.</zh>
   * <en>The default paint used to fill the zoom rectangle.</en>
   */
  public static final Paint DEFAULT_ZOOM_FILL_PAINT = new Color(0, 0, 255, 63) ;

  /**
   * <zh>ȱʡǷ.</zh>
   * <en>Default zoomable.</en>
   */
  public static final boolean DEFAULT_ZOOMABLE = true ;

  /**
   * <zh>panelʾ{@link IJChart}.</zh>
   * <en>The chart that is displayed in the panel.</en>
   */
  private IJChart chart ;

  /**
   * <zh>¼.</zh>
   * <en>Storage for registered (chart) mouse listeners.</en>
   */
  private EventListenerList chartMouseListeners ;

  /**
   * <zh>{@link IJChart}ڴ.</zh>
   * <en>A buffer for the rendered chart.</en>
   */
  private Image chartBuffer ;

  /**
   * <zh>ˢ».</zh>
   * <en>A flag that indicates that the buffer should be refreshed.</en>
   */
  private boolean refreshBuffer ;

  /**
   * <zh>ڴĿ.</zh>
   * <en>The width of the chart buffer.</en>
   */
  private int chartBufferWidth ;

  /**
   * <zh>ڴĸ߶.</zh>
   * <en>The height of the chart buffer.</en>
   */
  private int chartBufferHeight ;

  /**
   * <zh>СĻƿ.</zh>
   * <en>The minimum width for drawing a chart.</en>
   */
  private double minimumDrawWidth ;

  /**
   * <zh>СĻƸ߶.</zh>
   * <en>The minimum height for drawing a chart.</en>
   */
  private double minimumDrawHeight ;

  /**
   * <zh>x᷽.</zh>
   * <en>The scale factor used to draw the chart.</en>
   */
  private double scaleX ;

  /**
   * <zh>y᷽.</zh>
   * <en>The scale factor used to draw the chart.</en>
   */
  private double scaleY ;

  /**
   * <zh>chartĻƷ.</zh>
   * <en>The chart orientation.</en>
   */
  private ChartOrientation orientation ;

  /**
   * <zh>domain᷽Ƿ.</zh>
   * <en>A flag that controls whether or not domain axis zooming is enabled.</en>
   */
  private boolean domainZoomable ;

  /**
   * <zh>value᷽Ƿ.</zh>
   * <en>A flag that controls whether or not value axis zooming is enabled.</en>
   */
  private boolean valueZoomable ;

  /**
   * <zh>ŵʼ.</zh>
   * <en>The zoom rectangle starting point.</en>
   */
  private Point2D zoomPoint ;

  /**
   * <zh>ŵľ.</zh>
   * <en>The zoom rectangle.</en>
   */
  private transient Rectangle2D zoomRectangle ;

  /**
   * <zh>ſı߿paint.</zh>
   * <en>The paint used to draw the zoom rectangle outline.</en>
   */
  private transient Paint zoomBorderPaint ;

  /**
   * <zh>ſpaint.</zh>
   * <en>The paint used to fill the zoom rectangle.</en>
   */
  private transient Paint zoomFillPaint ;

  /**
   * <zh>Ƿž.</zh>
   * <en>Controls if the zoom rectangle is drawn as a border or filled.</en>
   */
  private boolean fillZoomRectangle ;

  /**
   * <zh>ŵСϷž.</zh>
   * <en>The minimum distance required to drag the mouse to trigger a zoom.</en>
   */
  private double zoomTriggerDistance ;

  /**
   * <zh>Ƿˮƽ.</zh>
   * <en>A flag that controls whether or not horizontal tracing is enabled.</en>
   */
  private boolean horizontalAxisTrace ;

  /**
   * <zh>Ƿٴֱ.</zh>
   * <en>A flag that controls whether or not vertical tracing is enabled.</en>
   */
  private boolean verticalAxisTrace ;

  /**
   * <zh>ֱ.</zh>
   * <en>A vertical trace line.</en>
   */
  private transient Line2D verticalTraceLine ;

  /**
   * <zh>ˮƽ.</zh>
   * <en>A horizontal trace line.</en>
   */
  private transient Line2D horizontalTraceLine ;

  /**
   * <zh>Ƿtooltipӳٷı.</zh>
   * <en>A flag that indicates if original tooltip delays are changed.</en>
   */
  private boolean ownToolTipDelaysActive ;

  /**
   * <zh>ʼtooltipӳʱ(λΪ).</zh>
   * <en>Original initial tooltip delay of ToolTipManager.sharedInstance().</en>
   */
  private int originalToolTipInitialDelay ;

  /**
   * <zh>tooltipٴʾӳʱ(λΪ).</zh>
   * <en>Original reshow tooltip delay of ToolTipManager.sharedInstance().</en>
   */
  private int originalToolTipReshowDelay ;

  /**
   * <zh>tooltipʧӳʱ(λΪ).</zh>
   * <en>Original dismiss tooltip delay of ToolTipManager.sharedInstance().</en>
   */
  private int originalToolTipDismissDelay ;

  /**
   * <zh>ʾtooltipӳʱ.</zh>
   * <en>Own initial tooltip delay to be used in this chart panel.</en>
   */
  private int ownToolTipInitialDelay ;

  /**
   * <zh>ٴʾtooltipӳʱ.</zh>
   * <en>Own reshow tooltip delay to be used in this chart panel.</en>
   */
  private int ownToolTipReshowDelay ;

  /**
   * <zh>tooltipʧӳʱ.</zh>
   * <en>Own dismiss tooltip delay to be used in this chart panel.</en>
   */
  private int ownToolTipDismissDelay ;

  /**
   * <zh>Ŵ.</zh>
   * <en>The factor used to zoom in on an axis range.</en>
   */
  private double zoomInFactor ;

  /**
   * <zh>С.</zh>
   * <en>The factor used to zoom out on an axis range.</en>
   */
  private double zoomOutFactor ;

  /**
   * <zh>ǷʹԴΪê.</zh>
   * <en>A flag that controls whether zoom operations are centred on the
   * current anchor point, or the centre point of the relevant axis.</en>
   */
  private boolean useAnchor ;

  /**
   * <zh>chartͼζ.</zh>
   * <en>The collection of chart shapes.</en>
   */
  private ChartShapeCollection chartShapes ;

  /**
   * <zh>panelһδС.</zh>
   * <en>The old size of this panel.</en>
   */
  private Rectangle2D oldPanelSize ;

  /**
   * <zh>ͼʾ.</zh>
   * <en>Rendering hints that will be used for chart drawing.  This should never be null.</en>
   */
  private transient RenderingHints renderingHints ;

  /**
   * <zh>Ƿı¼.</zh>
   * <en>A flag controls whether or not mouse changed.</en>
   */
  private boolean mouseChanged ;

  /**
   * <zh>굥¼.</zh>
   * <en>The point of mouse click.</en>
   */
  private Point mouseClickPoint ;

  /**
   * <zh>ʾtooltipͼεͼ,ȡ{@link ChartShapeDrawType}.</zh>
   * <en>Controls which type of shape can show tooltip.</en>
   */
  private List showTooltipShapeType ;

  /**
   * <zh>Ƿ.</zh>
   * <en>Controls whether or not zoomable.</en>
   */
  private boolean zoomable ;

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param chart IJChart <zh>IJChart.</zh>
   * <en>The IJChart object.</en>
   */
  public ChartPanel(IJChart chart) {
    this(chart, true) ;
  }

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param chart IJChart <zh>IJChart.</zh>
   * <en>The IJChart object.</en>
   * @param showTooltips boolean <zh>Ƿʾtooltip.</zh>
   * <en>Whether or not show tooltips.</en>
   */
  public ChartPanel(IJChart chart, boolean showTooltips) {
    this(chart, DEFAULT_WIDTH, DEFAULT_HEIGHT, DEFAULT_MINIMUM_DRAW_WIDTH,
         DEFAULT_MINIMUM_DRAW_HEIGHT, showTooltips) ;
  }

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param chart IJChart <zh>IJChart.</zh>
   * <en>The IJChart object.</en>
   * @param width int <zh>.</zh>
   * <en>The width.</en>
   * @param height int <zh>߶.</zh>
   * <en>The height.</en>
   * @param minimumDrawWidth int <zh>Сƿ.</zh>
   * <en>The minimum drawing width.</en>
   * @param minimumDrawHeight int <zh>СƸ߶.</zh>
   * <en>The minimum drawing height.</en>
   * @param showTooltips boolean <zh>Ƿʾtooltip.</zh>
   * <en>Whether or not show tooltips.</en>
   */
  public ChartPanel(IJChart chart, int width, int height, int minimumDrawWidth,
                    int minimumDrawHeight, boolean showTooltips) {
    this.setChart(chart) ;
    this.chartMouseListeners = new EventListenerList() ;
    setPreferredSize(new Dimension(width, height)) ;
    this.refreshBuffer = false ;
    this.minimumDrawWidth = minimumDrawWidth ;
    this.minimumDrawHeight = minimumDrawHeight ;
    this.zoomTriggerDistance = DEFAULT_ZOOM_TRIGGER_DISTANCE ;

    this.zoomPoint = null ;
    this.zoomRectangle = null ;
    this.zoomBorderPaint = DEFAULT_ZOOM_BORDER_PAINT ;
    this.zoomFillPaint = DEFAULT_ZOOM_FILL_PAINT ;
    this.fillZoomRectangle = DEFAULT_FILL_ZOOM_RECTANGLE ;
    this.horizontalAxisTrace = DEFAULT_HORIZONTAL_AXIS_TRACE ;
    this.verticalAxisTrace = DEFAULT_VERTICAL_AXIS_TRACE ;
    this.zoomOutFactor = DEFAULT_ZOOM_OUT_FACTOR ;
    this.zoomInFactor = DEFAULT_ZOOM_IN_FACTOR ;
    this.zoomable = DEFAULT_ZOOMABLE ;

    enableEvents(AWTEvent.MOUSE_EVENT_MASK) ;
    enableEvents(AWTEvent.MOUSE_MOTION_EVENT_MASK) ;
    setDisplayToolTips(showTooltips) ;
    addMouseListener(this) ;
    addMouseMotionListener(this) ;

    ToolTipManager ttm = ToolTipManager.sharedInstance() ;
    this.ownToolTipInitialDelay = ttm.getInitialDelay() ;
    this.ownToolTipDismissDelay = ttm.getDismissDelay() ;
    this.ownToolTipReshowDelay = ttm.getReshowDelay() ;

    this.useAnchor = false ;
    this.chartShapes = null ;
    this.oldPanelSize = null ;
    this.mouseClickPoint = null ;
    this.mouseChanged = false ;

    this.renderingHints = new RenderingHints(RenderingHints.KEY_ANTIALIASING,
                                             RenderingHints.VALUE_ANTIALIAS_ON) ;
    this.renderingHints.put(RenderingHints.KEY_TEXT_ANTIALIASING,
                            RenderingHints.VALUE_TEXT_ANTIALIAS_ON) ;

    this.showTooltipShapeType = new ArrayList() ;
    this.showTooltipShapeType.add(ChartShapeDrawType.AXIS_LABEL_SHAPE) ;
    this.showTooltipShapeType.add(ChartShapeDrawType.AXIS_TICK_LABEL_SHAPE) ;
    this.showTooltipShapeType.add(ChartShapeDrawType.AXIS_TICK_MARK_SHAPE) ;
    this.showTooltipShapeType.add(ChartShapeDrawType.TITLE_SHAPE) ;
    this.showTooltipShapeType.add(ChartShapeDrawType.LEGEND_SHAPE) ;
    this.showTooltipShapeType.add(ChartShapeDrawType.RENDER_DATA_ITEM_SHAPE) ;
  }

  /**
   * <zh>chart.</zh>
   * <en>Returns the chart contained in the panel.</en>
   *
   * @return <zh>chart (Ϊnull).</zh>
   * <en>The chart (maybe null).</en>
   */
  public IJChart getChart() {
    return this.chart ;
  }

  /**
   * <zh>chartʾ.</zh>
   * <en>Sets the chart that is displayed in the panel.</en>
   *
   * @param chart IJChart <zh>chart (Ϊnull).</zh>
   * <en>The chart (null permitted).</en>
   */
  public void setChart(IJChart chart) {
    if (this.chart != null) {
      this.chart.removeChangeListener(this) ;
      this.chart.removeProgressListener(this) ;
    }

    // add the new chart
    this.chart = chart ;
    if (this.chart != null) {
      this.chart.addChangeListener(this) ;
      this.chart.addProgressListener(this) ;
      AbstractPlot plot = chart.getPlot() ;

      this.domainZoomable = plot.isDomainZoomable() ;
      this.valueZoomable = plot.isValueZoomable() ;
      this.orientation = plot.getOrientation() ;
    } else {
      this.domainZoomable = false ;
      this.valueZoomable = false ;
    }

    this.chartShapes = null ;
    this.refreshBuffer = true ;
    repaint() ;
  }

  /**
   * <zh>СĻƿ.</zh>
   * <en>Returns the minimum width for drawing a chart.</en>
   *
   * @return double <zh>.</zh>
   * <en>The minimum drawing width.</en>
   */
  public double getMinimumDrawWidth() {
    return this.minimumDrawWidth ;
  }

  /**
   * <zh>СĻƿ.</zh>
   * <en>Sets the minimum width for drawing a chart.</en>
   *
   * @param width double <zh>.</zh>
   * <en>The width.</en>
   */
  public void setMinimumDrawWidth(double width) {
    this.minimumDrawWidth = width ;
  }

  /**
   * <zh>СĻƸ߶.</zh>
   * <en>Returns the minimum height for drawing a chart.</en>
   *
   * @return double <zh>߶.</zh>
   * <en>The minimum drawing height.</en>
   */
  public double getMinimumDrawHeight() {
    return this.minimumDrawHeight ;
  }

  /**
   * <zh>СĻƸ߶.</zh>
   * <en>Sets the minimum height for drawing a chart.</en>
   *
   * @param height double <zh>߶.</zh>
   * <en>The height.</en>
   */
  public void setMinimumDrawHeight(double height) {
    this.minimumDrawHeight = height ;
  }

  /**
   * <zh>x.</zh>
   * <en>Returns the X scale factor for the chart.  This will be 1.0 if no
   * scaling has been used.</en>
   *
   * @return double <zh>.</zh>
   * <en>The scale factor.</en>
   */
  public double getScaleX() {
    return this.scaleX ;
  }

  /**
   * <zh>y.</zh>
   * <en>Returns the Y scale factor for the chart.  This will be 1.0 if no
   * scaling has been used.</en>
   *
   * @return double <zh>.</zh>
   * <en>The scale factor.</en>
   */
  public double getScaleY() {
    return this.scaleY ;
  }

  /**
   * <zh>Ƿͼ.</zh>
   * <en>A convenience method that switches on mouse-based zooming.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setMouseZoomable(boolean flag) {
    setMouseZoomable(flag, true) ;
  }

  /**
   * <zh>Ƿͼ.</zh>
   * <en>A convenience method that switches on mouse-based zooming.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   * @param fillRectangle boolean <zh>ֵ.</zh>
   * <en>true if zoom rectangle is filled,false if rectangle is shown as outline only.</en>
   */
  public void setMouseZoomable(boolean flag, boolean fillRectangle) {
    setDomainZoomable(flag) ;
    setValueZoomable(flag) ;
    setFillZoomRectangle(fillRectangle) ;
  }

  /**
   * <zh>domain᷽Ƿ.</zh>
   * <en>Returns the flag that controls whether or not domain axis zooming is enabled.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDomainZoomable() {
    return this.domainZoomable ;
  }

  /**
   * <zh>domain᷽Ƿ.</zh>
   * <en>Sets the flag that controls whether or not domain axis zooming is enabled.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setDomainZoomable(boolean flag) {
    if (flag) {
      this.domainZoomable = flag && this.chart.getPlot().isDomainZoomable() ;
    } else {
      this.domainZoomable = false ;
    }
  }

  /**
   * <zh>value᷽Ƿ.</zh>
   * <en>Returns the flag that controls whether or not value axis zooming is enabled.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isValueZoomable() {
    return this.valueZoomable ;
  }

  /**
   * <zh>value᷽Ƿ.</zh>
   * <en>Sets the flag that controls whether or not value axis zooming is enabled.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setValueZoomable(boolean flag) {
    if (flag) {
      this.valueZoomable = flag && (this.chart.getPlot().isValueZoomable()) ;
    } else {
      this.valueZoomable = false ;
    }
  }

  /**
   * <zh>ػſı߿paint.</zh>
   * <en>Returns the paint used to draw the zoom rectangle border.</en>
   *
   * @return Paint <zh>paintĶ (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getZoomBorderPaint() {
    return this.zoomBorderPaint ;
  }

  /**
   * <zh>ûſı߿paint.</zh>
   * <en>Sets the paint used to draw the zoom rectangle border.</en>
   *
   * @param paint Paint <zh>paintĶ (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setZoomBorderPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.zoomBorderPaint = paint ;
  }

  /**
   * <zh>ſpaint.</zh>
   * <en>Returns the paint used to fill the zoom rectangle.</en>
   *
   * @return Paint <zh>paintĶ (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getZoomFillPaint() {
    return this.zoomFillPaint ;
  }

  /**
   * <zh>ſpaint.</zh>
   * <en>Sets the paint used to fill the zoom rectangle border.</en>
   *
   * @param paint Paint <zh>paintĶ (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setZoomFillPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.zoomFillPaint = paint ;
  }

  /**
   * <zh>Ƿž.</zh>
   * <en>Returns whether or not the zoom rectangle is filled.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isFillZoomRectangle() {
    return this.fillZoomRectangle ;
  }

  /**
   * <zh>Ƿž.</zh>
   * <en>Sets the flag that controls whether or not the zoom rectangle is filled.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setFillZoomRectangle(boolean flag) {
    this.fillZoomRectangle = flag ;
  }

  /**
   * <zh>شŵСϷž.</zh>
   * <en>Returns the minimum distance required to drag the mouse to trigger a zoom.</en>
   *
   * @return double <zh>.</zh>
   * <en>The distance.</en>
   */
  public double getZoomTriggerDistance() {
    return this.zoomTriggerDistance ;
  }

  /**
   * <zh>ôŵСϷž.</zh>
   * <en>Sets the minimum distance required to drag the mouse to trigger a zoom.</en>
   *
   * @param distance double <zh>.</zh>
   * <en>The distance.</en>
   */
  public void setZoomTriggerDistance(double distance) {
    this.zoomTriggerDistance = distance ;
  }

  /**
   * <zh>Ƿˮƽ.</zh>
   * <en>Returns the flag that controls whether or not horizontal tracing is enabled.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean getHorizontalAxisTrace() {
    return this.horizontalAxisTrace ;
  }

  /**
   * <zh>Ƿˮƽ.</zh>
   * <en>Sets the flag that controls whether or not horizontal tracing is enabled.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setHorizontalAxisTrace(boolean flag) {
    this.horizontalAxisTrace = flag ;
  }

  /**
   * <zh>ˮƽ.</zh>
   * <en>Returns the horizontal trace line.</en>
   *
   * @return Line2D <zh>ֱ (Ϊnull).</zh>
   * <en>The line (maybe null).</en>
   */
  public Line2D getHorizontalTraceLine() {
    return this.horizontalTraceLine ;
  }

  /**
   * <zh>ˮƽ.</zh>
   * <en>Sets the horizontal trace line.</en>
   *
   * @param line Line2D <zh>ֱ (Ϊnull).</zh>
   * <en>The line (null permitted).</en>
   */
  public void setHorizontalTraceLine(Line2D line) {
    this.horizontalTraceLine = line ;
  }

  /**
   * <zh>Ƿٴֱ.</zh>
   * <en>Returns the flag that controls whether or not vertical tracing is enabled.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isVerticalAxisTrace() {
    return this.verticalAxisTrace ;
  }

  /**
   * <zh>Ƿٴֱ.</zh>
   * <en>Sets the flag that controls whether or not vertical tracing is enabled.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setVerticalAxisTrace(boolean flag) {
    this.verticalAxisTrace = flag ;
  }

  /**
   * <zh>شֱ.</zh>
   * <en>Returns the vertical trace line.</en>
   *
   * @return Line2D <zh>ֱ (Ϊnull).</zh>
   * <en>The line (maybe null).</en>
   */
  protected Line2D getVerticalTraceLine() {
    return this.verticalTraceLine ;
  }

  /**
   * <zh>ôֱ.</zh>
   * <en>Sets the vertical trace line.</en>
   *
   * @param line Line2D <zh>ֱ (Ϊnull).</zh>
   * <en>The line (null permitted).</en>
   */
  protected void setVerticalTraceLine(Line2D line) {
    this.verticalTraceLine = line ;
  }

  /**
   * <zh>Ƿʾtooltip.</zh>
   * <en>Sets whether or not show tooltips.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setDisplayToolTips(boolean flag) {
    if (flag) {
      ToolTipManager.sharedInstance().registerComponent(this) ;
    } else {
      ToolTipManager.sharedInstance().unregisterComponent(this) ;
    }
  }

  /**
   * <zh>tooltipַ.</zh>
   * <en>Returns the tooltip text.</en>
   *
   * @param event MouseEvent <zh>¼ (Ϊnull).</zh>
   * <en>The mouse event (null not permitted).</en>
   *
   * @return String <zh>tooltipַ (Ϊnull).</zh>
   * <en>A tool tip or null if no tooltip is available (maybe null).</en>
   */
  public String getToolTipText(MouseEvent event) {
    if (event == null) {
      throw new IllegalArgumentException("Null 'event' argument.") ;
    }

    ChartShapeCollection shapes = this.chart.getChartShapes() ;
    Insets insets = getInsets() ;
    double x = (event.getX() - insets.left) / this.scaleX ;
    double y = (event.getY() - insets.top) / this.scaleY ;

    int size = this.showTooltipShapeType.size() ;
    for (int i = 0 ; i < size ; i++) {
      ChartShapeDrawType drawType = (ChartShapeDrawType)this.showTooltipShapeType.get(i) ;
      IChartShape shape = shapes.findChartShape(new Point2D.Double(x, y), drawType) ;
      if (shape != null) {
        return shape.getTooltipText() ;
      }
    }

    return null ;
  }

  /**
   * <zh>תһΪĻ.</zh>
   * <en>Translates a Java2D point on the chart to a screen location..</en>
   *
   * @param java2DPoint Point2D <zh>һ (Ϊnull).</zh>
   * <en>The Java2D point (null not permitted).</en>
   *
   * @return Point2D <zh>Ļ (Ϊnull).</zh>
   * <en>The screen location (never null).</en>
   */
  public Point2D translateJava2DToScreen(Point2D java2DPoint) {
    if (java2DPoint == null) {
      throw new IllegalArgumentException("Null 'java2DPoint' argument.") ;
    }

    Insets insets = getInsets() ;
    double x = java2DPoint.getX() * this.scaleX + insets.left ;
    double y = java2DPoint.getY() * this.scaleY + insets.top ;

    return new Point2D.Double(x, y) ;
  }

  /**
   * <zh>תĻΪjava.</zh>
   * <en>Translates a panel (component) location to a Java2D point.</en>
   *
   * @param screenPoint Point2D <zh>Ļ (Ϊnull).</zh>
   * <en>The screen location (null not permitted).</en>
   *
   * @return <zh>java (Ϊnull).</zh>
   * <en>The Java2D coordinates (never null).</en>
   */
  public Point2D translateScreenToJava2D(Point2D screenPoint) {
    if (screenPoint == null) {
      throw new IllegalArgumentException("Null 'screenPoint' argument.") ;
    }

    Insets insets = getInsets() ;
    double x = (screenPoint.getX() - insets.left) / this.scaleX ;
    double y = (screenPoint.getY() - insets.top) / this.scaleY ;

    return new Point2D.Double(x, y) ;
  }

  /**
   * <zh>һ,صͼΪźľ.</zh>
   * <en>Applies any scaling that is in effect for the chart drawing to the given rectangle.</en>
   *
   * @param rect Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The rectangle (null not permitted).</en>
   *
   * @return Rectangle2D <zh>źľ (Ϊnull).</zh>
   * <en>A new scaled rectangle (never null).</en>
   */
  public Rectangle2D scale(Rectangle2D rect) {
    if (rect == null) {
      throw new IllegalArgumentException("Null 'rect' argument.") ;
    }

    Insets insets = getInsets() ;
    double x = rect.getX() * getScaleX() + insets.left ;
    double y = rect.getY() * this.getScaleY() + insets.top ;
    double w = rect.getWidth() * this.getScaleX() ;
    double h = rect.getHeight() * this.getScaleY() ;

    return new Rectangle2D.Double(x, y, w, h) ;
  }

  /**
   * <zh>ĳӦͼ.</zh>
   * <en>Returns the shape which contains the specified point.</en>
   *
   * @param point Point2D <zh> (Ϊnull).</zh>
   * <en>The point (null not permitted).</en>
   *
   * @return IChartShape <zh>ͼ (Ϊnull).</zh>
   * <en>The shape (never null).</en>
   */
  public IChartShape getShapeForPoint(Point2D point) {
    return getShapeForPoint(point, ChartShapeDrawType.ANY_SHAPE) ;
  }

  /**
   * <zh>ĳͼ.</zh>
   * <en>Returns the shape which contains the specified point.</en>
   *
   * @param point Point2D <zh> (Ϊnull).</zh>
   * <en>The point (null not permitted).</en>
   * @param drawType ChartShapeDrawType <zh>ͼεĻ (Ϊnull).</zh>
   * <en>The shape draw shape (null permitted).</en>
   *
   * @return IChartShape <zh>ͼ (Ϊnull).</zh>
   * <en>The shape (never null).</en>
   */
  public IChartShape getShapeForPoint(Point2D point, ChartShapeDrawType drawType) {
    if (point == null) {
      throw new IllegalArgumentException("Null 'point' argument.") ;
    }
    if (drawType == null) {
      throw new IllegalArgumentException("Null 'drawType' argument") ;
    }

    ChartShapeCollection shapes = this.chart.getChartShapes() ;
    Insets insets = getInsets() ;
    double x = (point.getX() - insets.left) / this.scaleX ;
    double y = (point.getY() - insets.top) / this.scaleY ;

    return shapes.findChartShape(new Point2D.Double(x, y), drawType) ;
  }

  /**
   * <zh>Ƿˢ».</zh>
   * <en>Returns the flag that indicates that the buffer should be refreshed.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isRefreshBuffer() {
    return this.refreshBuffer ;
  }

  /**
   * <zh>Ƿˢ».</zh>
   * <en>Sets the flag that indicates that the buffer should be refreshed.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setRefreshBuffer(boolean flag) {
    this.refreshBuffer = flag ;
  }

  /**
   * <zh>һʾtooltipͼ.</zh>
   * <en>Add a shape type to display tooltip.</en>
   *
   * @param drawType ChartShapeDrawType <zh>ͼ (Ϊnull).</zh>
   * <en>The shape type (null not permitted).</en>
   */
  public void addShowTooltipShapeType(ChartShapeDrawType drawType) {
    if (drawType == null) {
      throw new IllegalArgumentException("Null 'drawType' argument") ;
    }

    if (!this.showTooltipShapeType.contains(drawType)) {
      this.showTooltipShapeType.add(drawType) ;
    }
  }

  /**
   * <zh>ɾһʾtooltipͼ.</zh>
   * <en>Remove a shape type to display tooltip.</en>
   *
   * @param drawType ChartShapeDrawType <zh>ͼ (Ϊnull).</zh>
   * <en>The shape type (null not permitted).</en>
   */
  public void removeShowTooltipShapeType(ChartShapeDrawType drawType) {
    if (drawType == null) {
      throw new IllegalArgumentException("Null 'drawType' argument") ;
    }

    this.showTooltipShapeType.remove(drawType) ;
  }

  /**
   * <zh>ɾеʾtooltipͼ.</zh>
   * <en>Clear all shape types to display tooltip.</en>
   */
  public void clearShowTooltipShapeType() {
    this.showTooltipShapeType.clear() ;
  }

  /**
   * <zh>еʾtooltipͼ.</zh>
   * <en>Returns all shape types to display tooltip.</en>
   *
   * @return List <zh>޸ĵͼͼ (Ϊnull).</zh>
   * <en>The shape type collection and unmodifable (never null).</en>
   */
  public List getShowTooltipShapeTypeList() {
    return Collections.unmodifiableList(this.showTooltipShapeType) ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Returns controls whether or not zoomable.</en>
   *
   * @return boolean <zh>Ƿ.</zh>
   * <en>Controls whether or not zoomable.</en>
   */
  public boolean isZoomable() {
    return this.zoomable ;
  }

  /**
   * <zh>Ƿ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets controls whether or not zoomable and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setZoomable(boolean flag) {
    this.zoomable = flag ;
  }

  /**
   * <zh>ƶ.</zh>
   * <en>Drawing the chart.</en>
   *
   * @param g Graphics <zh>ͼĶ.</zh>
   * <en>The graphics device.</en>
   */
  public void paintComponent(Graphics g) {
    if (this.mouseChanged) {
      boolean isAddCrosshair = false ;
      if (this.chart != null) {
        isAddCrosshair = this.chart.drawCrosshair(this.mouseClickPoint) ;
      }

      if (!isAddCrosshair) { //ʮûз仯
        this.mouseChanged = false ;
        return ;
      }
    }

    super.paintComponent(g) ;
    if (this.chart == null) {
      this.mouseChanged = false ;
      return ;
    }

    Graphics2D g2 = (Graphics2D) g ;
    g2.addRenderingHints(this.renderingHints) ;

    Rectangle2D visible = getVisibleRect() ;
    if (this.oldPanelSize == null || !visible.equals(this.oldPanelSize)) {
      this.chartShapes = null ;
      this.oldPanelSize = visible ;
    }
    Insets insets = getInsets() ;
    Rectangle2D available = new Rectangle2D.Double(insets.left, insets.top,
                                                   visible.getWidth() - insets.left - insets.right,
                                                   visible.getHeight() - insets.top - insets.bottom) ;

    boolean scale = false ;
    double drawWidth = available.getWidth() ;
    double drawHeight = available.getHeight() ;
    this.scaleX = 1.0 ;
    this.scaleY = 1.0 ;

    if (drawWidth < this.minimumDrawWidth) {
      this.scaleX = drawWidth / this.minimumDrawWidth ;
      drawWidth = this.minimumDrawWidth ;
      scale = true ;
    }

    if (drawHeight < this.minimumDrawHeight) {
      this.scaleY = drawHeight / this.minimumDrawHeight ;
      drawHeight = this.minimumDrawHeight ;
      scale = true ;
    }

    boolean clearBuffer = true ;
    if (this.chartBuffer == null || this.chartBufferWidth != drawWidth
        || this.chartBufferHeight != drawHeight) {
      this.chartBufferWidth = (int) drawWidth ;
      this.chartBufferHeight = (int) drawHeight ;

      this.chartBuffer = createImage(this.chartBufferWidth, this.chartBufferHeight) ;
      clearBuffer = false ;
      this.refreshBuffer = true ;
    }

    Rectangle2D chartArea = new Rectangle2D.Double(0.0, 0.0, drawWidth, drawHeight) ;
    if (this.refreshBuffer || (this.chartShapes != null && this.chartShapes.isShapeChanged())) {
      Graphics2D bufferG2 = (Graphics2D)this.chartBuffer.getGraphics() ;
      bufferG2.setRenderingHints(this.renderingHints) ;
      if (clearBuffer) {
        bufferG2.clearRect(0, 0, this.chartBufferWidth, this.chartBufferHeight) ;
      }
      if (scale) {
        bufferG2.transform(AffineTransform.getScaleInstance(this.scaleX, this.scaleY)) ;
        if (this.chartShapes != null) {
          this.chartShapes.draw(bufferG2) ;
        } else {
          if (!this.mouseChanged) {
            this.chart.draw(bufferG2, chartArea) ;
          }
          this.chartShapes = this.chart.getChartShapes() ;
          this.chartShapes.removeAllChangeListener() ;
          this.chartShapes.addChangeListener(this) ;
          this.chartShapes.draw(bufferG2) ;
        }
      } else {
        if (this.chartShapes != null) {
          this.chartShapes.draw(bufferG2) ;
        } else {
          if (!this.mouseChanged) {
            this.chart.draw(bufferG2, chartArea) ;
          }
          this.chartShapes = this.chart.getChartShapes() ;
          this.chartShapes.removeAllChangeListener() ;
          this.chartShapes.addChangeListener(this) ;
          this.chartShapes.draw(bufferG2) ;
        }
      }

      this.refreshBuffer = false ;
    }

    if (this.chartShapes != null) {
      this.chartShapes.setShapeChanged(false) ;
    }
    g2.drawImage(this.chartBuffer, insets.left, insets.top, this) ;
    drawZoomRectangle(g2) ;

    this.verticalTraceLine = null ;
    this.horizontalTraceLine = null ;
    this.mouseChanged = false ;
  }

  /**
   * <zh>chart仯¼.</zh>
   * <en>Receives notification of changes to the chart, and redraws the chart.</en>
   *
   * @param event ChartChangeEvent <zh>¼ (Ϊnull).</zh>
   * <en>The event (null permitted).</en>
   */
  public void chartChanged(ChartChangeEvent event) {
    this.refreshBuffer = true ;
    this.chartShapes = null ;
    repaint() ;
  }

  /**
   * <zh>յͼζ仯֪ͨ.</zh>
   * <en>Receives notification of an {@link com.ijchart.xychart.shape.IChartShape} change event.</en>
   *
   * @param event ChartShapeChangeEvent <zh>ͼζ仯Ķ</zh>
   * <en>Information about the event.</en>
   */
  public void chartShapeChanged(ChartShapeChangeEvent event) {
    this.refreshBuffer = true ;
    repaint() ;
  }

  /**
   * <zh>chartƹе¼.</zh>
   * <en>Receives notification of a chart progress event.</en>
   *
   * @param event ChartProgressEvent <zh>¼.</zh>
   * <en>The event.</en>
   */
  public void chartProgress(ChartProgressEvent event) {
    // does nothing - override if necessary
  }

  /**
   * <zh>¼.</zh>
   * <en>Invoked when the mouse enters a component.</en>
   *
   * @param event MouseEvent <zh>¼.</zh>
   * <en>The mouse event.</en>
   */
  public void mouseEntered(MouseEvent event) {
    if (!this.ownToolTipDelaysActive) {
      ToolTipManager ttm = ToolTipManager.sharedInstance() ;

      this.originalToolTipInitialDelay = ttm.getInitialDelay() ;
      ttm.setInitialDelay(this.ownToolTipInitialDelay) ;

      this.originalToolTipReshowDelay = ttm.getReshowDelay() ;
      ttm.setReshowDelay(this.ownToolTipReshowDelay) ;

      this.originalToolTipDismissDelay = ttm.getDismissDelay() ;
      ttm.setDismissDelay(this.ownToolTipDismissDelay) ;

      this.ownToolTipDelaysActive = true ;
    }

    Insets insets = getInsets() ;
    double x = (event.getX() - insets.left) / this.scaleX ;
    double y = (event.getY() - insets.top) / this.scaleY ;

    Object[] listeners = this.chartMouseListeners.getListeners(IChartMouseListener.class) ;
    if (listeners.length == 0) {
      return ;
    }

    ChartShapeCollection shapes = this.chart.getChartShapes() ;
    IChartShape shape = shapes.findChartShape(new Point2D.Double(x, y),
                                              ChartShapeDrawType.ANY_SHAPE) ;

    ChartMouseEvent chartEvent = new ChartMouseEvent(getChart(), event, shape) ;
    for (int i = listeners.length - 1 ; i >= 0 ; i--) {
      ((IChartMouseListener) listeners[i]).chartMouseEntered(chartEvent) ;
    }
  }

  /**
   * <zh>˳¼.</zh>
   * <en>Invoked when the mouse exits a component.</en>
   *
   * @param event MouseEvent <zh>¼.</zh>
   * <en>The mouse event.</en>
   */
  public void mouseExited(MouseEvent event) {
    if (this.ownToolTipDelaysActive) {
      // restore original tooltip dealys
      ToolTipManager ttm = ToolTipManager.sharedInstance() ;
      ttm.setInitialDelay(this.originalToolTipInitialDelay) ;
      ttm.setReshowDelay(this.originalToolTipReshowDelay) ;
      ttm.setDismissDelay(this.originalToolTipDismissDelay) ;
      this.ownToolTipDelaysActive = false ;
    }

    Insets insets = getInsets() ;
    double x = (event.getX() - insets.left) / this.scaleX ;
    double y = (event.getY() - insets.top) / this.scaleY ;

    Object[] listeners = this.chartMouseListeners.getListeners(IChartMouseListener.class) ;
    if (listeners.length == 0) {
      return ;
    }

    ChartShapeCollection shapes = this.chart.getChartShapes() ;
    IChartShape shape = shapes.findChartShape(new Point2D.Double(x, y),
                                              ChartShapeDrawType.ANY_SHAPE) ;

    ChartMouseEvent chartEvent = new ChartMouseEvent(getChart(), event, shape) ;
    for (int i = listeners.length - 1 ; i >= 0 ; i--) {
      ((IChartMouseListener) listeners[i]).chartMouseExited(chartEvent) ;
    }
  }

  /**
   * <zh>갴¼.</zh>
   * <en>Invoked when a mouse button has been pressed on a component.</en>
   *
   * @param event MouseEvent <zh>¼.</zh>
   * <en>The mouse event.</en>
   */
  public void mousePressed(MouseEvent event) {
    if (this.zoomRectangle == null) {
      Rectangle2D screenDataArea = getScreenDataArea(event.getX(), event.getY()) ;
      if (screenDataArea != null) {
        this.zoomPoint = getPointInRectangle(new Point2D.Double(event.getX(), event.getY()),
                                             screenDataArea) ;
      } else {
        this.zoomPoint = null ;
      }
    }

    Insets insets = getInsets() ;
    double x = (event.getX() - insets.left) / this.scaleX ;
    double y = (event.getY() - insets.top) / this.scaleY ;

    Object[] listeners = this.chartMouseListeners.getListeners(IChartMouseListener.class) ;
    if (listeners.length == 0) {
      return ;
    }

    ChartShapeCollection shapes = this.chart.getChartShapes() ;
    IChartShape shape = shapes.findChartShape(new Point2D.Double(x, y),
                                              ChartShapeDrawType.ANY_SHAPE) ;

    ChartMouseEvent chartEvent = new ChartMouseEvent(getChart(), event, shape) ;
    for (int i = listeners.length - 1 ; i >= 0 ; i--) {
      ((IChartMouseListener) listeners[i]).chartMousePressed(chartEvent) ;
    }
  }

  /**
   * <zh>ǿpointarea.</zh>
   * <en>Returns a point based on (x, y) but constrained to be within the bounds
   * of the given rectangle.</en>
   *
   * @param point Poing2D <zh>һ.</zh>
   * <en>The point.</en>
   * @param area Rectangle2D <zh>.</zh>
   * <en>The rectangle area.</en>
   *
   * @return Point2D <zh>ת.</zh>
   * <en>A point within the rectangle.</en>
   */
  private static Point2D getPointInRectangle(Point2D point, Rectangle2D area) {
    double x = Math.max(Math.ceil(area.getMinX()), Math.min(point.getX(), Math.floor(area.getMaxX()))) ;
    double y = Math.max(Math.ceil(area.getMinY()), Math.min(point.getY(), Math.floor(area.getMaxY()))) ;

    return new Point2D.Double(x, y) ;
  }

  /**
   * <zh>Ϸ¼.</zh>
   * <en>Invoked when a mouse button is pressed on a component and then dragged.</en>
   *
   * @param event MouseEvent <zh>¼.</zh>
   * <en>The mouse event.</en>
   */
  public void mouseDragged(MouseEvent event) {
    // if no initial zoom point was set, ignore dragging...
    if (this.zoomPoint == null) {
      return ;
    }

    boolean hZoom = false ;
    boolean vZoom = false ;
    if (this.orientation != null && this.orientation.equals(ChartOrientation.HORIZONTAL)) {
      hZoom = this.valueZoomable ;
      vZoom = this.domainZoomable ;
    } else if (this.orientation != null && this.orientation.equals(ChartOrientation.VERTICAL)) {
      hZoom = this.domainZoomable ;
      vZoom = this.valueZoomable ;
    }

    Rectangle2D scaledDataArea = getScreenDataArea((int)this.zoomPoint.getX(),
                                                   (int)this.zoomPoint.getY()) ;
    if (hZoom && vZoom) {
      // selected rectangle shouldn't extend outside the data area...
      double xmax = Math.min(event.getX(), scaledDataArea.getMaxX()) ;
      double ymax = Math.min(event.getY(), scaledDataArea.getMaxY()) ;
      this.zoomRectangle = new Rectangle2D.Double(this.zoomPoint.getX(), this.zoomPoint.getY(),
                                                  xmax - this.zoomPoint.getX(),
                                                  ymax - this.zoomPoint.getY()) ;
    } else if (hZoom) {
      double xmax = Math.min(event.getX(), scaledDataArea.getMaxX()) ;
      this.zoomRectangle = new Rectangle2D.Double(this.zoomPoint.getX(), scaledDataArea.getMinY(),
                                                  xmax - this.zoomPoint.getX(),
                                                  scaledDataArea.getHeight()) ;
    } else if (vZoom) {
      double ymax = Math.min(event.getY(), scaledDataArea.getMaxY()) ;
      this.zoomRectangle = new Rectangle2D.Double(scaledDataArea.getMinX(), this.zoomPoint.getY(),
                                                  scaledDataArea.getWidth(),
                                                  ymax - this.zoomPoint.getY()) ;
    }

    if (this.zoomable && this.zoomRectangle != null) {
      repaint() ;
    }
  }

  /**
   * <zh>ͷ¼.</zh>
   * <en>Invoked when a mouse button has been released on a component.</en>
   *
   * @param event MouseEvent <zh>¼.</zh>
   * <en>The mouse event.</en>
   */
  public void mouseReleased(MouseEvent event) {
    if (this.zoomable && this.zoomRectangle != null) {
      boolean hZoom = false ;
      boolean vZoom = false ;
      if (this.orientation.equals(ChartOrientation.HORIZONTAL)) {
        hZoom = this.valueZoomable ;
        vZoom = this.domainZoomable ;
      } else {
        hZoom = this.domainZoomable ;
        vZoom = this.valueZoomable ;
      }

      boolean zoomTrigger1 = hZoom && Math.abs(event.getX() - this.zoomPoint.getX())
                             >= this.zoomTriggerDistance ;
      boolean zoomTrigger2 = vZoom && Math.abs(event.getY() - this.zoomPoint.getY())
                             >= this.zoomTriggerDistance ;
      if (zoomTrigger1 || zoomTrigger2) {
        if ((hZoom && event.getX() < this.zoomPoint.getX())
            || (vZoom && (event.getY() < this.zoomPoint.getY()))) {
          restoreAutoBounds() ;
        } else {
          double x = 0, y = 0, w = 0, h = 0 ;
          Rectangle2D screenDataArea = getScreenDataArea((int)this.zoomPoint.getX(),
                                                         (int)this.zoomPoint.getY()) ;
          if (!vZoom) {
            x = this.zoomPoint.getX() ;
            y = screenDataArea.getMinY() ;
            w = Math.min(this.zoomRectangle.getWidth(),
                         screenDataArea.getMaxX() - this.zoomPoint.getX()) ;
            h = screenDataArea.getHeight() ;
          } else if (!hZoom) {
            x = screenDataArea.getMinX() ;
            y = this.zoomPoint.getY() ;
            w = screenDataArea.getWidth() ;
            h = Math.min(this.zoomRectangle.getHeight(),
                         screenDataArea.getMaxY() - this.zoomPoint.getY()) ;
          } else {
            x = this.zoomPoint.getX() ;
            y = this.zoomPoint.getY() ;
            w = Math.min(this.zoomRectangle.getWidth(),
                         screenDataArea.getMaxX() - this.zoomPoint.getX()) ;
            h = Math.min(this.zoomRectangle.getHeight(),
                         screenDataArea.getMaxY() - this.zoomPoint.getY()) ;
          }

          if (w > 0 && h > 0) {
            Rectangle2D zoomArea = new Rectangle2D.Double(x, y, w, h) ;
            zoom(zoomArea) ;
          }
        }
      }

      this.zoomRectangle = null ;
      repaint() ;
    }

    Insets insets = getInsets() ;
    double x = (event.getX() - insets.left) / this.scaleX ;
    double y = (event.getY() - insets.top) / this.scaleY ;

    Object[] listeners = this.chartMouseListeners.getListeners(IChartMouseListener.class) ;
    if (listeners.length == 0) {
      return ;
    }

    ChartShapeCollection shapes = this.chart.getChartShapes() ;
    IChartShape shape = shapes.findChartShape(new Point2D.Double(x, y),
                                              ChartShapeDrawType.ANY_SHAPE) ;

    ChartMouseEvent chartEvent = new ChartMouseEvent(getChart(), event, shape) ;
    for (int i = listeners.length - 1 ; i >= 0 ; i--) {
      ((IChartMouseListener) listeners[i]).chartMouseReleased(chartEvent) ;
    }
  }

  /**
   * <zh>¼.</zh>
   * <en>Invoked when the mouse button has been clicked (pressed and released) on a component.</en>
   *
   * @param event MouseEvent <zh>¼.</zh>
   * <en>The mouse event.</en>
   */
  public void mouseClicked(MouseEvent event) {
    this.mouseChanged = true ;
    this.mouseClickPoint = event.getPoint() ;
    if (this.chart == null) {
      return ;
    }

    Insets insets = getInsets() ;
    double x = (event.getX() - insets.left) / this.scaleX ;
    double y = (event.getY() - insets.top) / this.scaleY ;

    AbstractPlot plot = this.chart.getPlot() ;
    if (plot instanceof AbstractAxisPlot) {
      AbstractAxisPlot axisPlot = (AbstractAxisPlot) plot ;
      if (axisPlot.isDomainCrosshairVisible() || axisPlot.isValueCrosshairVisible()) {
        this.chart.setNotify(true) ; // force a redraw
      }
    }
    Object[] listeners = this.chartMouseListeners.getListeners(IChartMouseListener.class) ;
    if (listeners.length == 0) {
      return ;
    }

    ChartShapeCollection shapes = this.chart.getChartShapes() ;
    IChartShape shape = shapes.findChartShape(new Point2D.Double(x, y),
                                              ChartShapeDrawType.ANY_SHAPE) ;

    ChartMouseEvent chartEvent = new ChartMouseEvent(getChart(), event, shape) ;
    for (int i = listeners.length - 1 ; i >= 0 ; i--) {
      ((IChartMouseListener) listeners[i]).chartMouseClicked(chartEvent) ;
    }
  }

  /**
   * <zh>ƶ¼.</zh>
   * <en>Invoked when the mouse cursor has been moved onto a component
   *  but no buttons have been pushed.</en>
   *
   * @param event MouseEvent <zh>¼.</zh>
   * <en>The mouse event.</en>
   */
  public void mouseMoved(MouseEvent event) {
    if (this.chart == null) {
      return ;
    }

    Graphics2D g2 = (Graphics2D) getGraphics() ;
    if (this.horizontalAxisTrace) {
      drawHorizontalAxisTrace(g2, event.getX()) ;
    }
    if (this.verticalAxisTrace) {
      drawVerticalAxisTrace(g2, event.getY()) ;
    }
    g2.dispose() ;

    Object[] listeners = this.chartMouseListeners.getListeners(IChartMouseListener.class) ;
    if (listeners.length == 0) {
      return ;
    }

    Insets insets = getInsets() ;
    int x = (int) ((event.getX() - insets.left) / this.scaleX) ;
    int y = (int) ((event.getY() - insets.top) / this.scaleY) ;

    ChartShapeCollection shapes = this.chart.getChartShapes() ;
    IChartShape shape = shapes.findChartShape(new Point2D.Double(x, y),
                                              ChartShapeDrawType.ANY_SHAPE) ;

    ChartMouseEvent chartEvent = new ChartMouseEvent(getChart(), event, shape) ;
    for (int i = listeners.length - 1 ; i >= 0 ; i--) {
      ((IChartMouseListener) listeners[i]).chartMouseMoved(chartEvent) ;
    }
  }

  /**
   * <zh>ѡ.</zh>
   * <en>Zooms in on a selected region.</en>
   *
   * @param selection Rectangle2D <zh>.</zh>
   * <en>The selected region.</en>
   */
  public void zoom(Rectangle2D selection) {
    if (selection == null) {
      throw new IllegalArgumentException("Null 'selection' argument.") ;
    }
    if (selection.getWidth() == 0 || selection.getHeight() == 0) {
      throw new IllegalArgumentException("The width or height of 'selection' must be >0.") ;
    }
    if (this.chart == null || !this.zoomable || this.chart.getPlot() instanceof NoAxisPlot
        || this.chart.getPlot().getOrientation() == null) {
      return ;
    }

    Rectangle2D scaledDataArea = getScreenDataArea((int) selection.getCenterX(),
                                                   (int) selection.getCenterY()) ;
    if (selection.getHeight() > 0 && selection.getWidth() > 0) {
      Point2D center = new Point2D.Double(selection.getCenterX(), selection.getCenterY()) ;
      AbstractPlot plot = this.chart.getPlot() ;
      plot.zoomDomainAxes(selection, scaledDataArea, center) ;
      plot.zoomValueAxes(selection, scaledDataArea, center) ;
    }
  }

  /**
   * <zh>ڷ(x)ֵ(y)ϾָԶΧ.</zh>
   * <en>Restores the auto-range calculation on both axes.</en>
   */
  public void restoreAutoBounds() {
    restoreAutoDomainBounds() ;
    restoreAutoValueBounds() ;
  }

  /**
   * <zh>ڷ(x)ϻָԶΧ.</zh>
   * <en>Restores the auto-range calculation on the category axis.</en>
   */
  public void restoreAutoDomainBounds() {
    if (this.chart == null) {
      return ;
    }

    this.chart.getPlot().zoomDomainAxes(null, null, this.zoomPoint) ;
  }

  /**
   * <zh>ֵ(y)ϻָԶΧ.</zh>
   * <en>Restores the auto-range calculation on the value axis.</en>
   */
  public void restoreAutoValueBounds() {
    if (this.chart == null) {
      return ;
    }

    this.chart.getPlot().zoomValueAxes(null, null, this.zoomPoint) ;
  }

  /**
   * <zh>Ļϵ(ǵ).</zh>
   * <en>Returns the data area for the chart with the current scaling applied.</en>
   *
   * @return Rectangle2D <zh>ź (Ϊnull).</zh>
   * <en>The scaled data area (maybe null).</en>
   */
  public Rectangle2D getScreenDataArea() {
    if (this.chart == null) {
      return null ;
    }

    Rectangle2D dataArea = this.chart.getPlot().getDataArea() ;
    Insets insets = getInsets() ;
    double x = dataArea.getX() * this.scaleX + insets.left ;
    double y = dataArea.getY() * this.scaleY + insets.top ;
    double w = dataArea.getWidth() * this.scaleX ;
    double h = dataArea.getHeight() * this.scaleY ;

    return new Rectangle2D.Double(x, y, w, h) ;
  }

  /**
   * <zh>Ļϵ(ǵ).</zh>
   * <en>Returns the data area for the chart with the current scaling applied.</en>
   *
   * @param x int <zh>x.</zh>
   * <en>The x-coordinate.</en>
   * @param y int <zh>y.</zh>
   * <en>The y-coordinate.</en>
   *
   * @return Rectangle2D <zh>ź (Ϊnull).</zh>
   * <en>The scaled data area (maybe null).</en>
   */
  public Rectangle2D getScreenDataArea(int x, int y) {
    if (this.chart == null) {
      return null ;
    }

    AbstractPlot plot = this.chart.getPlot() ;
    Collection subPlots = plot.getSubPlots() ;
    if (subPlots == null || subPlots.size() == 0) {
      return getScreenDataArea() ;
    } else {
      Iterator iter = subPlots.iterator() ;
      while (iter.hasNext()) {
        AbstractPlot subPlot = (AbstractPlot) iter.next() ;
        Rectangle2D dataArea = scale(subPlot.getDataArea()) ;
        if (dataArea.contains(x, y)) {
          return dataArea ;
        }
      }
    }

    return null ;
  }

  /**
   * <zh>ʾtooltipӳʱ.</zh>
   * <en>Returns the initial tooltip delay to be used in this chart panel.</en>
   *
   * @return int <zh>ӳʱ.</zh>
   * <en>An integer representing the initial delay value,in milliseconds</en>
   */
  public int getInitialDelay() {
    return this.ownToolTipInitialDelay ;
  }

  /**
   * <zh>ʾtooltipӳʱ.</zh>
   * <en>Sets the initial tooltip delay to be used in this chart panel.</en>
   *
   * @param delay int <zh>ӳʱ.</zh>
   * <en>An integer representing the initial delay value,in milliseconds</en>
   */
  public void setInitialDelay(int delay) {
    this.ownToolTipInitialDelay = delay ;
  }

  /**
   * <zh>ٴʾtooltipӳʱ.</zh>
   * <en>Returns the reshow tooltip delay to be used in this chart panel.</en>
   *
   * @return int <zh>ӳʱ.</zh>
   * <en>An integer representing the reshow delay value,in milliseconds.</en>
   */
  public int getReshowDelay() {
    return this.ownToolTipReshowDelay ;
  }

  /**
   * <zh>ٴʾtooltipӳʱ.</zh>
   * <en>Sets the reshow tooltip delay to be used in this chart panel.</en>
   *
   * @param delay int <zh>ӳʱ.</zh>
   * <en>An integer representing the reshow delay value,in milliseconds.</en>
   */
  public void setReshowDelay(int delay) {
    this.ownToolTipReshowDelay = delay ;
  }

  /**
   * <zh>tooltipʧӳʱ.</zh>
   * <en>Returns the dismiss tooltip delay to be used in this chart panel.</en>
   *
   * @return int <zh>tooltipʧʱ.</zh>
   * <en>An integer representing the dismiss delay value,in milliseconds.</en>
   */
  public int getDismissDelay() {
    return this.ownToolTipDismissDelay ;
  }

  /**
   * <zh>tooltipʧӳʱ.</zh>
   * <en>Sets the dismiss tooltip delay to be used in this chart panel.</en>
   *
   * @param delay int <zh>tooltipʧʱ.</zh>
   * <en>An integer representing the dismiss delay value,in milliseconds.</en>
   */
  public void setDismissDelay(int delay) {
    this.ownToolTipDismissDelay = delay ;
  }

  /**
   * <zh>طŴ.</zh>
   * <en>Returns the zoom in factor.</en>
   *
   * @return double <zh>Ŵ.</zh>
   * <en>The zoom in factor.</en>
   */
  public double getZoomInFactor() {
    return this.zoomInFactor ;
  }

  /**
   * <zh>طŴ.</zh>
   * <en>Returns the zoom in factor.</en>
   *
   * @param factor double <zh>Ŵ.</zh>
   * <en>The zoom in factor.</en>
   */
  public void setZoomInFactor(double factor) {
    this.zoomInFactor = factor ;
  }

  /**
   * <zh>С.</zh>
   * <en>Returns the zoom out factor.</en>
   *
   * @return double <zh>С.</zh>
   * <en>The zoom out factor.</en>
   */
  public double getZoomOutFactor() {
    return this.zoomOutFactor ;
  }

  /**
   * <zh>С.</zh>
   * <en>Sets the zoom out factor.</en>
   *
   * @param factor double <zh>С.</zh>
   * <en>The zoom out factor.</en>
   */
  public void setZoomOutFactor(double factor) {
    this.zoomOutFactor = factor ;
  }

  /**
   * <zh>ʱ (Ϊnull).</zh>
   * <en>Returns the mouse click point (null permitted).</en>
   *
   * @return Point <zh> (Ϊnull).</zh>
   * <en>The mouse click point (maybe null).</en>
   */
  public Point getMouseClickPoint() {
    return this.mouseClickPoint ;
  }

  /**
   * <zh>ػͼʾ.</zh>
   * <en>Returns the collection of rendering hints for the chart.</en>
   *
   * @return RenderingHints <zh>ͼʾ.</zh>
   * <en>The rendering hints for the chart (never null).</en>
   */
  public RenderingHints getRenderingHints() {
    return this.renderingHints ;
  }

  /**
   * <zh>ûͼʾ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the rendering hints for the chart,and
   * sends a {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param renderingHints RenderingHints <zh>ͼʾ (Ϊnull).</zh>
   * <en>The rendering hints (null not permitted).</en>
   */
  public void setRenderingHints(RenderingHints renderingHints) {
    if (renderingHints == null) {
      throw new IllegalArgumentException("Null 'renderingHints' argument.") ;
    }

    this.renderingHints = renderingHints ;
  }

  /**
   * <zh>صǰͼλǷ񷴾.</zh>
   * <en>Returns a flag that indicates whether or not anti-aliasing is used when
   * the chart is drawn.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>The flag.</en>
   */
  public boolean getAntiAlias() {
    Object val = this.renderingHints.get(RenderingHints.KEY_ANTIALIASING) ;
    return RenderingHints.VALUE_ANTIALIAS_ON.equals(val) ;
  }

  /**
   * <zh>õǰͼλǷ񷴾,ֵΪtrue,ͼ,ǻٶȻ.</zh>
   * <en>Sets a flag that indicates whether or not anti-aliasing is used when the
   * chart is drawn.
   * <P>
   * Anti-aliasing usually improves the appearance of charts, but is slower.</en>
   *
   * @param flag <zh>ֵ.</zh>
   * <en>The new value of the flag.</en>
   */
  public void setAntiAlias(boolean flag) {
    Object val = this.renderingHints.get(RenderingHints.KEY_ANTIALIASING) ;
    if (val == null) {
      val = RenderingHints.VALUE_ANTIALIAS_DEFAULT ;
    }
    if ((!flag && RenderingHints.VALUE_ANTIALIAS_OFF.equals(val))
        || (flag && RenderingHints.VALUE_ANTIALIAS_ON.equals(val))) {
      // no change, do nothing
      return ;
    }

    if (flag) {
      this.renderingHints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON) ;
    } else {
      this.renderingHints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF) ;
    }
  }

  /**
   * <zh>صǰֻǷ񷴾.</zh>
   * <en>Returns the current value stored in the rendering hints table for
   * {@link RenderingHints#KEY_TEXT_ANTIALIASING}.</en>
   *
   * @return Object <zh>Ƿ񷴾ֵ.</zh>
   * <en> The hint value (possibly null).</en>
   */
  public Object getTextAntiAlias() {
    return this.renderingHints.get(RenderingHints.KEY_TEXT_ANTIALIASING) ;
  }

  /**
   * <zh>ûǷ񷴾,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the value in the rendering hints table for
   * {@link RenderingHints#KEY_TEXT_ANTIALIASING} to either
   * {@link RenderingHints#VALUE_TEXT_ANTIALIAS_ON} or
   * {@link RenderingHints#VALUE_TEXT_ANTIALIAS_OFF}, then sends a
   * {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>The new value of the flag.</en>
   */
  public void setTextAntiAlias(boolean flag) {
    if (flag) {
      setTextAntiAlias(RenderingHints.VALUE_TEXT_ANTIALIAS_ON) ;
    } else {
      setTextAntiAlias(RenderingHints.VALUE_TEXT_ANTIALIAS_OFF) ;
    }
  }

  /**
   * <zh>ûǷ񷴾,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the value in the rendering hints table for
   * {@link RenderingHints#KEY_TEXT_ANTIALIASING} to either
   * {@link RenderingHints#VALUE_TEXT_ANTIALIAS_ON} or
   * {@link RenderingHints#VALUE_TEXT_ANTIALIAS_OFF}, then sends a
   * {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param val Object <zh>ַݵֵ,ֵܵΪ{@link RenderingHints#KEY_TEXT_ANTIALIASING}
   * {@link RenderingHints#VALUE_TEXT_ANTIALIAS_ON}
   * {@link RenderingHints#VALUE_TEXT_ANTIALIAS_OFF} (Ϊnull).</zh>
   * <en>The new value (null permitted),the value maybe
   * {@link RenderingHints#KEY_TEXT_ANTIALIASING} or
   * {@link RenderingHints#VALUE_TEXT_ANTIALIAS_ON} or
   * {@link RenderingHints#VALUE_TEXT_ANTIALIAS_OFF} (null not permitted).</en>
   */
  public void setTextAntiAlias(Object val) {
    this.renderingHints.put(RenderingHints.KEY_TEXT_ANTIALIASING, val) ;
  }

  /**
   * <zh>Ŵ.</zh>
   * <en>Draws the zoom rectangle.</en>
   *
   * @param g2 Graphics2D
   */
  private void drawZoomRectangle(Graphics2D g2) {
    if (this.chart == null || !this.zoomable || this.zoomRectangle == null
        || this.chart.getPlot() instanceof NoAxisPlot) {
      return ;
    }

    if (this.fillZoomRectangle) {
      g2.setPaint(this.zoomFillPaint) ;
      g2.fill(this.zoomRectangle) ;
    } else {
      g2.setPaint(this.zoomBorderPaint) ;
      g2.draw(this.zoomRectangle) ;
    }
  }

  /**
   * <zh>ˮƽĸ.</zh>
   * <en>Draws a vertical line used to trace the mouse position to the horizontal axis..</en>
   *
   * @param g2 Graphics2D
   * @param x double <zh>ߵx.</zh>
   * <en>The x-coordinate of the trace line.</en>
   */
  private void drawHorizontalAxisTrace(Graphics2D g2, double x) {
    Rectangle2D dataArea = getScreenDataArea() ;
    g2.setXORMode(Color.orange) ;
    if (((int) dataArea.getMinX() < x) && (x < (int) dataArea.getMaxX())) {

      if (this.verticalTraceLine != null) {
        g2.draw(this.verticalTraceLine) ;
        this.verticalTraceLine.setLine(x, dataArea.getMinY(), x, dataArea.getMaxY()) ;
      } else {
        this.verticalTraceLine = new Line2D.Double(x, dataArea.getMinY(), x, dataArea.getMaxY()) ;
      }
      g2.draw(this.verticalTraceLine) ;
    }

    g2.setPaintMode() ;
  }

  /**
   * <zh>ƴֱĸ.</zh>
   * <en>Draws a horizontal line used to trace the mouse position to the horizontal axis..</en>
   *
   * @param g2 Graphics2D
   * @param y double <zh>ߵy.</zh>
   * <en>The y-coordinate of the trace line.</en>
   */
  private void drawVerticalAxisTrace(Graphics2D g2, double y) {
    Rectangle2D dataArea = getScreenDataArea() ;

    g2.setXORMode(Color.orange) ;
    if (((int) dataArea.getMinY() < y) && (y < (int) dataArea.getMaxY())) {

      if (this.horizontalTraceLine != null) {
        g2.draw(this.horizontalTraceLine) ;
        this.horizontalTraceLine.setLine(dataArea.getMinX(), y, dataArea.getMaxX(), y) ;
      } else {
        this.horizontalTraceLine = new Line2D.Double(dataArea.getMinX(), y, dataArea.getMaxX(), y) ;
      }
      g2.draw(this.horizontalTraceLine) ;
    }

    g2.setPaintMode() ;
  }

  /**
   * <zh>ChartMouseListener.</zh>
   * <en>Adds a listener.</en>
   *
   * @param listener ChartMouseListener <zh>.</zh>
   * <en>The listener.</en>
   */
  public void addChartMouseListener(IChartMouseListener listener) {
    if (listener == null) {
      throw new IllegalArgumentException("Null 'listener' argument.") ;
    }

    this.chartMouseListeners.add(IChartMouseListener.class, listener) ;
  }

  /**
   * <zh>ɾChartMouseListener.</zh>
   * <en>Removes a listener.</en>
   *
   * @param listener ChartMouseListener <zh>.</zh>
   * <en>The listener.</en>
   */
  public void removeChartMouseListener(IChartMouseListener listener) {
    this.chartMouseListeners.remove(IChartMouseListener.class, listener) ;
  }

  /**
   * <zh>еļ.</zh>
   * <en>Returns an array of the listeners of the given type registered with the panel.</en>
   *
   * @param listenerType Class <zh>.</zh>
   * <en>The listener class.</en>
   *
   * @return EventListener[] <zh>.</zh>
   * <en>An array of listeners.</en>
   */
  public EventListener[] getListeners(Class listenerType) {
    if (listenerType == IChartMouseListener.class) {
      // fetch listeners from local storage
      return this.chartMouseListeners.getListeners(listenerType) ;
    } else {
      return super.getListeners(listenerType) ;
    }
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writePaint(stream, this.zoomFillPaint) ;
    SerialUtil.writePaint(stream, this.zoomBorderPaint) ;
  }

  /**s
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.zoomFillPaint = SerialUtil.readPaint(stream) ;
    this.zoomBorderPaint = SerialUtil.readPaint(stream) ;

    this.chartMouseListeners = new EventListenerList() ;
    if (this.chart != null) {
      this.chart.addChangeListener(this) ;
    }
  }
}
