package com.ijchart.xychart.client ;

import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Font ;
import java.awt.GradientPaint ;
import java.awt.Graphics2D ;
import java.awt.Image ;
import java.awt.Paint ;
import java.awt.RenderingHints ;
import java.awt.Stroke ;
import java.awt.geom.AffineTransform ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.awt.image.BufferedImage ;
import java.io.File ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.OutputStream ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Collection ;
import java.util.Collections ;
import java.util.Iterator ;
import java.util.List ;

import javax.imageio.ImageIO ;
import javax.swing.event.EventListenerList ;

import com.ijchart.xychart.chartinfo.ChartShapeCollection ;
import com.ijchart.xychart.chartinfo.draw.IChangeChartShape ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.GradientPaintTransformType ;
import com.ijchart.xychart.event.ChartChangeEvent ;
import com.ijchart.xychart.event.ChartProgressEvent ;
import com.ijchart.xychart.event.PlotChangeEvent ;
import com.ijchart.xychart.event.TitleChangeEvent ;
import com.ijchart.xychart.listener.IChartChangeListener ;
import com.ijchart.xychart.listener.IChartProgressListener ;
import com.ijchart.xychart.listener.IPlotChangeListener ;
import com.ijchart.xychart.listener.ITitleChangeListener ;
import com.ijchart.xychart.plot.AbstractAxisPlot ;
import com.ijchart.xychart.plot.AbstractPlot ;
import com.ijchart.xychart.plot.NoAxisPlot ;
import com.ijchart.xychart.render.AbstractRender ;
import com.ijchart.xychart.shape.RectangleImageChartShape ;
import com.ijchart.xychart.shape.RectangleTextChartShape ;
import com.ijchart.xychart.title.AbstractTitle ;
import com.ijchart.xychart.title.ITitle ;
import com.ijchart.xychart.title.LegendTitle ;
import com.ijchart.xychart.title.TextTitle ;
import com.ijchart.xychart.ui.Align ;
import com.ijchart.xychart.ui.Rectangle ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.ui.RectangleInsets ;
import com.ijchart.xychart.ui.supplier.IDrawingSupplier ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;
import com.ijchart.xychart.util.StringUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>chart.</zh>
 * <en>This class used to draw charts.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class IJChart implements ITitleChangeListener, IPlotChangeListener, Serializable,
  Cloneable {

  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 5931509414010058549L ;

  /**
   * <zh>ȱʡtitle.</zh>
   * <en>The default font for titles.</en>
   */
  public static final Font DEFAULT_TITLE_FONT = new Font("SansSerif", Font.BOLD, 18) ;

  /**
   * <zh>ȱʡıɫ.</zh>
   * <en>The default background color.</en>
   */
  public static final Paint DEFAULT_BACKGROUND_PAINT = Color.white ;

  /**
   * <zh>ȱʡGradientPaint任.</zh>
   * <en>The defeault GradientPaint transform type.</en>
   */
  public static final GradientPaintTransformType DEFAULT_GRADIENT_PAINT_TRANSFORM_TYPE =
    GradientPaintTransformType.HORIZONTAL ;

  /**
   * <zh>ȱʡıͼƬ.</zh>
   * <en>The default background image.</en>
   */
  public static final Image DEFAULT_BACKGROUND_IMAGE = null ;

  /**
   * <zh>ȱʡıͼƬ뷽ʽ.</zh>
   * <en>The default background image alignment.</en>
   */
  public static final Align DEFAULT_BACKGROUND_IMAGE_ALIGNMENT = Align.FIT ;

  /**
   * <zh>ȱʡıͼƬalphaֵ.</zh>
   * <en>The default background image alpha.</en>
   */
  public static final float DEFAULT_BACKGROUND_IMAGE_ALPHA = 0.5f ;

  /**
   * <zh>ڿǷ֪ͨchartı¼ȱʡֵ.</zh>
   * <en>A default flag that can be used to enable/disable notification of chart change events.</en>
   */
  public static final boolean DEFAULT_NOTIFY = true ;

  /**
   * <zh>ȱʡ߿Ƿɼ.</zh>
   * <en>A default flag that controls whether or not the chart border is drawn. </en>
   */
  public static final boolean DEFAULT_BORDER_VISIBLE = false ;

  /**
   * <zh>ȱʡǰalpha͸ֵ.</zh>
   * <en>The default foreground alpha transparency.</en>
   */
  public static final float DEFAULT_BACKGROUND_ALPHA = 1.0f ;

  /**
   * <zh>߿Ƿɼ.</zh>
   * <en>A flag that controls whether or not the chart border is drawn. </en>
   */
  private boolean borderVisible ;

  /**
   * <zh>Ʊ߿Stroke.</zh>
   * <en>The stroke used to draw the chart border (if visible).</en>
   */
  private transient Stroke borderStroke ;

  /**
   * <zh>Ʊ߿Paint.</zh>
   * <en>The paint used to draw the chart border (if visible).</en>
   */
  private transient Paint borderPaint ;

  /**
   * <zh>chart߿ͻ֮Ŀհ.</zh>
   * <en>The padding between the chart border and the chart drawing area.</en>
   */
  private RectangleInsets padding ;

  /**
   * <zh>.</zh>
   * <en>The title font.</en>
   */
  private Font titleFont ;

  /**
   * <zh>.</zh>
   * <en>The chart title (optional).</en>
   */
  private ITitle title ;

  /**
   * <zh>ӱ,0.</zh>
   * <en>The chart subtitles (zero, one or many).  This field should never be null.</en>
   */
  private List subtitles ;

  /**
   * <zh>ͼ.</zh>
   * <en>The chart legend title.</en>
   */
  private LegendTitle legend ;

  /**
   * <zh>Ƿʾͼ.</zh>
   * <en>A boolean flag that controls whether or not the legend is visible.</en>
   */
  private boolean showLegend ;

  /**
   * <zh>Plot.</zh>
   * <en>Draws the visual representation of the data.</en>
   */
  private AbstractPlot plot ;

  /**
   * <zh>ƱPaint.</zh>
   * <en>Paint used to draw the background of the chart.</en>
   */
  private transient Paint backgroundPaint ;

  /**
   * <zh>Gradient Paint 任.</zh>
   * <en>Gradient Paint transform type.</en>
   */
  private GradientPaintTransformType gradientPaintTransformType ;

  /**
   * <zh>͸alphaֵ.</zh>
   * <en>The alpha transparency for the background paint.</en>
   */
  private float backgroundAlpha ;

  /**
   * <zh>ͼƬ.</zh>
   * <en>An optional background image for the chart. </en>
   */
  private transient Image backgroundImage ; // todo: not serialized yet

  /**
   * <zh>ͼƬ뷽ʽ.</zh>
   * <en>The alignment for the background image.</en>
   */
  private Align backgroundImageAlignment = Align.FIT ;

  /**
   * <zh>ͼƬalphaֵ.</zh>
   * <en>The alpha transparency for the background image.</en>
   */
  private float backgroundImageAlpha = 0.5f ;

  /**
   * <zh>ڿǷ֪ͨchartı¼.</zh>
   * <en>A flag that can be used to enable/disable notification of chart change events.</en>
   */
  private boolean notify ;

  /**
   * <zh>ͼͼζļ.</zh>
   * <en>The chart shapes collection.</en>
   */
  private ChartShapeCollection chartShapes ;

  /**
   * <zh>¼󼯺.</zh>
   * <en>Storage for registered change listeners.</en>
   */
  private transient EventListenerList changeListeners ;

  /**
   * <zh>ȼ󼯺.</zh>
   * <en>Storage for registered progress listeners.</en>
   */
  private transient EventListenerList progressListeners ;

  /**
   * <zh>ͼθı.</zh>
   * <en>The shape change object.</en>
   */
  private IChangeChartShape changeChartShape ;

  /**
   * <zh>һ,Ĭϲͼ.</zh>
   * <en>Creates a new chart based on the supplied plot.  The chart will have a legend added
   * automatically, but no title (although you can easily add one later).</en>
   *
   * @param plot AbstractPlot <zh>Plot (Ϊnull).</zh>
   * <en>The plot (null not permitted).</en>
   */
  public IJChart(AbstractPlot plot) {
    this(null, null, plot, true) ;
  }

  /**
   * <zh>һ,Ĭϲͼ.</zh>
   * <en>Creates a new chart with the given title and plot.  A default font
   * (@link DEFAULT_TITLE_FONT) is used for the title, and the chart will
   * have a legend added automatically.</en>
   *
   * @param title String <zh>ַ (Ϊnull).</zh>
   * <en>The chart title (null permitted).</en>
   * @param plot AbstractPlot <zh>Plot (Ϊnull).</zh>
   * <en>The plot (null not permitted).</en>
   */
  public IJChart(String title, AbstractPlot plot) {
    this(title, DEFAULT_TITLE_FONT, plot, true) ;
  }

  /**
   * <zh>һ,Ĭϲͼ.</zh>
   * <en>Creates a new chart with the given title and plot.  The
   * createLegend argument specifies whether or not a legend
   * should be added to the chart.</en>
   *
   * @param title String <zh>ַ (Ϊnull).</zh>
   * <en>The chart title (null permitted).</en>
   * @param titleFont Font <zh> (Ϊnull).</zh>
   * <en>The font for displaying the chart title (null permitted).</en>
   * @param plot AbstractPlot <zh>Plot (Ϊnull).</zh>
   * <en>The plot (null not permitted).</en>
   * @param createLegend <zh>Ƿ񴴽ͼ.</zh>
   * <en>A flag indicating whether or not a legend should be created for the chart.</en>
   */
  public IJChart(String title, Font titleFont, AbstractPlot plot, boolean createLegend) {
    if (plot == null) {
      throw new NullPointerException("Null 'plot' argument.") ;
    }

    this.notify = DEFAULT_NOTIFY ;
    this.borderVisible = DEFAULT_BORDER_VISIBLE ;
    this.borderStroke = new BasicStroke(1.0f) ;
    this.borderPaint = Color.black ;
    this.padding = RectangleInsets.INSETS_ZERO ;

    this.plot = plot ;
    this.plot.addChangeListener(this) ;

    this.subtitles = new ArrayList() ;

    if (createLegend) {
      LegendTitle legendTitle = new LegendTitle(this.plot.getRenders()) ;
      legendTitle.setPosition(RectangleEdge.BOTTOM) ;
      legendTitle.setBorderVisible(true) ;
      legendTitle.addChangeListener(this) ;

      this.legend = legendTitle ;
    }
    this.showLegend = createLegend ;

    // add the chart title, if one has been specified...
    if (!StringUtil.isEmptyWithTrim(title)) {
      if (titleFont == null) {
        this.titleFont = DEFAULT_TITLE_FONT ;
      } else {
        this.titleFont = titleFont ;
      }

      TextTitle textTitle = new TextTitle(title, this.titleFont) ;
      textTitle.addChangeListener(this) ;
      this.title = textTitle ;
    }

    this.backgroundPaint = DEFAULT_BACKGROUND_PAINT ;
    this.gradientPaintTransformType = DEFAULT_GRADIENT_PAINT_TRANSFORM_TYPE ;
    this.backgroundAlpha = DEFAULT_BACKGROUND_ALPHA ;

    this.backgroundImage = DEFAULT_BACKGROUND_IMAGE ;
    this.backgroundImageAlignment = DEFAULT_BACKGROUND_IMAGE_ALIGNMENT ;
    this.backgroundImageAlpha = DEFAULT_BACKGROUND_IMAGE_ALPHA ;

    this.chartShapes = null ;
    this.changeChartShape = null ;

    // create storage for listeners...
    this.progressListeners = new EventListenerList() ;
    this.changeListeners = new EventListenerList() ;
  }

  /**
   * <zh>ر߿Ƿɼ.</zh>
   * <en>Returns a flag that controls whether or not a border is drawn around the
   * outside of the chart.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isBorderVisible() {
    return this.borderVisible ;
  }

  /**
   * <zh>ñ߿Ƿɼ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets a flag that controls whether or not a border is drawn around the
   * outside of the chart,and sends a {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param visible boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setBorderVisible(boolean visible) {
    if (this.borderVisible != visible) {
      this.borderVisible = visible ;
      fireChartChanged() ;
    }
  }

  /**
   * <zh>ػƱ߿stroke.</zh>
   * <en>Returns the stroke used to draw the chart border (if visible).</en>
   *
   * @return Stroke <zh>Ʊ߿stroke (Ϊnull).</zh>
   * <en>The border stroke (never null).</en>
   */
  public Stroke getBorderStroke() {
    return this.borderStroke ;
  }

  /**
   * <zh>ûƱ߿stroke,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the stroke used to draw the chart border (if visible),and
   * sends a {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>Ʊ߿stroke (Ϊnull).</zh>
   * <en>The stroke (null not permitted).</en>
   */
  public void setBorderStroke(Stroke stroke) {
    if (stroke == null) {
      throw new IllegalArgumentException("Null 'stroke' argument.") ;
    }

    this.borderStroke = stroke ;
    fireChartChanged() ;
  }

  /**
   * <zh>ػƱ߿paint.</zh>
   * <en>Returns the paint used to draw the chart border (if visible).</en>
   *
   * @return Paint <zh>Ʊ߿paint (Ϊnull).</zh>
   * <en>The border paint (never null).</en>
   */
  public Paint getBorderPaint() {
    return this.borderPaint ;
  }

  /**
   * <zh>ûƱ߿paint,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the paint used to draw the chart border (if visible),and
   * sends a {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Ʊ߿paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setBorderPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.borderPaint = paint ;
    fireChartChanged() ;
  }

  /**
   * <zh>ͼı߿ͻ֮Ŀհ.</zh>
   * <en>Returns the padding between the chart border and the chart drawing area.</en>
   *
   * @return RectangleInsets <zh>RectangleInsets (Ϊnull).</zh>
   * <en>The padding (never null).</en>
   */
  public RectangleInsets getPadding() {
    return this.padding ;
  }

  /**
   * <zh>ͼı߿ͻ֮Ŀհ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the padding between the chart border and the chart drawing area,
   * and sends a {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param padding RectangleInsets <zh>RectangleInsets (Ϊnull).</zh>
   * <en>The padding (null not permitted).</en>
   */
  public void setPadding(RectangleInsets padding) {
    if (padding == null) {
      throw new IllegalArgumentException("Null 'padding' argument.") ;
    }

    this.padding = padding ;
    fireChartChanged() ;
  }

  /**
   * <zh>ͼ,ͼֻһ,Ҳö.</zh>
   * <en>Returns the main chart title.  Very often a chart will have just one
   * title, so we make this case simple by providing accessor methods for
   * the main title.  However, multiple titles are supported - see the
   * {@link #addSubtitle(ITitle)} method.</en>
   *
   * @return ITitle <zh>ITitle (Ϊnull).</zh>
   * <en>The chart title (possibly null).</en>
   */
  public ITitle getTitle() {
    return this.title ;
  }

  /**
   * <zh>,м{@link ChartChangeEvent}¼.벻ƱɰѴ˶Ϊnull.</zh>
   * <en>Sets the main title for the chart and sends a {@link ChartChangeEvent}
   * to all registered listeners.  If you do not want a title for the
   * chart, set it to null.  If you want more than one title on
   * a chart, use the {@link #addSubtitle(ITitle)} method.</en>
   *
   * @param title ITitle <zh>ITitle (Ϊnull).</zh>
   * <en>The title (null permitted).</en>
   */
  public void setTitle(ITitle title) {
    if (!(this.title == null && title == null)) {
      this.title = title ;
      if (this.title != null && title instanceof AbstractTitle) {
        AbstractTitle abstractTitle = (AbstractTitle) title ;
        if (!abstractTitle.hasListener(this)) {
          abstractTitle.addChangeListener(this) ;
        }
      }

      fireChartChanged() ;
    }
  }

  /**
   * <zh>ñ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the chart title and sends a {@link ChartChangeEvent} to all
   * registered listeners.  This is a convenience method that ends up calling
   * the {@link #setTitle(ITitle)} method.  If there is an existing title,
   * its text is updated, otherwise a new title using the default font is
   * added to the chart.  If text is null the chart title is set to null.</en>
   *
   * @param text <zh> (Ϊnull).</zh>
   * <en>The title text (null permitted).</en>
   */
  public void setTitle(String text) {
    if (!StringUtil.isEmpty(text)) {
      if (this.title == null) {
        TextTitle textTitle = new TextTitle(text, this.titleFont) ;
        textTitle.addChangeListener(this) ;
        setTitle(textTitle) ;
      } else {
        if (this.title instanceof TextTitle) {
          ((TextTitle)this.title).setText(text) ;
          fireChartChanged() ;
        }
      }
    } else {
      setTitle((ITitle)null) ;
    }
  }

  /**
   * <zh>һӱ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Adds a chart subtitle, and notifies registered listeners that the chart
   * has been modified.</en>
   *
   * @param subtitle ITitle <zh>ӱ (Ϊnull).</zh>
   * <en>The subtitle (null not permitted).</en>
   */
  public void addSubtitle(ITitle subtitle) {
    if (subtitle == null) {
      throw new IllegalArgumentException("Null 'subtitle' argument.") ;
    }

    if (subtitle instanceof AbstractTitle) {
      AbstractTitle abstractTitle = (AbstractTitle) subtitle ;
      if (!abstractTitle.hasListener(this)) {
        abstractTitle.addChangeListener(this) ;
      }
    }
    this.subtitles.add(subtitle) ;
    fireChartChanged() ;
  }

  /**
   * <zh>ָӱ.</zh>
   * <en>Returns a chart subtitle.</en>
   *
   * @param index int <zh>ֵ,ڵ0.</zh>
   * <en>The index of the chart subtitle (zero based).</en>
   *
   * @return ITitle <zh>ӱ.</zh>
   * <en>A chart subtitle.</en>
   */
  public ITitle getSubtitle(int index) {
    if ((index < 0) || (index >= this.subtitles.size())) {
      throw new IllegalArgumentException("Index out of range.") ;
    }

    return (ITitle)this.subtitles.get(index) ;
  }

  /**
   * <zh>һضλһӱ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Adds a subtitle at a particular position in the subtitle list, and sends
   * a {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param index int <zh>ֵ,ڱ[0,#getSubtitleCount()].</zh>
   * <en>The index (in the range 0 to {@link #getSubtitleCount()}).</en>
   * @param subtitle ITitle <zh>ӱ (Ϊnull).</zh>
   * <en>The subtitle to add (null not permitted).</en>
   */
  public void addSubtitle(int index, ITitle subtitle) {
    if (index < 0 || index > getSubtitleCount()) {
      throw new IllegalArgumentException("The 'index' argument is out of range.") ;
    }
    if (subtitle == null) {
      throw new IllegalArgumentException("Null 'subtitle' argument.") ;
    }

    if (subtitle instanceof AbstractTitle) {
      AbstractTitle abstractTitle = (AbstractTitle) subtitle ;
      if (!abstractTitle.hasListener(this)) {
        abstractTitle.addChangeListener(this) ;
      }
    }

    this.subtitles.add(index, subtitle) ;
    fireChartChanged() ;
  }

  /**
   * <zh>ɾеӱ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Clears all subtitles from the chart and sends a {@link ChartChangeEvent}
   * to all registered listeners.</en>
   */
  public void clearSubtitles() {
    Iterator iterator = this.subtitles.iterator() ;
    while (iterator.hasNext()) {
      ITitle t = (ITitle) iterator.next() ;
      if (t instanceof AbstractTitle) {
        ((AbstractTitle) t).removeChangeListener(this) ;
      }
    }

    this.subtitles.clear() ;
    fireChartChanged() ;
  }

  /**
   * <zh>ɾָӱ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Removes the specified subtitle and sends a {@link ChartChangeEvent} to
   * all registered listeners.</en>
   *
   * @param title ITitle <zh>ӱ (Ϊnull).</zh>
   * <en>The subtitle.</en>
   */
  public void removeSubTitle(ITitle title) {
    if (title == null) {
      throw new IllegalArgumentException("Null 'title' argument.") ;
    }

    if (title instanceof AbstractTitle) {
      AbstractTitle abstractTitle = (AbstractTitle) title ;
      abstractTitle.removeChangeListener(this) ;
    }

    this.subtitles.remove(title) ;
    fireChartChanged() ;
  }

  /**
   * <zh>ɾеӱ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Removes all subtitle and sends a {@link ChartChangeEvent} to
   * all registered listeners.</en>
   */
  public void removeAllSubTitle() {
    if (this.subtitles.size() > 0) {
      for (int i = 0 ; i < this.subtitles.size() ; i++) {
        Object obj = this.subtitles.get(i) ;
        if (obj instanceof AbstractTitle) {
          AbstractTitle abstractTitle = (AbstractTitle) obj ;
          abstractTitle.removeChangeListener(this) ;
        }
      }

      this.subtitles.clear() ;
      fireChartChanged() ;
    }
  }

  /**
   * <zh>ӱ󼯺,˼ǲ޸ĵ.</zh>
   * <en>Returns the list of subtitles for the chart,and this collection is unmodifiable.</en>
   *
   * @return Collection <zh>ӱ󼯺 (Ϊ,Ϊnull).</zh>
   * <en>The subtitle list (possibly empty, but never null).</en>
   */
  public Collection getSubtitles() {
    return Collections.unmodifiableCollection(this.subtitles) ;
  }

  /**
   * <zh>ӱĸ.</zh>
   * <en>Returns the number of titles for the chart.</en>
   *
   * @return int <zh>ӱĸ.</zh>
   * <en>The number of titles for the chart.</en>
   */
  public int getSubtitleCount() {
    return this.subtitles.size() ;
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Returns the chart legend title.</en>
   *
   * @return LegendTitle <zh>ͼ.</zh>
   * <en>The chart legend title.</en>
   */
  public LegendTitle getLegendTitle() {
    return this.legend ;
  }

  /**
   * <zh>ͼ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the chart legend title, and notifies registered listeners that the chart
   * has been modified.</en>
   *
   * @param legend LegendTitle <zh>ͼ (Ϊnull).</zh>
   * <en>The legend title (null not permitted).</en>
   */
  public void setLegendTitle(LegendTitle legend) {
    if (legend == null) {
      throw new IllegalArgumentException("Null 'legend' argument.") ;
    }

    if (!legend.hasListener(this)) {
      legend.addChangeListener(this) ;
    }

    this.legend = legend ;
    fireChartChanged() ;
  }

  /**
   * <zh>Ƿʾͼ.</zh>
   * <en>Returns whether or not the legend is visible.</en>

   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isShowLegend() {
    return this.showLegend ;
  }

  /**
   * <zh>Ƿʾͼ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets whether or not the legend is visible,and
   * sends a {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setShowLegend(boolean flag) {
    if (this.showLegend != flag) {
      this.showLegend = flag ;
      fireChartChanged() ;
    }
  }

  /**
   * <zh>ͼPlot.</zh>
   * <en>Returns the plot for the chart.  The plot is a class responsible for
   * coordinating the visual representation of the data, including the axes (if any).</en>
   *
   * @return AbstractPlot <zh>Plot.</zh>
   * <en>The plot.</en>
   */
  public AbstractPlot getPlot() {
    return this.plot ;
  }

  /**
   * <zh>ػƱPaint.</zh>
   * <en>Returns the paint used to draw the background of the chart.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (possibly null).</en>
   */
  public Paint getBackgroundPaint() {
    return this.backgroundPaint ;
  }

  /**
   * <zh>ûƱPaint,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the paint used to fill the chart background and sends a
   * {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (possibly null).</en>
   */
  public void setBackgroundPaint(Paint paint) {
    if (!(this.backgroundPaint == null && paint == null)) {
      this.backgroundPaint = paint ;
      fireChartChanged() ;
    }
  }

  /**
   * <zh>ͼڲɫչ.</zh>
   * <en>Get the ChartShape GradientPaintTransformType type.</en>
   *
   * @return GradientPaintTransformType
   */
  public GradientPaintTransformType getGradientPaintTransformType() {
    return this.gradientPaintTransformType ;
  }

  /**
   * <zh>ͼڲɫչ,PaintΪGradientPaint,
   * м{@link com.ijchart.xychart.event.ChartShapeChangeEvent}¼.</zh>
   * <en>Sets the ChartShape GradientPaintTransformType type,use only when fill Paint is
   *  GradientPaint,and sends a {@link com.ijchart.xychart.event.ChartShapeChangeEvent}
   *  to all registered listeners.</en>
   *
   * @param transformType GradientPaintTransformType <zh>任 (Ϊnull).</zh>
   * <en>The gradient paint transform type (null permitted).</en>
   */
  public void setGradientPaintTransformType(GradientPaintTransformType transformType) {
    this.gradientPaintTransformType = transformType ;
    fireChartChanged() ;
  }

  /**
   * <zh>رalphaֵ.</zh>
   * <en>Returns the alpha transparency of the plot area background.</en>
   *
   * @return <zh>͸alphaֵ.</zh>
   * <en>The alpha transparency.</en>
   */
  public float getBackgroundAlpha() {
    return this.backgroundAlpha ;
  }

  /**
   * <zh>ñ͸alphaֵ,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the alpha transparency of the plot area background, and notifies
   * registered listeners that the plot has been modified.</en>
   *
   * @param alpha <zh>alphaֵ,ֵڱ[0.0,1.0].</zh>
   * <en>The new alpha value (in the range 0.0f to 1.0f).</en>
   */
  public void setBackgroundAlpha(float alpha) {
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    if (this.backgroundAlpha != alpha) {
      this.backgroundAlpha = alpha ;
      fireChartChanged() ;
    }
  }

  /**
   * <zh>رͼƬ.</zh>
   * <en>Returns the background image for the chart. </en>
   *
   * @return Image <zh>ͼƬ (Ϊnull).</zh>
   * <en>The image (possibly null).</en>
   */
  public Image getBackgroundImage() {
    return this.backgroundImage ;
  }

  /**
   * <zh>ñͼƬ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the background image for the chart and sends a
   * {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param image Image <zh>ͼƬ (Ϊnull).</zh>
   * <en>The image (null permitted).</en>
   */
  public void setBackgroundImage(Image image) {
    if (!(this.backgroundImage == null && image == null)) {
      this.backgroundImage = image ;
      fireChartChanged() ;
    }
  }

  /**
   * <zh>رͼƬĶ뷽ʽ.</zh>
   * <en>Returns the background image alignment. Alignment constants are defined
   * in the {@link Align} class.</en>
   *
   * @return Align <zh>ͼƬĶ뷽ʽ.</zh>
   * <en>The alignment.</en>
   */
  public Align getBackgroundImageAlignment() {
    return this.backgroundImageAlignment ;
  }

  /**
   * <zh>ñͼƬĶ뷽ʽ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the background alignment.  Alignment options are defined by the
   * {@link Align} class.</en>
   *
   * @param alignment Align <zh>뷽ʽ.</zh>
   * <en>The alignment.</en>
   */
  public void setBackgroundImageAlignment(Align alignment) {
    if (!this.backgroundImageAlignment.equals(alignment)) {
      this.backgroundImageAlignment = alignment ;
      fireChartChanged() ;
    }
  }

  /**
   * <zh>رͼƬalphaֵ.</zh>
   * <en>Returns the alpha-transparency for the chart's background image.</en>
   *
   * @return float <zh>ͼƬalphaֵ.</zh>
   * <en>The alpha-transparency for background image.</en>
   */
  public float getBackgroundImageAlpha() {
    return this.backgroundImageAlpha ;
  }

  /**
   * <zh>ñͼƬalphaֵ,м{@link ChartChangeEvent}¼.</zh>
   * <en>Sets the alpha-transparency for the chart's background image.
   * Registered listeners are notified that the chart has been changed.</en>
   *
   * @param alpha float <zh>ͼƬalphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha-transparency for background image,must between 0.0 and 1.0 .</en>
   */
  public void setBackgroundImageAlpha(float alpha) {
    if (this.backgroundImageAlpha != alpha) {
      this.backgroundImageAlpha = alpha ;
      fireChartChanged() ;
    }
  }

  /**
   * <zh>¼仯ʱǷ֪ͨ仯.</zh>
   * <en>Returns a flag that controls whether or not change events are sent to
   * registered listeners.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isNotify() {
    return this.notify ;
  }

  /**
   * <zh>¼仯ʱǷ֪ͨ仯.</zh>
   * <en>Sets a flag that controls whether or not listeners receive
   * {@link ChartChangeEvent} notifications.</en>
   *
   * @param notify boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setNotify(boolean notify) {
    this.notify = notify ;
    // if the flag is being set to true, there maybe queued up changes...
    if (notify) {
      fireChartChanged() ;
    }
  }

  /**
   * <zh>ͼͼεļ.</zh>
   * <en>Returns the collection of chart shapes.</en>
   *
   * @return ChartShapeCollection <zh>ͼͼεļ (Ϊnull).</zh>
   * <en>The collection of chart shapes (never null).</en>
   */
  public ChartShapeCollection getChartShapes() {
    return this.chartShapes ;
  }

  /**
   * <zh>ͼθı.</zh>
   * <en>Returns the shape change object.</en>
   *
   * @return IChangeChartShape <zh>ͼθı (Ϊnull).</zh>
   * <en>The shape change object (maybe null).</en>
   */
  public IChangeChartShape getChangeChartShape() {
    return this.changeChartShape ;
  }

  /**
   * <zh>ͼθı,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the shape change object,and notifies
   * registered listeners that the plot has been modified.</en>
   *
   * @param change IChangeChartShape <zh>ͼθı.</zh>
   * <en>The shape change object.</en>
   */
  public void setChangeChartShape(IChangeChartShape change) {
    if (!(this.changeChartShape == null && change == null)) {
      this.changeChartShape = change ;
      fireChartChanged() ;
    }
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Draws the chart on a Java 2D graphics device (such as the screen or a printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param chartArea Rectangle2D <zh>ͼdata() (Ϊnull).</zh>
   * <en>The area within which the data should be drawn (null not permitted).</en>
   */
  public void draw(Graphics2D g2, Rectangle2D chartArea) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (chartArea == null) {
      throw new IllegalArgumentException("Null 'chartArea' argument.") ;
    }

    //еrenderһIDrawingSupplier
    Iterator iter = this.plot.getRenders().iterator() ;
    IDrawingSupplier supplier = this.plot.getDrawingSupplier() ;
    supplier.setCanReset(true) ;
    supplier.reset() ;

    while (iter.hasNext()) {
      Object obj = iter.next() ;
      if (obj != null && obj instanceof AbstractRender) {
        ((AbstractRender) obj).setDrawingSupplier(this.plot.getDrawingSupplier(), false) ;
      }
    }

    if (this.chartShapes != null) {
      this.chartShapes.clearShapesListener() ; //ɾͼεļ
    }
    this.chartShapes = new ChartShapeCollection() ;
    this.chartShapes.setDrawArea(chartArea) ;
    this.chartShapes.setChangeChartShape(this.changeChartShape) ;

    notifyListeners(new ChartProgressEvent(this, this,
                                           ChartProgressEvent.PROGRESS_CALCULATE_STARTED)) ;

    drawBackground(chartArea) ;
    drawBackgroundImage(chartArea) ;
    drawBorder(chartArea) ;

    this.padding.trim(chartArea) ;

    drawTitles(g2, chartArea) ; //Ʊ
    drawLengend(g2, chartArea) ; //ͼ

    supplier.setCanReset(true) ;
    supplier.reset() ;
    this.plot.draw(g2, chartArea, this.chartShapes) ;

    notifyListeners(new ChartProgressEvent(this, this,
                                           ChartProgressEvent.PROGRESS_CALCULATE_FINISHED)) ;

    this.chartShapes.setIJChart(this) ;
  }

  /**
   * <zh>ʮ.</zh>
   * <en>Draws the crosshair on a Java 2D graphics device (such as the screen or a printer).</en>
   *
   * @param mouseClickPoint Point2D <zh>ʱ (Ϊnull).</zh>
   * <en>The mouse click point (null permitted).</en>
   *
   * @return boolean <zh>Ƿʮ߷˱仯.</zh>
   * <en>Returns whether or not crosshair is changed.</en>
   */
  public boolean drawCrosshair(Point2D mouseClickPoint) {
    if (this.plot instanceof NoAxisPlot) {
      return false ;
    }

    return this.plot.drawCrosshair(mouseClickPoint) ;
  }

  /**
   * <zh>Ʊɫ,ƱPaintGradientPaint,gradientĻƷͼĻƷһ.</zh>
   * <en>Fills the specified area with the background paint.  If the background
   * paint is an instance of GradientPaint, the gradient will
   * run in the direction suggested by the plot's orientation.</en>
   *
   * @param chartArea <zh> (Ϊnull).</zh>
   * <en>The background draw area (null not permitted).</en>
   */
  protected void drawBackground(Rectangle2D chartArea) {
    if (this.backgroundPaint == null) {
      return ;
    }

    Paint p = this.backgroundPaint ;
    if (p instanceof GradientPaint && this.plot instanceof AbstractAxisPlot) {
      ChartOrientation orientation = ((AbstractAxisPlot)this.plot).getOrientation() ;
      GradientPaint gp = (GradientPaint) p ;

      if (orientation.equals(ChartOrientation.VERTICAL)) {
        p = new GradientPaint((float) chartArea.getCenterX(),
                              (float) chartArea.getMaxY(), gp.getColor1(),
                              (float) chartArea.getCenterX(), (float) chartArea.getMinY(),
                              gp.getColor2()) ;
      } else if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        p = new GradientPaint((float) chartArea.getMinX(),
                              (float) chartArea.getCenterY(), gp.getColor1(),
                              (float) chartArea.getMaxX(), (float) chartArea.getCenterY(),
                              gp.getColor2()) ;
      }
    }

    //ɫͼ
    RectangleTextChartShape shape = new RectangleTextChartShape(chartArea, null) ;
    shape.setDrawBorder(false) ;
    shape.setFillPaint(p) ;
    shape.setFillShape(true) ;
    shape.setFillAlpha(this.backgroundAlpha) ;
    if (this.gradientPaintTransformType != null) {
      shape.setGradientPaintTransformType(this.gradientPaintTransformType) ;
    }

    this.chartShapes.setBackgroundShape(shape) ;
  }

  /**
   * <zh>ָڻƱͼƬ.</zh>
   * <en>Draws the background image (if there is one) aligned within the specified area.</en>
   *
   * @param chartArea <zh> (Ϊnull).</zh>
   * <en>The background draw area (null not permitted).</en>
   */
  protected void drawBackgroundImage(Rectangle2D chartArea) {
    if (this.backgroundImage != null) {
      Rectangle2D dest = new Rectangle2D.Double(0.0, 0.0, this.backgroundImage.getWidth(null),
                                                this.backgroundImage.getHeight(null)) ;
      Rectangle2D imageRect = Align.align(dest, chartArea, this.backgroundImageAlignment) ;
      imageRect.setRect(imageRect.getX(), imageRect.getY(), imageRect.getWidth() + 1,
                        imageRect.getHeight() + 1) ;

      //ͼƬͼ
      RectangleImageChartShape shape = new RectangleImageChartShape(imageRect, this.backgroundImage) ;
      shape.setDrawBorder(false) ;
      shape.setFillShape(false) ; //ͼ
      shape.setImageAlpha(this.backgroundImageAlpha) ;

      this.chartShapes.setBackgroundImageShape(shape) ;
    }
  }

  /**
   * <zh>߿.</zh>
   * <en>Draws the chart border.</en>
   *
   * @param chartArea <zh> (Ϊnull).</zh>
   * <en>The background draw area (null not permitted).</en>
   */
  protected void drawBorder(Rectangle2D chartArea) {
    if (this.borderVisible && this.borderStroke != null && this.borderPaint != null) {
      RectangleTextChartShape shape = new RectangleTextChartShape(chartArea, null) ;
      shape.setDrawBorder(true) ;
      shape.setBorderPaint(this.borderPaint) ;
      shape.setBorderStroke(this.borderStroke) ;
      shape.setFillShape(false) ;

      this.chartShapes.setBorderShape(shape) ;
    }
  }

  /**
   * <zh>Ʊ.</zh>
   * <en>Draws the titles on a Java 2D graphics device (such as the screen or a printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param chartArea Rectangle2D <zh>ͼdata() (Ϊnull).</zh>
   * <en>The area within which the data should be drawn (null not permitted).</en>
   */
  protected void drawTitles(Graphics2D g2, Rectangle2D chartArea) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (chartArea == null) {
      throw new IllegalArgumentException("Null 'chartArea' argument.") ;
    }

    List tmpTitles = new ArrayList() ;
    tmpTitles.add(this.title) ;
    tmpTitles.addAll(this.subtitles) ;

    for (int i = 0 ; i < tmpTitles.size() ; i++) {
      Object obj = tmpTitles.get(i) ;
      if (obj == null) {
        continue ;
      }

      if (obj instanceof AbstractTitle) {
        Rectangle titleRect = ((ITitle) obj).draw(g2, chartArea, this.chartShapes) ;
        if (titleRect == null) {
          continue ;
        }

        AbstractTitle tmpTitle = (AbstractTitle) obj ;
        //ݱλüchart
        RectangleEdge titlePosition = tmpTitle.getPosition() ;
        //ļӼspaceΪӼ
        final double space = 1 ;
        if (titlePosition.equals(RectangleEdge.TOP)) {
          chartArea.setRect(chartArea.getX(), chartArea.getY() + titleRect.getHeight() + space,
                            chartArea.getWidth(), chartArea.getHeight() - titleRect.getHeight()) ;
        } else if (titlePosition.equals(RectangleEdge.BOTTOM)) {
          chartArea.setRect(chartArea.getX(), chartArea.getY(), chartArea.getWidth(),
                            chartArea.getHeight() - titleRect.getHeight() - space) ;
        } else if (titlePosition.equals(RectangleEdge.LEFT)) {
          chartArea.setRect(chartArea.getX() + titleRect.getHeight() + space, chartArea.getY(),
                            chartArea.getWidth() - titleRect.getHeight(), chartArea.getHeight()) ;
        } else if (titlePosition.equals(RectangleEdge.RIGHT)) {
          chartArea.setRect(chartArea.getX(), chartArea.getY(),
                            chartArea.getWidth() - titleRect.getHeight() - space,
                            chartArea.getHeight()) ;
        }
      }
    }
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Draws the legend on a Java 2D graphics device (such as the screen or a printer).</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param chartArea Rectangle2D <zh>ͼdata() (Ϊnull).</zh>
   * <en>The area within which the data should be drawn (null not permitted).</en>
   */
  protected void drawLengend(Graphics2D g2, Rectangle2D chartArea) {
    if (g2 == null) {
      throw new IllegalArgumentException("Null 'g2' argument.") ;
    }
    if (chartArea == null) {
      throw new IllegalArgumentException("Null 'chartArea' argument.") ;
    }

    if (isShowLegend() && this.legend != null) {
      Rectangle titleRect = this.legend.draw(g2, chartArea, this.chartShapes) ;
      if (titleRect == null) {
        return ;
      }

      //ݱλüchart
      RectangleEdge titlePosition = this.legend.getPosition() ;
      //ļӼspaceΪӼ
      final double space = 1 ;
      if (titlePosition.equals(RectangleEdge.TOP)) {
        chartArea.setRect(chartArea.getX(), chartArea.getY() + titleRect.getHeight() + space,
                          chartArea.getWidth(), chartArea.getHeight() - titleRect.getHeight()) ;
      } else if (titlePosition.equals(RectangleEdge.BOTTOM)) {
        chartArea.setRect(chartArea.getX(), chartArea.getY(), chartArea.getWidth(),
                          chartArea.getHeight() - titleRect.getHeight() - space) ;
      } else if (titlePosition.equals(RectangleEdge.LEFT)) {
        chartArea.setRect(chartArea.getX() + titleRect.getWidth() + space, chartArea.getY(),
                          chartArea.getWidth() - titleRect.getWidth(), chartArea.getHeight()) ;
      } else if (titlePosition.equals(RectangleEdge.RIGHT)) {
        chartArea.setRect(chartArea.getX(), chartArea.getY(),
                          chartArea.getWidth() - titleRect.getWidth() - space, chartArea.getHeight()) ;
      }
    }
  }

  /**
   * <zh>ͼƵһͼƬ.</zh>
   * <en>Creates and returns a buffered image into which the chart has been drawn.</en>
   *
   * @param imageWidth int <zh>ͼƬĿ.</zh>
   * <en>The image width.</en>
   * @param imageHeight int <zh>ͼƬĸ߶.</zh>
   * <en>The image height.</en>
   *
   * @return Image <zh>ͼƬ (Ϊnull).</zh>
   * <en>A image (never null).</en>
   */
  public BufferedImage createBufferedImage(int imageWidth, int imageHeight) {
    return createBufferedImage(imageWidth, imageHeight, imageWidth, imageHeight,
                               BufferedImage.TYPE_INT_RGB) ;
  }

  /**
   * <zh>ͼƵһͼƬ.</zh>
   * <en>Creates and returns a buffered image into which the chart has been drawn.</en>
   *
   * @param imageWidth int <zh>ͼƬĿ.</zh>
   * <en>The image width.</en>
   * @param imageHeight int <zh>ͼƬĸ߶.</zh>
   * <en>The image height.</en>
   * @param imageType int <zh>ͼƬ,BufferedImage.TYPE_INT_RGB .</zh>
   * <en>The image type.</en>
   *
   * @return Image <zh>ͼƬ (Ϊnull).</zh>
   * <en>A image (never null).</en>
   */
  public BufferedImage createBufferedImage(int imageWidth, int imageHeight, int imageType) {
    return createBufferedImage(imageWidth, imageHeight, imageWidth, imageHeight, imageType) ;
  }

  /**
   * <zh>ͼƵһͼƬ.</zh>
   * <en>Creates and returns a buffered image into which the chart has been drawn.</en>
   *
   * @param imageWidth int <zh>ͼƬĿ.</zh>
   * <en>The image width.</en>
   * @param imageHeight int <zh>ͼƬĸ߶.</zh>
   * <en>The image height.</en>
   * @param drawWidth int <zh>ʵʻƵĿ (ᱻŵӦͼƬ).</zh>
   * <en>The width for drawing the chart (will be scaled to fit image).</en>
   * @param drawHeight int <zh>ʵʻƵĸ߶ (ᱻŵӦͼƬ߶).</zh>
   * <en>The height for drawing the chart (will be scaled to fit image).</en>
   *
   * @return Image <zh>ͼƬ (Ϊnull).</zh>
   * <en>A image (never null).</en>
   */
  public BufferedImage createBufferedImage(int imageWidth, int imageHeight, double drawWidth,
                                           double drawHeight) {
    return createBufferedImage(imageWidth, imageHeight, drawWidth, drawHeight,
                               BufferedImage.TYPE_INT_RGB) ;
  }

  /**
   * <zh>ͼƵһͼƬ.</zh>
   * <en>Creates and returns a buffered image into which the chart has been drawn.</en>
   *
   * @param imageWidth int <zh>ͼƬĿ.</zh>
   * <en>The image width.</en>
   * @param imageHeight int <zh>ͼƬĸ߶.</zh>
   * <en>The image height.</en>
   * @param drawWidth int <zh>ʵʻƵĿ (ᱻŵӦͼƬ).</zh>
   * <en>The width for drawing the chart (will be scaled to fit image).</en>
   * @param drawHeight int <zh>ʵʻƵĸ߶ (ᱻŵӦͼƬ߶).</zh>
   * <en>The height for drawing the chart (will be scaled to fit image).</en>
   * @param imageType int <zh>ͼƬ,BufferedImage.TYPE_INT_RGB .</zh>
   * <en>The image type.</en>
   *
   * @return Image <zh>ͼƬ (Ϊnull).</zh>
   * <en>A image (never null).</en>
   */
  public BufferedImage createBufferedImage(int imageWidth, int imageHeight, double drawWidth,
                                           double drawHeight, int imageType) {
    double tmpWidth = Math.max(drawWidth, ChartPanel.DEFAULT_MINIMUM_DRAW_WIDTH) ;
    double tmpHeight = Math.max(drawHeight, ChartPanel.DEFAULT_MINIMUM_DRAW_HEIGHT) ;

    BufferedImage image = new BufferedImage(imageWidth, imageHeight, imageType) ;
    Graphics2D g2 = image.createGraphics() ;
    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON) ;
    g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
                        RenderingHints.VALUE_TEXT_ANTIALIAS_ON) ;
    double scaleX = imageWidth / tmpWidth ;
    double scaleY = imageHeight / tmpHeight ;
    AffineTransform st = AffineTransform.getScaleInstance(scaleX, scaleY) ;
    g2.transform(st) ;

    draw(g2, new Rectangle2D.Double(0, 0, tmpWidth, tmpHeight)) ;
    this.chartShapes.draw(g2) ;
    g2.dispose() ;

    return image ;
  }

  /**
   * <zh>һڼͼ仯¼ļ.</zh>
   * <en>Registers an object for notification of changes to the chart.</en>
   *
   * @param listener IChartChangeListener <zh>ͼ仯 (Ϊnull).</zh>
   * <en>The listener (null not permitted).</en>
   */
  public void addChangeListener(IChartChangeListener listener) {
    if (listener == null) {
      throw new IllegalArgumentException("Null 'listener' argument.") ;
    }

    this.changeListeners.add(IChartChangeListener.class, listener) ;
  }

  /**
   * <zh>ɾһͼ仯ļ.</zh>
   * <en>Deregisters an object for notification of changes to the chart.</en>
   *
   * @param listener IChartChangeListener <zh>ͼ仯 (Ϊnull).</zh>
   * <en>The listener (null not permitted).</en>
   */
  public void removeChangeListener(IChartChangeListener listener) {
    if (listener == null) {
      throw new IllegalArgumentException("Null 'listener' argument.") ;
    }

    this.changeListeners.remove(IChartChangeListener.class, listener) ;
  }

  /**
   * <zh>֪ͨмͼ˱仯.</zh>
   * <en>Sends a default {@link ChartChangeEvent} to all registered listeners.</en>
   */
  public void fireChartChanged() {
    ChartChangeEvent event = new ChartChangeEvent(this) ;
    notifyListeners(event) ;
  }

  /**
   * <zh>֪ͨмͼ˱仯.</zh>
   * <en>Sends a {@link ChartChangeEvent} to all registered listeners.</en>
   *
   * @param event ChartChangeEvent <zh>ͼ仯¼.</zh>
   * <en>Information about the event that triggered the notification.</en>
   */
  protected void notifyListeners(ChartChangeEvent event) {
    if (this.notify) {
      Object[] listeners = this.changeListeners.getListenerList() ;
      for (int i = listeners.length - 2 ; i >= 0 ; i -= 2) {
        if (listeners[i] == IChartChangeListener.class) {
          ((IChartChangeListener) listeners[i + 1]).chartChanged(event) ;
        }
      }
    }
  }

  /**
   * <zh>һ.</zh>
   * <en>Registers an object for notification of progress events relating to the chart.</en>
   *
   * @param listener IChartProgressListener <zh>ͼƽȼ (Ϊnull).</zh>
   * <en>The object being registered (null not permitted).</en>
   */
  public void addProgressListener(IChartProgressListener listener) {
    if (listener == null) {
      throw new IllegalArgumentException("Null 'listener' argument.") ;
    }
    this.progressListeners.add(IChartProgressListener.class, listener) ;
  }

  /**
   * <zh>ɾһ.</zh>
   * <en>Deregisters an object for notification of changes to the chart.</en>
   *
   * @param listener IChartProgressListener <zh>ͼƽȼ (Ϊnull).</zh>
   * <en>The object being deregistered (null not permitted).</en>
   */
  public void removeProgressListener(IChartProgressListener listener) {
    if (listener == null) {
      throw new IllegalArgumentException("Null 'listener' argument.") ;
    }

    this.progressListeners.remove(IChartProgressListener.class, listener) ;
  }

  /**
   * <zh>ɾнȼ.</zh>
   * <en>Deregisters all object for notification of progress changes.</en>
   */
  public void removeAllProgressListener() {
    Object[] listeners = this.progressListeners.getListenerList() ;

    for (int i = listeners.length - 2 ; i >= 0 ; i -= 2) {
      this.progressListeners.remove(IChartProgressListener.class,
                                    (IChartProgressListener) listeners[i + 1]) ;
    }
  }

  /**
   * <zh>֪ͨмͼ仯¼.</zh>
   * <en>Sends a {@link ChartProgressEvent} to all registered listeners.</en>
   *
   * @param event ChartProgressEvent <zh>ͼ仯¼.</zh>
   * <en>Information about the event that triggered the notification.</en>
   */
  public void notifyListeners(ChartProgressEvent event) {
    Object[] listeners = this.progressListeners.getListenerList() ;
    for (int i = listeners.length - 2 ; i >= 0 ; i -= 2) {
      if (listeners[i] == IChartProgressListener.class) {
        ((IChartProgressListener) listeners[i + 1]).chartProgress(event) ;
      }
    }
  }

  /**
   * <zh>յtitle仯¼.</zh>
   * <en>Receives notification of a chart title change event.</en>
   *
   * @param event TitleChangeEvent <zh>title仯¼.</zh>
   * <en>The title change event.</en>
   */
  public void titleChanged(TitleChangeEvent event) {
    notifyListeners(new ChartChangeEvent(event.getTitle())) ;
  }

  /**
   * <zh>յPlot仯¼.</zh>
   * <en>Receives notification of a plot change event.</en>
   *
   * @param event PlotChangeEvent <zh>plot仯¼.</zh>
   * <en>The plot change event.</en>
   */
  public void plotChanged(PlotChangeEvent event) {
    notifyListeners(new ChartChangeEvent(this)) ;
  }

  /**
   * <zh>chart洢ΪJPEGʽͼƬ.</zh>
   * <en>Save chart to JPEG image.</en>
   *
   * @param filePath String <zh>ͼƬļ· (Ϊnull).</zh>
   * <en>The image file path (null not permitted).</en>
   * @param width int <zh>ͼƬĿ.</zh>
   * <en>The image width.</en>
   * @param height int <zh>ͼƬĸ߶.</zh>
   * <en>The image height.</en>
   *
   * @throws IOException
   */
  public void saveChartAsJPEG(String filePath, int width, int height) throws IOException {
    if (filePath == null) {
      throw new IllegalArgumentException("Null 'filePath' argument.") ;
    }

    File file = new File(filePath) ;
    saveChartAsJPEG(file, width, height) ;
  }

  /**
   * <zh>chart洢ΪJPEGʽͼƬ.</zh>
   * <en>Save chart to JPEG image.</en>
   *
   * @param stream OutputStream <zh>ͼƬļ (Ϊnull).</zh>
   * <en>An OutputStream to be written to (null not permitted).</en>
   * @param width int <zh>ͼƬĿ.</zh>
   * <en>The image width.</en>
   * @param height int <zh>ͼƬĸ߶.</zh>
   * <en>The image height.</en>
   *
   * @throws IOException
   */
  public void saveChartAsJPEG(OutputStream stream, int width, int height) throws IOException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument.") ;
    }

    BufferedImage bufferImg = createBufferedImage(width, height) ;
    try {
      ImageIO.write(bufferImg, "JPG", stream) ;
    } catch (IOException ex) {
      throw ex ;
    }
  }

  /**
   * <zh>chart洢ΪJPEGʽͼƬ.</zh>
   * <en>Save chart to JPEG image.</en>
   *
   * @param imageFile File <zh>ͼƬļ (Ϊnull).</zh>
   * <en>The image file (null not permitted).</en>
   * @param width int <zh>ͼƬĿ.</zh>
   * <en>The image width.</en>
   * @param height int <zh>ͼƬĸ߶.</zh>
   * <en>The image height.</en>
   *
   * @throws IOException
   */
  public void saveChartAsJPEG(File imageFile, int width, int height) throws IOException {
    if (imageFile == null) {
      throw new IllegalArgumentException("Null 'imageFile' argument.") ;
    }

    BufferedImage bufferImg = createBufferedImage(width, height) ;
    try {
      ImageIO.write(bufferImg, "JPG", imageFile) ;
    } catch (IOException ex) {
      throw ex ;
    }
  }

  /**
   * <zh>chart洢ΪPNGʽͼƬ.</zh>
   * <en>Save chart to PNG image.</en>
   *
   * @param filePath String <zh>ͼƬļ· (Ϊnull).</zh>
   * <en>The image file path (null not permitted).</en>
   * @param width int <zh>ͼƬĿ.</zh>
   * <en>The image width.</en>
   * @param height int <zh>ͼƬĸ߶.</zh>
   * <en>The image height.</en>
   *
   * @throws IOException
   */
  public void saveChartAsPNG(String filePath, int width, int height) throws IOException {
    if (filePath == null) {
      throw new IllegalArgumentException("Null 'filePath' argument.") ;
    }

    File file = new File(filePath) ;
    saveChartAsPNG(file, width, height) ;
  }

  /**
   * <zh>chart洢ΪPNGʽͼƬ.</zh>
   * <en>Save chart to PNG image.</en>
   *
   * @param stream OutputStream <zh>ͼƬļ (Ϊnull).</zh>
   * <en>An OutputStream to be written to (null not permitted).</en>
   * @param width int <zh>ͼƬĿ.</zh>
   * <en>The image width.</en>
   * @param height int <zh>ͼƬĸ߶.</zh>
   * <en>The image height.</en>
   *
   * @throws IOException
   */
  public void saveChartAsPNG(OutputStream stream, int width, int height) throws IOException {
    if (stream == null) {
      throw new IllegalArgumentException("Null 'stream' argument.") ;
    }

    BufferedImage bufferImg = createBufferedImage(width, height) ;
    try {
      ImageIO.write(bufferImg, "PNG", stream) ;
    } catch (IOException ex) {
      throw ex ;
    }
  }

  /**
   * <zh>chart洢ΪPNGʽͼƬ.</zh>
   * <en>Save chart to PNG image.</en>
   *
   * @param imageFile File <zh>ͼƬļ (Ϊnull).</zh>
   * <en>The image file (null not permitted).</en>
   * @param width int <zh>ͼƬĿ.</zh>
   * <en>The image width.</en>
   * @param height int <zh>ͼƬĸ߶.</zh>
   * <en>The image height.</en>
   *
   * @throws IOException
   */
  public void saveChartAsPNG(File imageFile, int width, int height) throws IOException {
    if (imageFile == null) {
      throw new IllegalArgumentException("Null 'imageFile' argument.") ;
    }

    BufferedImage bufferImg = createBufferedImage(width, height) ;
    try {
      ImageIO.write(bufferImg, "PNG", imageFile) ;
    } catch (IOException ex) {
      throw ex ;
    }
  }

  /**
   * Tests this chart for equality with another object.
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof IJChart)) {
      return false ;
    }

    IJChart that = (IJChart) obj ;
    if (this.borderVisible != that.borderVisible) {
      return false ;
    }
    if (!EqualsUtil.equals(this.borderStroke, that.borderStroke)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.borderPaint, that.borderPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.padding, that.padding)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.titleFont, that.titleFont)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.title, that.title)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.subtitles, that.subtitles)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.legend, that.legend)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.plot, that.plot)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.backgroundPaint, that.backgroundPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.gradientPaintTransformType, that.gradientPaintTransformType)) {
      return false ;
    }
    if (this.backgroundAlpha != that.backgroundAlpha) {
      return false ;
    }
    if (!EqualsUtil.equals(this.backgroundImage, that.backgroundImage)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.backgroundImageAlignment, that.backgroundImageAlignment)) {
      return false ;
    }
    if (this.backgroundImageAlpha != that.backgroundImageAlpha) {
      return false ;
    }
    if (this.notify != that.notify) {
      return false ;
    }
    if (!EqualsUtil.equals(this.chartShapes, that.chartShapes)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.changeChartShape, that.changeChartShape)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcode.</zh>
   * <en>Returns a hash code for this object.</en>
   *
   * @return <zh>hashcode.</zh>
   * <en>A hash code.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(borderVisible) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.borderStroke) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.borderPaint) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.padding) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.titleFont) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.title) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.subtitles) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.legend) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.plot) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.backgroundPaint) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.gradientPaintTransformType) ;
    result = 29 * result + (int) (100 * backgroundAlpha) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.backgroundImage) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.backgroundImageAlignment) ;
    result = 29 * result + (int) (29 * backgroundImageAlpha) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.notify) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.chartShapes) ;
    result = 29 * result + 29 * HashcodeUtil.hashCode(this.changeChartShape) ;

    return result ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;

    SerialUtil.writeStroke(stream, this.borderStroke) ;
    SerialUtil.writePaint(stream, this.borderPaint) ;
    SerialUtil.writePaint(stream, this.backgroundPaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;

    this.borderStroke = SerialUtil.readStroke(stream) ;
    this.borderPaint = SerialUtil.readPaint(stream) ;
    this.backgroundPaint = SerialUtil.readPaint(stream) ;

    this.progressListeners = new EventListenerList() ;
    this.changeListeners = new EventListenerList() ;

    // register as a listener with sub-components...
    if (this.title != null) {
      if (this.title instanceof AbstractTitle) {
        ((AbstractTitle)this.title).addChangeListener(this) ;
      }
    }

    //sub titles
    for (int i = 0 ; i < this.subtitles.size() ; i++) {
      Object obj = this.subtitles.get(i) ;
      if (obj != null && obj instanceof AbstractTitle) {
        ((AbstractTitle) obj).addChangeListener(this) ;
      }
    }

    //legend titles
    if (this.legend != null) {
      this.legend.addChangeListener(this) ;
    }

    this.plot.addChangeListener(this) ;
  }

  /**
   * <zh>¡.</zh>
   * <en>Clones the object, and takes care of listeners.</en>
   *
   * @return A clone.
   *
   * @throws CloneNotSupportedException if the chart is not cloneable.
   */
  public Object clone() throws CloneNotSupportedException {
    IJChart chart = (IJChart)super.clone() ;

    if (this.title != null) {
      if (chart.title instanceof AbstractTitle) {
        ((AbstractTitle) chart.title).addChangeListener(chart) ;
      }
    }

    for (int i = 0 ; i < chart.getSubtitleCount() ; i++) {
      Object subtitle = chart.subtitles.get(i) ;
      if (subtitle != null && subtitle instanceof AbstractTitle) {
        ((AbstractTitle) subtitle).addChangeListener(chart) ;
      }
    }

    if (this.plot != null) {
      ((AbstractPlot) chart.plot).addChangeListener(chart) ;
    }

    chart.progressListeners = new EventListenerList() ;
    chart.changeListeners = new EventListenerList() ;

    return chart ;
  }

}
