package com.ijchart.xychart.data ;

import java.io.Serializable ;
import java.util.Date ;

import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ͼݶ,ⲿǲ޸ĵ.</zh>
 * <en>Grantt ChartData object,this class is immutable.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class GranttChartData extends AbstractChartData implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -3697322462590448487L ;

  /**
   * <zh>ʼʱ.</zh>
   * <en>The start date/time.</en>
   */
  private Date start ;

  /**
   * <zh>ʱ</zh>
   * <en>The end date/time.</en>
   */
  private Date end ;

  /**
   * <zh>ɰٷֱ.</zh>
   * <en>The percent complete.</en>
   */
  private Double percentComplete ;

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param start long <zh>ʼʱ.</zh>
   * <en>The start date/time in milliseconds.</en>
   * @param end long <zh>ʱ</zh>
   * <en>The end date/time in milliseconds.</en>
   */
  public GranttChartData(long start, long end) {
    this(new Date(start), new Date(end), null) ;
  }

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param start long <zh>ʼʱ.</zh>
   * <en>The start date/time in milliseconds.</en>
   * @param end long <zh>ʱ</zh>
   * <en>The end date/time in milliseconds.</en>
   * @param percentComplete <zh>ɰٷֱ.</zh>
   * <en>The percent complete.</en>
   */
  public GranttChartData(long start, long end, Double percentComplete) {
    this(new Date(start), new Date(end), percentComplete) ;
  }

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param start long <zh>ʼʱ.</zh>
   * <en>The start date/time in milliseconds.</en>
   * @param end long <zh>ʱ</zh>
   * <en>The end date/time in milliseconds.</en>
   */
  public GranttChartData(Date start, Date end) {
    this(start, end, null) ;
  }

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param start long <zh>ʼʱ.</zh>
   * <en>The start date/time in milliseconds.</en>
   * @param end long <zh>ʱ</zh>
   * <en>The end date/time in milliseconds.</en>
   * @param percentComplete <zh>ɰٷֱ.</zh>
   * <en>The percent complete.</en>
   */
  public GranttChartData(Date start, Date end, Double percentComplete) {
    if (start == null) {
      throw new IllegalArgumentException("Null 'start' argument.") ;
    }
    if (end == null) {
      throw new IllegalArgumentException("Null 'end' argument.") ;
    }
    if (start.getTime() > end.getTime()) {
      throw new IllegalArgumentException("Requires start <= end.") ;
    }
    if (percentComplete != null) {
      double complete = percentComplete.doubleValue() ;
      if (!(complete >= 0.0 && complete <= 1.0)) {
        throw new IllegalArgumentException("'percentComplete' must be in [0.0,1.0].") ;
      }
    }

    this.start = start ;
    this.end = end ;
    this.percentComplete = percentComplete ;
  }

  /**
   * <zh>ؿʼʱ.</zh>
   * <en>Returns the start date/time.</en>
   *
   * @return Date <zh>ʼʱ (Ϊnull).</zh>
   * <en>The start date/time (never null).</en>
   */
  public Date getStart() {
    return this.start ;
  }

  /**
   * <zh>ؽʱ.</zh>
   * <en>Returns the end date/time.</en>
   *
   * @return Date <zh>ʱ (Ϊnull).</zh>
   * <en>The end date/time (never null).</en>
   */
  public Date getEnd() {
    return this.end ;
  }

  /**
   * <zh>ɰٷֱ.</zh>
   * <en>Returns the percent complete.</en>
   *
   * @return Double <zh>ɰٷֱ.</zh>
   * <en>The percent complete.</en>
   */
  public Double getPercentComplete() {
    return this.percentComplete ;
  }

  /**
   * <zh>رǩַ.</zh>
   * <en>Returns the label text.</en>
   *
   * @return String <zh>ǩַ (Ϊnull).</zh>
   * <en>The label text (maybe null).</en>
   */
  public String getLabelText() {
    return "(" + formatDate(this.start) + "," + formatDate(this.end) + ")" ;
  }

  /**
   * <zh>ͼȵareaĺ.</zh>
   * <en>Returns the function arguments of hotspot area.</en>
   *
   * @return String <zh> (Ϊnull).</zh>
   * <en>The function arguments (maybe null).</en>
   */
  public String getAreaFunctionArguments() {
    return formatDateForFunction(this.start) + "," + formatDateForFunction(this.end) ;
  }

  /**
   * <zh>ͼȵareaurl.</zh>
   * <en>Returns the url for hotspot area.</en>
   *
   * @return String <zh>url (Ϊnull).</zh>
   * <en>The url fragment (maybe null).</en>
   */
  public String getAreaURLFragment() {
    return "start=" + formatDate(this.start) + "&end=" + formatDate(this.end) ;
  }

  /**
   * <zh>ʵComparableӿڵcompareTo().</zh>
   * <en>Implements the compareTo() method of interface Comparable.</en>
   *
   * @param o Object <zh>ҪȽϵĶ.</zh>
   * <en>The Object to be compared.</en>
   * @return int <zh>ֵΪʾС,0ʾ,ʾ.</zh>
   * <en>A negative integer, zero, or a positive integer as this object	is less than, equal to, or
   * greater than the specified object.</en>
   *
   * @throws ClassCastException if the specified object's type prevents it
   *         from being compared to this Object.
   */
  public int compareTo(Object o) {
    if (o == null) {
      return -1 ;
    } else {
      if (!(o instanceof GranttChartData)) {
        throw new ClassCastException("Must be 'GranttChartData'.") ;
      }

      GranttChartData other = (GranttChartData) o ;
      if (this.start == null) {
        return -1 ;
      }

      return this.start.compareTo(other.start) ;
    }
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof GranttChartData)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    GranttChartData that = (GranttChartData) obj ;
    if (EqualsUtil.equals(this.start, that.start)) {
      return false ;
    }
    if (EqualsUtil.equals(this.end, that.end)) {
      return false ;
    }
    if (EqualsUtil.equals(this.percentComplete, that.percentComplete)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() ;
    result = 37 * result + HashcodeUtil.hashCode(this.start) ;
    result = 37 * result + HashcodeUtil.hashCode(this.end) ;
    result = 37 * result + HashcodeUtil.hashCode(this.percentComplete) ;

    return (int) result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[start=" + this.start
      + ",end=" + this.end
      + ",percentComplete=" + percentComplete
      + "]" ;
  }

}
