package com.ijchart.xychart.data ;

import java.io.Serializable ;

import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>(x,y,z)ֵͼݶ,ⲿǲ޸ĵ.</zh>
 * <en>The (x,y,z) Number ChartData object,this class is immutable.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class XYZNumberChartData extends XYNumberChartData implements IXYChartData,
  Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 6909017890845971993L ;

  /**
   * <zh>zֵ.</zh>
   * <en>The z number data.</en>
   */
  private Number z ;

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param x double <zh>xֵ.</zh>
   * <en>The x value.</en>
   * @param y double <zh>yֵ.</zh>
   * <en>The y value.</en>
   * @param z double <zh>zֵ.</zh>
   * <en>The z value.</en>
   */
  public XYZNumberChartData(double x, double y, double z) {
    this(new Double(x), new Double(y), new Double(z)) ;
  }

  /**
   * <zh>.</zh>
   * <en>Constructor.</en>
   *
   * @param x Number <zh>xֵ (Ϊnull).</zh>
   * <en>The x Number object (null not permitted).</en>
   * @param y Number <zh>yֵ (Ϊnull).</zh>
   * <en>The y Number object (null not permitted).</en>
   * @param z Number <zh>zֵ (Ϊnull).</zh>
   * <en>The z Number object (null not permitted).</en>
   */
  public XYZNumberChartData(Number x, Number y, Number z) {
    super(x, y) ;
    if (z == null) {
      throw new IllegalArgumentException("Null 'z' argument.") ;
    }

    this.z = z ;
  }

  /**
   * <zh>õdouble͵zֵ.</zh>
   * <en>Returns the double z value.</en>
   *
   * @return double <zh>double͵ֵ.</zh>
   * <en>The double value.</en>
   */
  public double getZ() {
    return this.z.doubleValue() ;
  }

  /**
   * <zh>zֶ.</zh>
   * <en>Returns the z Number data.</en>
   *
   * @return Number <zh>Ϊnull.</zh>
   * <en>Never null.</en>
   */
  public Number getZData() {
    return this.z ;
  }

  /**
   * <zh>رǩַ.</zh>
   * <en>Returns the label text.</en>
   *
   * @return String <zh>ǩַ (Ϊnull).</zh>
   * <en>The label text (never null).</en>
   */
  public String getLabelText() {
    return "(" + formatNumber(getX()) + "," + formatNumber(getY())
      + "," + formatNumber(getZ()) + ")" ;
  }

  /**
   * <zh>ͼȵareaĺ.</zh>
   * <en>Returns the function arguments of hotspot area.</en>
   *
   * @return String <zh> (Ϊnull).</zh>
   * <en>The function arguments (maybe null).</en>
   */
  public String getAreaFunctionArguments() {
    return formatNumberForFunction(getX()) + ","
      + formatNumberForFunction(getY()) + ","
      + formatNumberForFunction(getZ()) ;
  }

  /**
   * <zh>ͼȵareaurl.</zh>
   * <en>Returns the url for hotspot area.</en>
   *
   * @return String <zh>url (Ϊnull).</zh>
   * <en>The url fragment (maybe null).</en>
   */
  public String getAreaURLFragment() {
    return "x=" + formatNumber(getX()) + "&y=" + formatNumber(getY())
      + "&z=" + formatNumber(getZ()) ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof XYZNumberChartData)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    XYZNumberChartData that = (XYZNumberChartData) obj ;
    if (EqualsUtil.equals(this.z, that.z)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = 37 * result + HashcodeUtil.hashCode(this.z) ;

    return result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[x=" + getX() + ",y=" + getY() + ",z=" + getZ() + "]" ;
  }

}
