package com.ijchart.xychart.dataset ;

import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Collections ;
import java.util.List ;

import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.dataset.data.KeyValuePair ;
import com.ijchart.xychart.util.EqualsUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ӿ{@link IKeyValueDataset}ȱʡʵ.ڲݴ洢ʹList,Ա˹.</zh>
 * <en>The default implemention of {@link IKeyValueDataset} interface.The use of internal
 *  is List,so this dataset can not be too large.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class DefaultKeyValueDataset extends AbstractDataset implements IKeyValueDataset,
  Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -5608149959318627177L ;

  /**
   * <zh>洢õList.</zh>
   * <en>The list for storage data.</en>
   */
  private List dataList ;

  /**
   * <zh>ûвĹ.</zh>
   * <en>No argument constructor.</en>
   */
  public DefaultKeyValueDataset() {
    dataList = new ArrayList() ;
  }

  /**
   * <zh>ָϳʼСĹ.</zh>
   * <en>Constructs an empty dataset with the specified initial capacity.</en>
   * @param initialCapacity int
   */
  public DefaultKeyValueDataset(int initialCapacity) {
    dataList = new ArrayList(initialCapacity) ;
  }

  /**
   * <zh>мֵ</zh>
   * <en>Returns all the keys.</en>
   *
   * @return <zh>мֵ</zh>
   * <en>All the keys list.</en>
   */
  public List getAllKeys() {
    List keyList = new ArrayList() ;

    for (int i = 0 ; i < dataList.size() ; i++) {
      KeyValuePair pair = (KeyValuePair) dataList.get(i) ;
      keyList.add(pair.getKey()) ;
    }

    return keyList ;
  }

  /**
   * <zh>ݼֵӦֵ.ûҵ򷵻null.</zh>
   * <en>Returns the data associate with this key.</en>
   *
   * @param key Object
   * @return IChartData
   */
  public IChartData getDataByKey(Object key) {
    for (int i = 0 ; i < dataList.size() ; i++) {
      KeyValuePair pair = (KeyValuePair) dataList.get(i) ;
      if (pair.getKey().equals(key)) {
        return (IChartData) pair.getValue() ;
      }
    }

    return null ;
  }

  /**
   * <zh>ȡֵ,Χ׳쳣.</zh>
   * <en>Get chartdata by index,if index out of bound a Exception will be throw.</en>
   * @param index int <zh>ֵ.</zh>
   *  <en>The data index</en>
   * @return IChartData
   */
  public IChartData getDataByIndex(int index) {
    if (!(index >= 0 && index < this.dataList.size())) {
      throw new IllegalArgumentException("Argument 'index' must be between 0 and " + this.dataList.size()) ;
    }

    KeyValuePair pair = (KeyValuePair)this.dataList.get(index) ;

    return (IChartData) pair.getValue() ;
  }

  /**
   * <zh>Ÿıֵ,Χ׳쳣.</zh>
   * <en>Sets chartdata by index,if index out of bound a Exception will be throw.</en>
   * @param index int <zh>ֵ.</zh>
   *  <en>The data index</en>
   * @param data IChartData <zh>ֵ.</zh>
   * <en>The {@link IChartData} object.</en>
   */
  public void setDataByIndex(int index, IChartData data) {
    if (!(index >= 0 && index < this.dataList.size())) {
      throw new IllegalArgumentException("Argument 'index' must be between 0 and " + this.dataList.size()) ;
    }

    KeyValuePair pair = (KeyValuePair)this.dataList.get(index) ;
    pair.setValue(data) ;
    fireDatasetChanged() ;
  }

  /**
   * <zh>ݼֵıֵ.</zh>
   * <en>Sets chartdata by key.</en>
   * @param key Object <zh>ֵ (Ϊnull).</zh>
   *  <en>The key object (null not permitted).</en>
   * @param data IChartData <zh>ֵ.</zh>
   * <en>The {@link IChartData} object.</en>
   * @return boolean <zh>Ƿıֵ.</zh>
   * <en>If of not changed data.</en>
   */
  public boolean setDataByKey(Object key, IChartData data) {
    if (key == null) {
      throw new IllegalArgumentException("Null 'data' argument.") ;
    }

    int index = getKeyIndex(key) ;
    if (index < 0) {
      return false ;
    }

    setDataByIndex(index, data) ;
    return true ;
  }

  /**
   * <zh>ݼ.˼ǲɸı.</zh>
   * <en> Returns a list (possibly empty) of the values.The returned list
   *  should be unmodifiable.</en>
   * @return List <zh>ݼ.</zh>
   * <en>The list of values.</en>
   */
  public List getData() {
    return Collections.unmodifiableList(this.dataList) ;
  }

  /**
   * <zh>ؼֵĸ.</zh>
   * <en>Returns the keys count.</en>
   *
   * @return int
   */
  public int getDataSize() {
    return dataList.size() ;
  }

  /**
   * <zh>ؼֵмֵе.</zh>
   * <en>Returns the index for a given key.</en>
   *
   * @param key Object <zh>ֵ</zh>
   * <en>The key object.</en>
   * @return int <zh>ֵмֵе,ûҵ򷵻-1.</zh>
   * <en>The column index, or -1 if the key is unrecognised.</en>
   */
  public int getKeyIndex(Object key) {
    for (int i = 0 ; i < dataList.size() ; i++) {
      KeyValuePair pair = (KeyValuePair) dataList.get(i) ;
      if (pair.getKey().equals(key)) {
        return i ;
      }

    }

    return -1 ;
  }

  /**
   * <zh>һֵݶԵ.</zh>
   * <en>Add a value associate a key to the storage list.</en>
   *
   * @param key Object <zh>ֵ (Ϊnull).</zh>
   * <en>The key object (null not permitted).</en>
   * @param value Object <zh>ֵ (Ϊnull).</zh>
   * <en>The value object (null permitted).</en>
   */
  public void addChartData(Object key, IChartData value) {
    this.dataList.add(new KeyValuePair(key, value)) ;
    fireDatasetChanged() ;
  }

  /**
   * <zh>ݼֵɾݼ.</zh>
   * <en>Remove data by key object.</en>
   * @param key Object <zh>ֵ (Ϊnull)</zh>
   * <en>The key object,(null not permitted).</en>
   * @return int <zh>ҵļֵڼеλ,ûҵ-1.</zh>
   * <en>The key index in the collection,if not found then return -1.</en>
   */
  public int removeChartDataByKey(Object key) {
    if (key == null) {
      throw new IllegalArgumentException("Null 'key' argument.") ;
    }

    for (int i = 0 ; i < dataList.size() ; i++) {
      KeyValuePair pair = (KeyValuePair) dataList.get(i) ;
      if (pair.getKey().equals(key)) {
        this.dataList.remove(i) ;
        fireDatasetChanged() ;

        return i ;
      }
    }

    return -1 ;
  }

  /**
   * <zh>ɾݼ.</zh>
   * <en>Remove data by index.</en>
   * @param index int <zh>.</zh>
   * <en>The index</en>
   */
  public void removeChartDataByIndex(int index) {
    if (!(index >= 0 && index < this.dataList.size())) {
      throw new IllegalArgumentException("Argument 'index' must be between 0 and " + this.dataList.size()) ;
    }

    this.dataList.remove(index) ;
    fireDatasetChanged() ;
  }

  /**
   * <zh>ݼ.</zh>
   * <en>Clear all data in the dataset.</en>
   */
  public void removeAllChartData() {
    this.dataList.clear() ;
    fireDatasetChanged() ;
  }

  /**
   * <zh>жеֵǷc.</zh>
   * <en>Test whether or not all data is Class c in this dataset.</en>
   *
   * @param c Class <zh>Ҫжϵclass (Ϊnull).</zh>
   * <en>A class (null not permitted).</en>
   * @return boolean <zh>ֵctrue,򷵻false.</zh>
   * <en>Returns true if all data equals Class c in this dataset,false otherwise.</en>
   */
  public boolean isThisDataType(Class c) {
    if (c == null) {
      throw new IllegalArgumentException("Null 'c' argument.") ;
    }

    for (int i = 0 ; i < dataList.size() ; i++) {
      KeyValuePair pair = (KeyValuePair) dataList.get(i) ;
      if (pair != null && !c.isInstance(pair.getValue())) {
        return false ;
      }
    }

    return true ;
  }

  /**
   * <zh>¡.</zh>
   * <en>Clone this object.</en>
   * @return Object
   * @throws java.lang.CloneNotSupportedException
   */
  public Object clone() throws CloneNotSupportedException {
    DefaultKeyValueDataset clone = (DefaultKeyValueDataset)super.clone() ;
    clone.dataList = (ArrayList) ((ArrayList)this.dataList).clone() ;

    return clone ;
  }

  /**
   * <zh>equals.</zh>
   * <en>Override equals method.</en>
   * @param obj Object
   * @return boolean
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof DefaultKeyValueDataset)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    DefaultKeyValueDataset other = (DefaultKeyValueDataset) obj ;
    if (!EqualsUtil.equals(this.dataList, other.dataList)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hascodeֵ.</zh>
   * <en>Get hascode.</en>
   * @return int
   */
  public int hashCode() {
    int result = super.hashCode() ;

    for (int i = 0 ; i < this.dataList.size() ; i++) {
      result = 29 * result + this.dataList.get(i).hashCode() ;
    }

    return result ;
  }
}
