package com.ijchart.xychart.dataset ;

import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Arrays ;
import java.util.Collection ;
import java.util.Collections ;
import java.util.Iterator ;
import java.util.List ;

import com.ijchart.xychart.constant.SortType ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.util.EqualsUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ʵֽӿIMatrixDataset.</zh>
 * <en>The default implement of interface IMatrixDataset.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class DefaultMatrixDataset extends AbstractDataset implements IMatrixDataset, Serializable,
  Cloneable {

  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 8482292279759724354L ;

  /**
   * <zh>洢RowKeyArrayList.</zh>
   * <en>The ArrayList for storage RowKeys.</en>
   */
  private List rowKeyList = null ;

  /**
   * <zh>洢ColKeyArrayList.</zh>
   * <en>The ArrayList for storage ColKeys.</en>
   */
  private List colKeyList = null ;

  /**
   * <zh>洢ݵArrayList.</zh>
   * <en>The ArrayList for storage data.</en>
   */
  private List dataList = null ;

  /**
   * <zh>ȱʡ.</zh>
   * <en>No argument constructor.</en>
   */
  public DefaultMatrixDataset() {
    this.rowKeyList = new ArrayList() ;
    this.colKeyList = new ArrayList() ;
    this.dataList = new ArrayList() ;
  }

  /**
   * <zh>һRowKey.</zh>
   * <en>Add a RowKey.</en>
   *
   * @param rowKey Object <zh>RowKey (Ϊnull)</zh>
   * <en>RowKey (null permitted).</en>
   */
  public void addRowKey(Object rowKey) {
    this.rowKeyList.add(rowKey) ;
    fireDatasetChanged() ;
  }

  /**
   * <zh>ıRowKeyֵ.</zh>
   * <en>Sets the RowKey value by row index.</en>
   *
   * @param index int <zh>.</zh>
   * <en>The row index.</en>
   * @param rowKey Object <zh>µRowKeyֵ (Ϊnull)</zh>
   * <en>New RowKey value (null permitted).</en>
   */
  public void setRowKey(int index, Object rowKey) {
    checkRowKeyIndex(index) ;

    this.rowKeyList.set(index, rowKey) ;
    fireDatasetChanged() ;
  }

  /**
   * <zh>ɾRowKey.</zh>
   * <en>Delete RowKey by index.</en>
   *
   * @param index int <zh>.</zh>
   * <en>The row index.</en>
   *
   * @return Object <zh>ɾrowKey.</zh>
   * <en>The deleted ColKey object.</en>
   */
  public Object deleteRowKey(int index) {
    checkRowKeyIndex(index) ;

    Object rowKey = this.rowKeyList.remove(index) ;
    fireDatasetChanged() ;

    return rowKey ;
  }

  /**
   * <zh>ɾRowKey.</zh>
   * <en>Delete this RowKey.</en>
   *
   * @param rowKey Object <zh>ҪɾRowKey (Ϊnull).</zh>
   *<en>The RowKey (null permitted).</en>
   *
   * @return boolean <zh>Ƿɾһ.</zh>
   * <en>Returns true if this list contained the specified element.</en>
   */
  public boolean deleteRowKey(Object rowKey) {
    boolean b = this.rowKeyList.remove(rowKey) ;
    fireDatasetChanged() ;

    return b ;
  }

  /**
   * <zh>ȡRolKey.</zh>
   * <en>Get RolKey by index.</en>
   *
   * @param index int <zh>ֵ.</zh>
   * <en>The row index.</en>
   *
   * @return Object <zh>ֵӦRolKey(Ϊnull).</zh>
   * <en>The RolKey for the index(maybe null).</en>
   */
  public Object getRowKey(int index) {
    try {
      return rowKeyList.get(index) ;
    } catch (Exception ex) {
      return null ;
    }
  }

  /**
   * <zh>һColKey.</zh>
   * <en>Add a ColKey.</en>
   *
   * @param colKey Object <zh>ColKey (Ϊnull)</zh>
   * <en>ColKey (null permitted).</en>
   */
  public void addColKey(Object colKey) {
    this.colKeyList.add(colKey) ;
    fireDatasetChanged() ;
  }

  /**
   * <zh>ıColKeyֵ.</zh>
   * <en>Sets the ColKey value by index.</en>
   *
   * @param index int <zh>.</zh>
   * <en>The index.</en>
   * @param colKey Object <zh>µColKeyֵ (Ϊnull)</zh>
   * <en>New ColKey value (null permitted).</en>
   */
  public void setColKey(int index, Object colKey) {
    checkColKeyIndex(index) ;

    this.colKeyList.set(index, colKey) ;
    fireDatasetChanged() ;
  }

  /**
   * <zh>ɾColKey.</zh>
   * <en>Delete ColKey by index.</en>
   *
   * @param index int <zh>.</zh>
   * <en>The index.</en>
   *
   * @return Object <zh>ɾColKey.</zh>
   * <en>The deleted ColKey object.</en>
   */
  public Object deleteColKey(int index) {
    checkColKeyIndex(index) ;

    Object colKey = this.colKeyList.remove(index) ;
    fireDatasetChanged() ;

    return colKey ;
  }

  /**
   * <zh>ɾColKey.</zh>
   * <en>Delete this ColKey.</en>
   *
   * @param colKey Object <zh>ҪɾColKey (Ϊnull).</zh>
   * <en>The ColKey (null permitted).</en>
   *
   * @return boolean <zh>Ƿɾһ.</zh>
   * <en>Returns true if this list contained the specified element.</en>
   */
  public boolean deleteColKey(Object colKey) {
    boolean b = this.colKeyList.remove(colKey) ;
    fireDatasetChanged() ;

    return b ;
  }

  /**
   * <zh>ȡColKey.</zh>
   * <en>Get ColKey by index.</en>
   *
   * @param index int <zh>ֵ.</zh>
   * <en>The index.</en>
   *
   * @return Object <zh>ֵӦColKey(Ϊnull).</zh>
   * <en>The ColKey for the index(maybe null).</en>
   */
  public Object getColKey(int index) {
    try {
      return colKeyList.get(index) ;
    } catch (Exception ex) {
      return null ;
    }
  }

  /**
   * <zh>һ</zh>
   * <en>Add one row to data list.</en>
   */
  public void addRowData() {
    addRowData(1) ;
  }

  /**
   * <zh>һӼ</zh>
   * <en>Add multiple rows to data list.</en>
   *
   * @param count int <zh>.</zh>
   * <en>.</en>
   */
  public void addRowData(int count) {
    for (int i = 0 ; i < count ; i++) {
      this.dataList.add(new ArrayList()) ;
    }

    if (count > 0) {
      fireDatasetChanged() ;
    }
  }

  /**
   * <zh>ɾһ.</zh>
   * <en>Delete the datas of a dataset row.</en>
   *
   * @param rowIndex int <zh>.</zh>
   * <en>The row index.</en>
   */
  public void deleteRow(int rowIndex) {
    checkDataIndex(rowIndex) ;

    this.dataList.remove(rowIndex) ;
    fireDatasetChanged() ;
  }

  /**
   * <zh>ĳһ{@link IChartData}ֵ.</zh>
   * <en>Add a {@link IChartData} to dataset.</en>
   *
   * @param rowIndex int <zh>.</zh>
   * <en>The row index.</en>
   * @param chartData IChartData <zh>{@link IChartData}(Ϊnull).</zh>
   * <en>The {@link IChartData} object (null permitted).</en>
   */
  public void addData(int rowIndex, IChartData chartData) {
    checkDataIndex(rowIndex) ;

    List rowData = (List)this.dataList.get(rowIndex) ;
    rowData.add(chartData) ;

    fireDatasetChanged() ;
  }

  /**
   * <zh>ĳӶ{@link IChartData}ֵ.</zh>
   * <en>Add {@link IChartData} datas to dataset.</en>
   *
   * @param rowIndex int <zh>.</zh>
   * <en>The row index.</en>
   * @param chartDatas Collection <zh>{@link IChartData}(Ϊnull).</zh>
   * <en>The {@link IChartData} object (null permitted).</en>
   */
  public void addDatas(int rowIndex, Collection chartDatas) {
    checkDataIndex(rowIndex) ;

    Iterator iter = chartDatas.iterator() ;
    while (iter.hasNext()) {
      Object obj = iter.next() ;

      if (obj != null && !(obj instanceof IChartData)) {
        throw new IllegalArgumentException("Chartdata must be 'IChartData'.") ;
      }
    }

    List rowData = (List)this.dataList.get(rowIndex) ;
    rowData.addAll(chartDatas) ;

    fireDatasetChanged() ;
  }

  /**
   * <zh>ĳӶ{@link IChartData}ֵ.</zh>
   * <en>Add {@link IChartData} datas to dataset.</en>
   *
   * @param rowIndex int <zh>.</zh>
   * <en>The row index.</en>
   * @param chartDatas IChartData <zh>{@link IChartData}(Ϊnull).</zh>
   * <en>The {@link IChartData} object (null permitted).</en>
   */
  public void addDatas(int rowIndex, IChartData[] chartDatas) {
    checkDataIndex(rowIndex) ;

    List rowData = (List)this.dataList.get(rowIndex) ;
    rowData.addAll(Arrays.asList(chartDatas)) ;

    fireDatasetChanged() ;
  }

  /**
   * <zh>rowIndexcolIndexеֵ.</zh>
   * <en>Change the dataset value by row index and column index.</en>
   *
   * @param rowIndex int <zh>.</zh>
   * <en>The row index.</en>
   * @param colIndex int <zh>.</zh>
   * <en>The column index.</en>
   * @param chartData IChartData <zh>{@link IChartData}(Ϊnull).</zh>
   * <en>The {@link IChartData} object (null permitted).</en>
   */
  public void setData(int rowIndex, int colIndex, IChartData chartData) {
    checkDataIndex(rowIndex, colIndex) ;

    List rowData = (List)this.dataList.get(rowIndex) ;
    if (rowData == null) {
      rowData = new ArrayList() ;
      this.dataList.set(rowIndex, rowData) ;
    }

    rowData.set(colIndex, chartData) ;
    fireDatasetChanged() ;
  }

  /**
   * <zh>ݼ</zh>
   * <en>Returns the row count of this dataset.</en>
   *
   * @return int <zh>ݼ,ûݷ0.</zh>
   * <en>The row count of this dataset,no data return 0.</en>
   */
  public int getRowCount() {
    return this.dataList.size() ;
  }

  /**
   * <zh>ݼ.</zh>
   * <en>The column count of this dataset.</en>
   *
   * @return int <zh>ݼ.</zh>
   * <en>The column count.</en>
   */
  public int getColumnCount() {
    int result = 0 ;
    for (int i = 0 ; i < this.dataList.size() ; i++) {
      List list = (List)this.dataList.get(i) ;
      if (list != null) {
        result = Math.max(result, list.size()) ;
      }
    }

    return result ;
  }

  /**
   * <zh>ݼ.</zh>
   * <en>The column count of this dataset.</en>
   *
   * @param rowIdx int <zh>.</zh>
   * <en>The row index.</en>
   * @return int <zh>ݼ.</zh>
   * <en>The column count.</en>
   */
  public int getColumnCount(int rowIdx) {
    int dataSize = this.dataList.size() ;
    if (rowIdx >= dataSize) {
      return getColumnCount() ;
    }

    List list = (List)this.dataList.get(rowIdx) ;
    return (list == null) ? 0 : list.size() ;
  }

  /**
   * <zh>õάеĳֵ,rowIndexcolIndexΧ׳쳣.</zh>
   * <en>Get the data of matrix by rowIndex and colIndex,an Exception will be throw when index out
   * of bound.</en>
   *
   * @param rowIndex int <zh>ֵ,һ0,ڶ1,.</zh>
   * <en>The row index.The first is at index 0,the next is at 1, and so on.</en>
   * @param colIndex int <zh>ֵ,һ0,ڶ1,.</zh>
   * <en>The column index.The first is at index 0,the next is at 1, and so on.</en>
   *
   * @return IChartData <zh>иõֵ(Ϊnull).</zh>
   * <en>The data of matrix (maybe null).</en>
   */
  public IChartData getData(int rowIndex, int colIndex) {
    checkDataIndex(rowIndex) ;

    List rowData = (List)this.dataList.get(rowIndex) ;
    try {
      return (IChartData) rowData.get(colIndex) ;
    } catch (Exception ex) {
      return null ;
    }
  }

  /**
   * <zh>õĳһе.</zh>
   * <en>Returns all datas of the specified row index.</en>
   *
   * @param rowIndex int <zh>ֵ,һ0,ڶ1,.</zh>
   * <en>The row index.The first is at index 0,the next is at 1, and so on.</en>
   *
   * @return List <zh>ݼ,˼ⲿ޸.</zh>
   * <en>The datas and unmodifiable.</en>
   */
  public List getRowDatas(int rowIndex) {
    checkDataIndex(rowIndex) ;

    return Collections.unmodifiableList((List)this.dataList.get(rowIndex)) ;
  }

  /**
   * <zh>ݿrowKey,ĳûrowKey,Ϊnull.</zh>
   * <en>Returns the rowKeys of this dataset,if a row line data is not rowKey, is null.</en>
   *
   * @return List <zh>ûrowKeyؿList.</zh>
   * <en>A list (never null).</en>
   */
  public List getDatasetRowKeys() {
    List result = new ArrayList() ;

    int rowCount = getRowCount() ;
    for (int i = 0 ; i < rowCount ; i++) {
      result.add(getRowKey(i)) ;
    }

    return result ;
  }

  /**
   * <zh>жеֵǷc.</zh>
   * <en>Test whether or not all data is Class c in this dataset.</en>
   *
   * @param c Class <zh>Ҫжϵclass (Ϊnull).</zh>
   * <en>A class (null not permitted).</en>
   *
   * @return boolean <zh>ֵctrue,򷵻false.</zh>
   * <en>Returns true if all data equals Class c in this dataset,false otherwise.</en>
   */
  public boolean isThisDataType(Class c) {
    if (c == null) {
      throw new IllegalArgumentException("Null 'c' argument.") ;
    }

    for (int i = 0 ; i < dataList.size() ; i++) {
      List list = (List) dataList.get(i) ;
      for (int j = 0 ; j < list.size() ; j++) {
        Object obj = list.get(j) ;
        if (obj != null && !c.isInstance(obj)) {
          return false ;
        }
      }
    }

    return true ;
  }

  /**
   * <zh>ݼǷΪ.</zh>
   * <en>Returns whether or not the dataset is empty.</en>
   *
   * @return boolean <zh>յݼtrue,Ϊfalse.</zh>
   * <en>Returns true if the axis dataset is empty,otherwise false.</en>
   */
  public boolean isEmpty() {
    if (getRowCount() != 0 || getColumnCount() != 0) {
      return false ;
    }

    if (this.dataList.size() == 0) {
      return true ;
    }

    for (int i = 0 ; i < this.dataList.size() ; i++) {
      List rowData = (List)this.dataList.get(i) ;
      for (int j = 0 ; j < rowData.size() ; j++) {
        if (rowData.get(j) != null) {
          return false ;
        }
      }
    }

    return true ;
  }

  /**
   * <zh>ݼеһݽ,ݼеݱʵ{@link java.lang.Comparable}ӿ.</zh>
   * <en>Sort the datas of a dataset row,the data must implements
   *  {@link java.lang.Comparable} interface.</en>
   *
   * @param rowIndex int <zh>ֵ,һ0,ڶ1,.</zh>
   * <en>The row index.The first is at index 0,the next is at 1, and so on.</en>
   * @param sortType SortType <zh> (Ϊnull).</zh>
   * <en>The sort type,ascending or descending.</en>
   */
  public void sortRowData(int rowIndex, SortType sortType) {
    if (sortType == null) {
      throw new IllegalArgumentException("Null 'sortType' argument.") ;
    }

    int datasetSize = this.dataList.size() ;
    if (rowIndex < datasetSize) {
      List datas = (List)this.dataList.get(rowIndex) ;
      int datasSize = datas.size() ;
      List notNullDatas = new ArrayList(datasSize) ;

      int nullObjCount = 0 ;
      for (int i = 0 ; i < datasSize ; i++) {
        Object obj = datas.get(i) ;
        if (obj == null) {
          nullObjCount++ ;
        } else {
          notNullDatas.add(obj) ;
        }
      }

      Collections.sort(notNullDatas) ;
      datas.clear() ;

      if (sortType.equals(SortType.ASCENDING)) {
        for (int i = 0 ; i < nullObjCount ; i++) {
          datas.add(null) ;
        }
        datas.addAll(notNullDatas) ;
      } else {
        Collections.reverse(notNullDatas) ;
        datas.addAll(notNullDatas) ;
        for (int i = 0 ; i < nullObjCount ; i++) {
          datas.add(null) ;
        }
      }

    }
  }

  /**
   * <zh>ݿcolKey,ĳûcolKey,Ϊnull.</zh>
   * <en>Returns the colKeys of this dataset,if a column line data is not colKey, is null.</en>
   *
   * @return List <zh>ûcolKeyؿList.</zh>
   * <en>A list (never null).</en>
   */
  protected List getDatasetColKeys() {
    List result = new ArrayList() ;

    int colCount = getColumnCount() ;
    for (int i = 0 ; i < colCount ; i++) {
      result.add(getColKey(i)) ;
    }

    return result ;
  }

  /**
   * <zh>ǷԽ.</zh>
   * <en>Check whether of not out of bound.</en>
   *
   * @param rowIndex int <zh>ֵ.</zh>
   * <en>The row index.</en>
   */
  private void checkDataIndex(int rowIndex) {
    if (rowIndex < 0) {
      throw new IllegalArgumentException("rowIndex must be great than 0.") ;
    }
    if (rowIndex > this.dataList.size() - 1) {
      throw new IndexOutOfBoundsException("Row index:" + rowIndex + ",Size:" + this.dataList.size()) ;
    }
  }

  /**
   * <zh>кǷԽ.</zh>
   * <en>Check whether of not out of bound.</en>
   *
   * @param rowIndex int
   * @param colIndex int
   */
  private void checkDataIndex(int rowIndex, int colIndex) {
    checkDataIndex(rowIndex) ;

    if (colIndex < 0) {
      throw new IllegalArgumentException("colIndex must be great than 0.") ;
    }

    List rowData = (List)this.dataList.get(rowIndex) ;

    if (colIndex > rowData.size() - 1) {
      throw new IndexOutOfBoundsException("Column index:" + colIndex + ",size:" + rowData.size()) ;
    }
  }

  /**
   * <zh>Ƿ񳬳Χ.</zh>
   * <en>Check if the given row index is in range.</en>
   * @param index int <zh>.</zh>
   * <en>The row index.</en>
   */
  private void checkRowKeyIndex(int index) {
    if (index < 0) {
      throw new IllegalArgumentException("index must be great than 0.") ;
    }
    if (index > this.rowKeyList.size() - 1) {
      throw new IndexOutOfBoundsException("Index:" + index + ",Size:" + this.rowKeyList.size()) ;
    }
  }

  /**
   * <zh>Ƿ񳬳Χ.</zh>
   * <en>Check if the given column index is in range.</en>
   * @param index int
   */
  private void checkColKeyIndex(int index) {
    if (index < 0) {
      throw new IllegalArgumentException("index must be great than 0.") ;
    }
    if (index > this.colKeyList.size() - 1) {
      throw new IndexOutOfBoundsException("Index:" + index + ",Size:" + this.colKeyList.size()) ;
    }
  }

  /**
   * <zh>hascodeֵ.</zh>
   * <en>Get hascode.</en>
   * @return int
   */
  public int hashCode() {
    int result = super.hashCode() ;

    for (int i = 0 ; i < this.rowKeyList.size() ; i++) {
      if (rowKeyList.get(i) != null) {
        result = 29 * result + rowKeyList.get(i).hashCode() ;
      }
    }

    for (int i = 0 ; i < this.colKeyList.size() ; i++) {
      if (colKeyList.get(i) != null) {
        result = 29 * result + colKeyList.get(i).hashCode() ;
      }
    }

    //ÿݵhashcode
//    for (int i = 0 ; i < this.dataList.size() ; i++) {
//      List rowList = (List)this.dataList.get(i) ;
//      if (rowList.get(i) != null) {
//        result = 29 * result + rowList.get(i).hashCode() ;
//      }
//    }

    return result ;
  }

  /**
   * <zh>equals.</zh>
   * <en>Override equals method.</en>
   * @param obj Object
   * @return boolean
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof DefaultMatrixDataset)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    DefaultMatrixDataset other = (DefaultMatrixDataset) obj ;
    if (!EqualsUtil.equals(this.rowKeyList, other.rowKeyList)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.colKeyList, other.colKeyList)) {
      return false ;
    }

    if (this.dataList.size() != other.dataList.size()) {
      return false ;
    }

    for (int i = 0 ; i < dataList.size() ; i++) {
      if (!EqualsUtil.equals((List)this.dataList.get(i), (List) other.dataList.get(i))) {
        return false ;
      }
    }

    return true ;
  }

}
