package com.ijchart.xychart.event ;

import java.awt.event.MouseEvent ;
import java.io.Serializable ;
import java.util.EventObject ;

import com.ijchart.xychart.client.IJChart ;
import com.ijchart.xychart.shape.IChartShape ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>chart¼.</zh>
 * <en>A mouse event for a chart that is displayed in
 *  a {@link com.ijchart.xychart.client.ChartPanel}.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class ChartMouseEvent extends EventObject implements Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 6773809220715647191L ;

  /**
   * <zh>chart.</zh>
   * <en>The chart object.</en>
   */
  private IJChart chart ;

  /**
   * <zh>¼.</zh>
   * <en>The mouse event.</en>
   */
  private MouseEvent trigger ;

  /**
   * <zh>chartͼζ.</zh>
   * <en>The chart shape.</en>
   */
  private IChartShape shape ;

  /**
   * Constructs a new event.
   *
   * @param chart  the source chart (null not permitted).
   * @param trigger  the mouse event that triggered this event
   *                 (null not permitted).
   * @param shape  the chart shape (if any) under the mouse point
   *                (null permitted).
   */
  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param chart IJChart <zh>chart.</zh>
   * <en>The chart object.</en>
   * @param trigger MouseEvent <zh>¼.</zh>
   * <en>The mouse event.</en>
   * @param shape IChartShape <zh>chartͼζ.</zh>
   * <en>The chart shape.</en>
   */
  public ChartMouseEvent(IJChart chart, MouseEvent trigger, IChartShape shape) {
    super(chart) ;

    this.chart = chart ;
    this.trigger = trigger ;
    this.shape = shape ;
  }

  /**
   * <zh>chart.</zh>
   * <en>Returns the chart object.</en>
   *
   * @return IJChart
   */
  public IJChart getChart() {
    return this.chart ;
  }

  /**
   * <zh>¼.</zh>
   * <en>Returns the mouse event.</en>
   *
   * @return MouseEvent <zh>¼.</zh>
   * <en>The mouse event.</en>
   */
  public MouseEvent getTrigger() {
    return this.trigger ;
  }

  /**
   * <zh>chartͼζ.</zh>
   * <en>Returns the chart shape.</en>
   *
   * @return IChartShape <zh>chartͼζ.</zh>
   * <en>The chart shape.</en>
   */
  public IChartShape getShape() {
    return this.shape ;
  }

}
