package com.ijchart.xychart.render.category ;

import java.awt.geom.GeneralPath ;
import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.axis.category.CategoryAxis ;
import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.data.NumberChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.shape.GeneralPathChartShape ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.ui.IEffect3D ;
import com.ijchart.xychart.ui.RectangleEdge ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>3D͵ͼ.</zh>
 * <en>Draw 3D line shapes.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class Line3DRender extends AbstractCategoryRender implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -6355303315488370760L ;

  /**
   * <zh>3DʾЧ.</zh>
   * <en>The IEffect3D object.</en>
   */
  private IEffect3D effect ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param effect3D IEffect3D <zh>3DʾЧ (Ϊnull).</zh>
   * <en>The IEffect3D object (null not permitted).</en>
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public Line3DRender(IEffect3D effect3D, IMatrixDataset dataset) {
    super(dataset) ;

    if (effect3D == null) {
      throw new IllegalArgumentException("Null 'effect3D' argument.") ;
    }
    this.effect = effect3D ;
  }

  /**
   * <zh>3DʾЧ.</zh>
   * <en>The IEffect3D object.</en>
   *
   * @return IEffect3D
   */
  public IEffect3D getEffect3D() {
    return this.effect ;
  }

  /**
   * <zh>ο3DʾЧ,
   * м{@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the IEffect3D object and sends a
   *  {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param effect3D IEffect3D <zh>3DʾЧ.</zh>
   * <en>The IEffect3D object.</en>
   */
  public void setEffect3D(IEffect3D effect3D) {
    if (effect3D == null) {
      throw new IllegalArgumentException("Null 'effect3D' argument.") ;
    }

    this.effect = effect3D ;
    fireRenderChanged() ;
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param categoryAxis CategoryAxis <zh>category (Ϊnull).</zh>
   * <en>The category axis (null not permitted).</en>
   * @param valueAxis AbstractValueAxis <zh>value (Ϊnull).</zh>
   * <en>The value axis (null not permitted).</en>
   * @param valueAxisEdge RectangleEdge <zh>valueλ (Ϊnull).</zh>
   * <en>The value axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(CategoryAxis categoryAxis, AbstractValueAxis valueAxis,
                           RectangleEdge valueAxisEdge, Rectangle2D dataArea,
                           ChartOrientation orientation, RenderShapeCollection renderShapes,
                           float alpha) {
    if (categoryAxis == null) {
      throw new IllegalArgumentException("Null 'categoryAxis' argument.") ;
    }
    if (valueAxis == null) {
      throw new IllegalArgumentException("Null 'valueAxis' argument.") ;
    }
    if (valueAxisEdge == null) {
      throw new IllegalArgumentException("Null 'valueAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }
    int columns = dataset.getColumnCount() ;
    int rows = dataset.getRowCount() ;

    this.fillPaints = getFillPaints(columns) ;
    this.borderPaints = getBorderPaints(columns) ;
    this.borderStrokes = getBorderStrokes(columns) ;

    //ߺͼ
    if (orientation.equals(ChartOrientation.VERTICAL)) { //ֱ
      double gapSize = 0 ; //С
      if (columns > 1) {
        gapSize = dataArea.getWidth() * categoryAxis.getCategoryMargin() / (columns - 1) ;
      }

      //ÿϵеĿ
      double categoryWidth = dataArea.getWidth()
                             * (1 - categoryAxis.getLowerMargin() - categoryAxis.getUpperMargin()
                                - ((columns == 1) ? 0 : categoryAxis.getCategoryMargin()))
                             / columns ;

      double startX = dataArea.getX() + dataArea.getWidth() * categoryAxis.getLowerMargin()
                      + 0.5 * categoryWidth ;
      for (int i = 0 ; i < rows ; i++) {
        double lastY = Double.NaN ;
        int currRowCols = dataset.getColumnCount(i) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof NumberChartData) {
            double value = ((NumberChartData) data).getValue() ;
            double y = valueAxis.valueToJava2D(dataArea, valueAxisEdge, value, false) ;
            if (j == 0) {
              lastY = y ;
              continue ;
            }

            //ֵ᷶Χֵ
            if (y == AbstractValueAxis.INVALID_COORDINATE_VALUE || Double.isNaN(lastY)
                || lastY == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
              renderShapes.addDataItemShape(null) ;
              lastY = Double.NaN ;
              continue ;
            }

            double sign = valueAxis.isInverted() ? -1 : 1 ; //+-
            double x1 = startX + (j - 1) * (categoryWidth + gapSize) ;
            double y1 = lastY ;
            double x2 = x1 - sign * this.effect.getXOffset() ;
            double y2 = y1 + this.effect.getYOffset() ;

            double x3 = startX + j * (categoryWidth + gapSize) ;
            double y3 = y ;
            double x4 = x3 - sign * this.effect.getXOffset() ;
            double y4 = y + this.effect.getYOffset() ;
            lastY = y ;

            GeneralPath path = new GeneralPath() ;
            path.moveTo((float) x1, (float) y1) ;
            path.lineTo((float) x2, (float) y2) ;
            path.lineTo((float) x4, (float) y4) ;
            path.lineTo((float) x3, (float) y3) ;
            path.closePath() ;

            Rectangle2D drawArea = new Rectangle2D.Double(x1, Math.min(y1, y4),
                                                          Math.abs(x3 - x1 - this.effect.getXOffset()),
                                                          Math.abs(y4 - y1)) ;

            Object[] params = {path, drawArea} ;
            IChartShape shape = createItemShape(orientation, valueAxis, i, j - 1, null,
                                                fillPaints[j - 1], borderPaints[j - 1],
                                                borderStrokes[j - 1], alpha, params) ;
            renderShapes.addDataItemShape(shape) ;
          } else {
            lastY = Double.NaN ;
            renderShapes.addDataItemShape(null) ;
          }
        }
      }

    } else if (orientation.equals(ChartOrientation.HORIZONTAL)) { //ˮƽ
      double gapSize = 0 ; //С
      if (columns > 1) {
        gapSize = dataArea.getHeight() * categoryAxis.getCategoryMargin() / (columns - 1) ;
      }

      //ÿϵеĿ
      double categoryWidth = dataArea.getHeight()
                             * (1 - categoryAxis.getLowerMargin() - categoryAxis.getUpperMargin()
                                - ((columns == 1) ? 0 : categoryAxis.getCategoryMargin()))
                             / columns ;

      double startY = dataArea.getY() + dataArea.getHeight() * categoryAxis.getLowerMargin()
                      + 0.5 * categoryWidth ;
      for (int i = 0 ; i < rows ; i++) {
        double lastX = Double.NaN ;
        int currRowCols = dataset.getColumnCount(i) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof NumberChartData) {
            double value = ((NumberChartData) data).getValue() ;
            double x = valueAxis.valueToJava2D(dataArea, valueAxisEdge, value, false) ;
            if (j == 0) {
              lastX = x ;
              continue ;
            }

            //ֵ᷶Χֵ
            if (x == AbstractValueAxis.INVALID_COORDINATE_VALUE || Double.isNaN(lastX)
                || lastX == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
              renderShapes.addDataItemShape(null) ;
              lastX = Double.NaN ;
              continue ;
            }

            double sign = valueAxis.isInverted() ? 1 : -1 ; //+-
            double x1 = lastX ;
            double y1 = startY + (j - 1) * (categoryWidth + gapSize) ;
            double x2 = x1 - this.effect.getXOffset() ;
            double y2 = y1 + sign * this.effect.getYOffset() ;

            double x3 = x ;
            double y3 = startY + j * (categoryWidth + gapSize) ;
            double x4 = x - this.effect.getXOffset() ;
            double y4 = y3 + sign * this.effect.getYOffset() ;

            lastX = x ;

            GeneralPath path = new GeneralPath() ;
            path.moveTo((float) x1, (float) y1) ;
            path.lineTo((float) x2, (float) y2) ;
            path.lineTo((float) x4, (float) y4) ;
            path.lineTo((float) x3, (float) y3) ;
            path.closePath() ;

            Rectangle2D drawArea = new Rectangle2D.Double(Math.min(x1, x3), y1,
                                                          Math.abs(x3 - x1 - this.effect.getXOffset()),
                                                          Math.abs(y4 - y1)) ;

            Object[] params = {path, drawArea} ;
            IChartShape shape = createItemShape(orientation, valueAxis, i, j - 1, null,
                                                fillPaints[j - 1], borderPaints[j - 1],
                                                borderStrokes[j - 1], alpha, params) ;
            renderShapes.addDataItemShape(shape) ;
          } else {
            lastX = Double.NaN ;
            renderShapes.addDataItemShape(null) ;
          }
        }
      }

    }

  }

  /**
   * <zh>ͼĸ.</zh>
   * <en>Returns the count of legends.</en>
   *
   * @return int <zh>ͼĸ.</zh>
   * <en>The count of legends.</en>
   */
  protected int getLegendItemCount() {
    int count = super.getLegendItemCount() ;

    return (count > 0) ? Math.max(1, count - 1) : 0 ;
  }

  /**
   * <zh>ֻͼ,ͼε.</zh>
   * <en>Create a shape with no attribute specified.</en>
   *
   * @param rect Rectangle2D <zh>ͼľ (Ϊnull).</zh>
   * <en>The bar rectangle (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   * @param appendParam Object <zh>ӵĲ,ڴݶĶ (Ϊnull).</zh>
   * <en>The append parameter,used for special propose (null permitted).</en>
   *
   * @return IChartShape <zh>ͼ.</zh>
   * <en>The shape.</en>
   */
  protected IChartShape createItemShapeOnly(Rectangle2D rect, String text,
                                            Object appendParam) {
    if (appendParam == null) {
      throw new IllegalArgumentException("Null 'appendParam' argument.") ;
    }

    Object[] params = (Object[]) appendParam ;
    GeneralPath path = (GeneralPath) params[0] ;
    Rectangle2D drawArea = (Rectangle2D) params[1] ;

    GeneralPathChartShape shape = new GeneralPathChartShape(path, drawArea, null) ;
    shape.setText(text) ;

    return shape ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof Line3DRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    Line3DRender that = (Line3DRender) obj ;
    if (this.effect != that.effect) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = 37 * result + this.effect.hashCode() ;

    return result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[effect=" + this.effect + "]" ;
  }

}
