package com.ijchart.xychart.render.paint ;

import java.awt.Color ;
import java.awt.Paint ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Collections ;
import java.util.List ;

import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ֵpaint.</zh>
 * <en>A paint scale that uses a lookup table to associate paint instances with
 *  data value ranges.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class LookupPaintScale implements IPaintScale, Serializable, Cloneable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -801375752813773389L ;

  /**
   * <zh>ȱʡľ̬paint.</zh>
   * <en>The default static paint object.</en>
   */
  public static final transient Paint DEFAULT_PAINT = Color.lightGray ;

  /**
   * <zh>Сֵ߽.</zh>
   * <en>The lower bound.</en>
   */
  private double lowerBound ;

  /**
   * <zh>ֵ߽.</zh>
   * <en>The upper bound.</en>
   */
  private double upperBound ;

  /**
   * <zh>ȱʡpaint.</zh>
   * <en>The default paint.</en>
   */
  private transient Paint defaultPaint ;

  /**
   * <zh>ұ.</zh>
   * <en>The lookup table.</en>
   */
  private transient List lookupTable ;

  /**
   * <zh>ͼͼʱʾķ̶ֵ.</zh>
   * <en>The symbols for drawing legend shape.</en>
   */
  private String[] legendSymbols ;

  /**
   * <zh>ȱʡĹ.</zh>
   * <en>The default constructor.</en>
   */
  public LookupPaintScale() {
    this(DEFAULT_LOWER_BOUND, DEFAULT_UPPER_BOUND) ;
  }

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param lowerBound <zh>Сֵ߽.</zh>
   * <en>The lower bound.</en>
   * @param upperBound <zh>ֵ߽.</zh>
   * <en>The upper bound.</en>
   */
  public LookupPaintScale(double lowerBound, double upperBound) {
    this(lowerBound, upperBound, DEFAULT_PAINT) ;
  }

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param lowerBound <zh>Сֵ߽.</zh>
   * <en>The lower bound.</en>
   * @param upperBound <zh>ֵ߽.</zh>
   * <en>The upper bound.</en>
   * @param defaultPaint <zh>ȱʡpaint (Ϊnull).</zh>
   * <en>The default paint object (null not permitted).</en>
   */
  public LookupPaintScale(double lowerBound, double upperBound, Paint defaultPaint) {
    this(lowerBound, upperBound, defaultPaint, null) ;
  }

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param lowerBound <zh>Сֵ߽.</zh>
   * <en>The lower bound.</en>
   * @param upperBound <zh>ֵ߽.</zh>
   * <en>The upper bound.</en>
   * @param defaultPaint <zh>ȱʡpaint (Ϊnull).</zh>
   * <en>The default paint object (null not permitted).</en>
   * <zh>ͼͼʱʾķ̶ֵ.</zh>
   * <en>The symbols for drawing legend shape.</en>
   * @param legendSymbols <zh>ͼǩ.</zh>
   * <en>The legend symbols.</en>
   */
  public LookupPaintScale(double lowerBound, double upperBound, Paint defaultPaint,
                          String[] legendSymbols) {
    if (lowerBound >= upperBound) {
      throw new IllegalArgumentException("Must lowerBound < upperBound.") ;
    }
    if (defaultPaint == null) {
      throw new IllegalArgumentException("Null 'defaultPaint' argument.") ;
    }
/** @todo ໹ҪĽԲμxyblockchartdemo3. */
    this.lowerBound = lowerBound ;
    this.upperBound = upperBound ;
    this.defaultPaint = defaultPaint ;
    this.lookupTable = new ArrayList() ;
    this.legendSymbols = legendSymbols ;
  }

  /**
   * <zh>Сֵ߽.</zh>
   * <en>The lower bound.</en>
   *
   * @return double <zh>Сֵ߽.</zh>
   * <en>The lower bound.</en>
   */
  public double getLowerBound() {
    return this.lowerBound ;
  }

  /**
   * <zh>ֵ߽.</zh>
   * <en>The upper bound.</en>
   *
   * @return double <zh>ֵ߽.</zh>
   * <en>The upper bound.</en>
   */
  public double getUpperBound() {
    return this.upperBound ;
  }

  /**
   * <zh>ȱʡpaint.</zh>
   * <en>Returns the default paint.</en>
   *
   * @return Paint <zh>paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getDefaultPaint() {
    return this.defaultPaint ;
  }

  /**
   * <zh>ػͼͼʱʾķ̶ֵ.</zh>
   * <en>Returns the symbols for drawing legend shape.</en>
   *
   * @return String[] <zh>ַ (Ϊnull).</zh>
   * <en>The string array.</en>
   */
  public String[] getLegendSymbols() {
    return legendSymbols ;
  }

  /**
   * <zh>һֵpaintĶӦϵ.</zh>
   * <en>Adds an value and a paint object to lookup table.</en>
   *
   * @param value double <zh>ֵ.</zh>
   * <en>The value.</en>
   * @param paint Paint <zh>paint.</zh>
   * <en>The paint object.</en>
   */
  public void add(double value, Paint paint) {
    ValuePaintPair item = new ValuePaintPair(value, paint) ;
    int index = Collections.binarySearch(this.lookupTable, item) ;
    if (index >= 0) {
      this.lookupTable.set(index, item) ;
    } else {
      this.lookupTable.add( -(index + 1), item) ;
    }
  }

  /**
   * <zh>ĳֵӦpaint.</zh>
   * <en>Returns a Paint instance for the specified value.</en>
   *
   * @param value double <zh>ֵ.</zh>
   * <en>The value.</en>
   * @return Paint <zh>paint (Ϊnull).</zh>
   * <en>The paint object (never null).</en>
   */
  public Paint getPaint(double value) {
    if (value < this.lowerBound || value > this.upperBound) {
      return this.defaultPaint ;
    }

    int count = this.lookupTable.size() ;
    if (count == 0) {
      return this.defaultPaint ;
    }

    ValuePaintPair item = (ValuePaintPair)this.lookupTable.get(0) ;
    if (value < item.value) {
      return this.defaultPaint ;
    }

    int low = 0 ;
    int high = this.lookupTable.size() - 1 ;
    while (high - low > 1) {
      int current = (low + high) / 2 ;
      item = (ValuePaintPair)this.lookupTable.get(current) ;
      if (value >= item.value) {
        low = current ;
      } else {
        high = current ;
      }
    }

    if (high > low) {
      item = (ValuePaintPair)this.lookupTable.get(high) ;
      if (value < item.value) {
        item = (ValuePaintPair)this.lookupTable.get(low) ;
      }
    }

    return (item != null) ? item.paint : this.defaultPaint ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof LookupPaintScale)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    LookupPaintScale that = (LookupPaintScale) obj ;
    if (this.lowerBound != that.lowerBound) {
      return false ;
    }
    if (this.upperBound != that.upperBound) {
      return false ;
    }
    if (!EqualsUtil.equals(this.defaultPaint, that.defaultPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.lookupTable, that.lookupTable)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.legendSymbols, that.legendSymbols)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() ;
    result = 37 * result + this.lowerBound ;
    result = 37 * result + this.upperBound ;
    result = 37 * result + HashcodeUtil.hashCode(this.defaultPaint) ;
    result = 37 * result + HashcodeUtil.hashCode(this.lookupTable) ;
    result = 37 * result + HashcodeUtil.hashCode(this.legendSymbols) ;

    return (int) result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[lowerBound=" + this.lowerBound
      + ",upperBound=" + this.upperBound
      + ",defaultPaint=" + this.defaultPaint
      + ",lookupTable size=" + this.lookupTable.size()
      + "]" ;
  }

  /**
   * <zh>л</zh>
   * <en>Provides serialization support</en>
   * @param stream <zh></zh>
   * <en>the output stream</en>
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writePaint(stream, this.defaultPaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.defaultPaint = SerialUtil.readPaint(stream) ;
  }

  /**
   * <zh>һֵpaint.</zh>
   * <en>A value and a paint object.</en>
   */
  class ValuePaintPair implements Comparable, Serializable {
    /**
     * <zh>л汾ID.</zh>
     * <en>For serialization.</en>
     */
    private static final long serialVersionUID = 698920578512361571L ;

    /**
     * <zh>ֵ.</zh>
     * <en>The value.</en>
     */
    private double value ;

    /**
     * <zh>paint.</zh>
     * <en>The paint.</en>
     */
    private transient Paint paint ;

    /**
     * <zh>.</zh>
     * <en>The constructor.</en>
     *
     * @param value double <zh>ֵ.</zh>
     * <en>The value.</en>
     * @param paint Paint <zh>paint.</zh>
     * <en>The paint.</en>
     */
    public ValuePaintPair(double value, Paint paint) {
      this.value = value ;
      this.paint = paint ;
    }

    /* (non-Javadoc)
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public int compareTo(Object obj) {
      ValuePaintPair that = (ValuePaintPair) obj ;
      double d1 = this.value ;
      double d2 = that.value ;

      if (d1 > d2) {
        return 1 ;
      }
      if (d1 < d2) {
        return -1 ;
      }

      return 0 ;
    }

    /**
     * <zh>hashcodeֵ.</zh>
     * <en>Returns a hash code value for the object.</en>
     *
     * @return int <zh>hashcodeֵ.</zh>
     * <en>A hash code value for this object.</en>
     */
    public int hashCode() {
      double hashcode = super.hashCode() ;
      hashcode = 37 * hashcode + this.value ;
      hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.paint) ;

      return (int) hashcode ;
    }

    /**
     * Tests this item for equality with an arbitrary object.
     *
     * @param obj  the object (<code>null</code> permitted).
     *
     * @return A boolean.
     */
    public boolean equals(Object obj) {
      if (obj == this) {
        return true ;
      }
      if (!(obj instanceof ValuePaintPair)) {
        return false ;
      }

      ValuePaintPair that = (ValuePaintPair) obj ;
      if (this.value != that.value) {
        return false ;
      }
      if (!EqualsUtil.equals(this.paint, that.paint)) {
        return false ;
      }

      return true ;
    }

    /**
     * Provides serialization support.
     *
     * @param stream  the output stream.
     *
     * @throws IOException  if there is an I/O error.
     */
    private void writeObject(ObjectOutputStream stream) throws IOException {
      stream.defaultWriteObject() ;
      SerialUtil.writePaint(stream, this.paint) ;
    }

    /**
     * Provides serialization support.
     *
     * @param stream  the input stream.
     *
     * @throws IOException  if there is an I/O error.
     * @throws ClassNotFoundException  if there is a classpath problem.
     */
    private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
      stream.defaultReadObject() ;
      this.paint = SerialUtil.readPaint(stream) ;
    }
  }
}
