package com.ijchart.xychart.render.xy ;

import java.awt.Color ;
import java.awt.Paint ;
import java.awt.geom.Ellipse2D ;
import java.awt.geom.GeneralPath ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.ArrayList ;
import java.util.Collections ;
import java.util.Iterator ;
import java.util.List ;

import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.axis.value.BoxAndWhiskerValueAxis ;
import com.ijchart.xychart.axis.value.NumberValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.IAngle ;
import com.ijchart.xychart.constant.LegendType ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.data.IntervalXBoxAndWhiskerData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.dataset.SingleRowDataset ;
import com.ijchart.xychart.range.NumberRange ;
import com.ijchart.xychart.render.category.outlier.Outlier ;
import com.ijchart.xychart.render.category.outlier.OutlierList ;
import com.ijchart.xychart.render.category.outlier.OutlierListCollection ;
import com.ijchart.xychart.shape.BoxAndWhiskerChartShape ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>BoxAndWhisker()ͼ.</zh>
 * <en>Draw Box and Whisker shapes.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class XYBoxAndWhiskerRender extends AbstractXYRender implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 8238411775397449615L ;

  /**
   * <zh>ȱʡмߺƽֵǵPaint.</zh>
   * <en>The default color used to paint the median line and average marker.</en>
   */
  public static final Paint DEFAULT_ARTIFACT_PAINT = Color.black ;

  /**
   * <zh>мߺƽֵǵPaint.</zh>
   * <en>The color used to paint the median line and average marker.</en>
   */
  private transient Paint artifactPaint ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset SingleRowDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public XYBoxAndWhiskerRender(SingleRowDataset dataset) {
    super(dataset) ;

    this.artifactPaint = DEFAULT_ARTIFACT_PAINT ;
    setBaseItemDrawBorder(true) ;
  }

  /**
   * <zh>ػмߺƽֵǵPaint.</zh>
   * <en>Returns the color used to paint the median line and average marker.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getArtifactPaint() {
    return this.artifactPaint ;
  }

  /**
   * <zh>ûмߺƽֵǵPaint,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>The color used to paint the median line and average marker,and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setArtifactPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.artifactPaint = paint ;
    fireRenderChanged() ;
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param xAxis AbstractValueAxis <zh>xֵ (Ϊnull).</zh>
   * <en>The x value axis (null not permitted).</en>
   * @param xAxisEdge RectangleEdge <zh>xλ(Ϊnull).</zh>
   * <en>The x axis edge.</en>
   * @param yAxis AbstractValueAxis <zh>yֵ (Ϊnull).</zh>
   * <en>The y value axis (null not permitted).</en>
   * @param yAxisEdge RectangleEdge <zh>yλ (Ϊnull).</zh>
   * <en>The y axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(AbstractValueAxis xAxis, RectangleEdge xAxisEdge,
                           AbstractValueAxis yAxis, RectangleEdge yAxisEdge,
                           Rectangle2D dataArea, ChartOrientation orientation,
                           RenderShapeCollection renderShapes, float alpha) {
    if (xAxis == null) {
      throw new IllegalArgumentException("Null 'xAxis' argument.") ;
    }
    if (xAxisEdge == null) {
      throw new IllegalArgumentException("Null 'xAxisEdge' argument.") ;
    }
    if (yAxis == null) {
      throw new IllegalArgumentException("Null 'yAxis' argument.") ;
    }
    if (yAxisEdge == null) {
      throw new IllegalArgumentException("Null 'yAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }
    int rows = dataset.getRowCount() ;
    boolean axisInverted = yAxis.isInverted() ; //valueǷת

    double minTickValue = Double.NaN ;
    double maxTickValue = Double.NaN ;
    if (yAxis instanceof NumberValueAxis) {
      NumberValueAxis axis = (NumberValueAxis) yAxis ;
      String[] numberTicks = axis.getNumberTicks() ;
      minTickValue = Double.parseDouble(numberTicks[0]) ;
      minTickValue = yAxis.valueToJava2D(dataArea, yAxisEdge, minTickValue, false) ;
      maxTickValue = Double.parseDouble(numberTicks[numberTicks.length - 1]) ;
      maxTickValue = yAxis.valueToJava2D(dataArea, yAxisEdge, maxTickValue, false) ;
    }

    this.fillPaints = getFillPaints(rows) ;
    getBorderPaints(rows) ;
    this.borderPaints = this.fillPaints ;
    this.borderStrokes = getBorderStrokes(rows) ;

    //BoxAndWhiskerͼ
    if (orientation.equals(ChartOrientation.VERTICAL)) { //ֱ
      if (dataset.getRowCount() > 0) {
        int currRowCols = dataset.getColumnCount(0) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          double yyMean = 0.0 ;
          double yyOutlier = 0.0 ;
          double aRadius = 0 ; // average radius

          IChartData data = dataset.getData(0, j) ;
          if (data != null && data instanceof IntervalXBoxAndWhiskerData) {
            IntervalXBoxAndWhiskerData boxData = (IntervalXBoxAndWhiskerData) data ;
            double startXValue = boxData.getStartX() ;
            double endXValue = boxData.getEndX() ;
            double yQ1 = boxData.getQ1() ;
            double yQ3 = boxData.getQ3() ;
            double yMax = boxData.getMaxRegularValue() ;
            double yMin = boxData.getMinRegularValue() ;
            if (!Double.isNaN(startXValue) && !Double.isNaN(endXValue) &&
                !Double.isNaN(yQ1) && !Double.isNaN(yQ3)
                && !Double.isNaN(yMax) && !Double.isNaN(yMin)) {
              double startX = xAxis.valueToJava2D(dataArea, xAxisEdge, startXValue, false) ;
              double endX = xAxis.valueToJava2D(dataArea, xAxisEdge, endXValue, false) ;
              double yyQ1 = yAxis.valueToJava2D(dataArea, yAxisEdge, yQ1, false) ;
              double yyQ3 = yAxis.valueToJava2D(dataArea, yAxisEdge, yQ3, false) ;
              double yyMax = yAxis.valueToJava2D(dataArea, yAxisEdge, yMax, false) ;
              double yyMin = yAxis.valueToJava2D(dataArea, yAxisEdge, yMin, false) ;
              //ֵ᷶Χֵ
              if (startX == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || endX == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || yyQ1 == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || yyQ3 == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || yyMax == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || yyMin == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
                renderShapes.addDataItemShape(null) ;
                continue ;
              }

              double yMean = boxData.getMean() ;
              double reallyBarWidth = Math.abs(endX - startX) ;
              double x11 = Math.min(startX, endX) ;
              Ellipse2D avgEllipse = null ;
              if (!Double.isNaN(yMean)) {
                yyMean = yAxis.valueToJava2D(dataArea, yAxisEdge, yMean, false) ;
                aRadius = reallyBarWidth / 4 ;
                if (isBetween(minTickValue, maxTickValue, yyMean + aRadius)
                    && isBetween(minTickValue, maxTickValue, yyMean - aRadius)) {
                  avgEllipse = new Ellipse2D.Double(x11 + aRadius, yyMean - aRadius, aRadius * 2,
                                                    aRadius * 2) ;
                }
              }

              double yMedian = boxData.getMedian() ;
              double yyMedian = Double.NaN ;
              if (!Double.isNaN(yMedian)) {
                yyMedian = yAxis.valueToJava2D(dataArea, yAxisEdge, yMedian, false) ;
              }

              // draw yOutliers...
              if (yAxis instanceof BoxAndWhiskerValueAxis) {
                BoxAndWhiskerValueAxis bawAxis = (BoxAndWhiskerValueAxis) yAxis ;
                NumberRange range = bawAxis.getZoomDatasetRange() ;
                if (range == null) {
                  range = bawAxis.getDatasetRange() ;
                }
                double maxAxisValue = yAxis.valueToJava2D(dataArea, yAxisEdge,
                                                          range.getUpperBound(), false) ;
                double minAxisValue = yAxis.valueToJava2D(dataArea, yAxisEdge,
                                                          range.getLowerBound(), false) ;

                // draw outliers
                double oRadius = reallyBarWidth / 3 ; // outlier radius
                List outliers = new ArrayList() ;
                OutlierListCollection outlierListCollection = new OutlierListCollection() ;

                // From outlier array sort out which are outliers and put these into a
                // list If there are any farouts, set the flag on the
                // OutlierListCollection
                List yOutliers = boxData.getOutliers() ;
                List ellipseList = new ArrayList() ;
                GeneralPath highFarBorders = null ;
                GeneralPath lowFarBorders = null ;
                if (yOutliers != null) {
                  for (int m = 0 ; m < yOutliers.size() ; m++) {
                    double outlier = ((Number) yOutliers.get(m)).doubleValue() ;
                    double minOutlier = boxData.getMinOutlier() ;
                    double maxOutlier = boxData.getMaxOutlier() ;
                    double minRegular = boxData.getMinRegularValue() ;
                    double maxRegular = boxData.getMaxRegularValue() ;
                    if (outlier > maxOutlier) {
                      outlierListCollection.setHighFarOut(true) ;
                    } else if (outlier < minOutlier) {
                      outlierListCollection.setLowFarOut(true) ;
                    } else if (outlier > maxRegular) {
                      yyOutlier = yAxis.valueToJava2D(dataArea, yAxisEdge, outlier, false) ;
                      outliers.add(new Outlier(x11 + reallyBarWidth / 2.0, yyOutlier, oRadius)) ;
                    } else if (outlier < minRegular) {
                      yyOutlier = yAxis.valueToJava2D(dataArea, yAxisEdge, outlier, false) ;
                      outliers.add(new Outlier(x11 + reallyBarWidth / 2.0, yyOutlier, oRadius)) ;
                    }
                    Collections.sort(outliers) ;
                  }

                  // Process outliers. Each outlier is either added to the
                  // appropriate outlier list or a new outlier list is made
                  for (Iterator iterator = outliers.iterator() ; iterator.hasNext() ; ) {
                    Outlier outlier = (Outlier) iterator.next() ;
                    outlierListCollection.add(outlier) ;
                  }

                  for (Iterator iterator = outlierListCollection.iterator() ; iterator.hasNext() ; ) {
                    OutlierList list = (OutlierList) iterator.next() ;
                    Outlier outlier = list.getAveragedOutlier() ;
                    Point2D point = outlier.getPoint() ;

                    if (list.isMultiple()) {
                      if (axisInverted) {
                        if (isBetween(minTickValue, maxTickValue, point.getY() + 2 * oRadius)
                            && isBetween(minTickValue, maxTickValue, point.getY() + oRadius)) {
                          Ellipse2D dot1 = new Ellipse2D.Double(point.getX() - (reallyBarWidth / 2)
                                                                + oRadius, point.getY() + oRadius,
                                                                oRadius,
                                                                oRadius) ;
                          ellipseList.add(dot1) ;
                          Ellipse2D dot2 = new Ellipse2D.Double(point.getX() + (reallyBarWidth / 2),
                                                                point.getY() + oRadius, oRadius,
                                                                oRadius) ;

                          ellipseList.add(dot2) ;
                        }
                      } else {
                        if (isBetween(minTickValue, maxTickValue, point.getY())
                            && isBetween(minTickValue, maxTickValue, point.getY() + oRadius)) {
                          Ellipse2D dot1 = new Ellipse2D.Double(point.getX()
                                                                - (reallyBarWidth / 2)
                                                                + oRadius, point.getY(), oRadius,
                                                                oRadius) ;
                          ellipseList.add(dot1) ;
                          Ellipse2D dot2 = new Ellipse2D.Double(point.getX()
                                                                + (reallyBarWidth / 2),
                                                                point.getY(), oRadius, oRadius) ;

                          ellipseList.add(dot2) ;
                        }
                      }
                    } else {
                      if (axisInverted) {
                        if (isBetween(minTickValue, maxTickValue, point.getY() + 2 * oRadius)
                            && isBetween(minTickValue, maxTickValue, point.getY() + oRadius)) {
                          Ellipse2D dot = new Ellipse2D.Double(point.getX() + oRadius / 2,
                                                               point.getY() + oRadius, oRadius,
                                                               oRadius) ;
                          ellipseList.add(dot) ;
                        }
                      } else {
                        if (isBetween(minTickValue, maxTickValue, point.getY())
                            && isBetween(minTickValue, maxTickValue, point.getY() + oRadius)) {
                          Ellipse2D dot = new Ellipse2D.Double(point.getX() + oRadius / 2,
                                                               point.getY(), oRadius, oRadius) ;
                          ellipseList.add(dot) ;
                        }
                      }
                    }
                  }

                  // draw farout indicators
                  if (outlierListCollection.isHighFarOut()) {
                    double xx = x11 + reallyBarWidth / 2.0 ;
                    int sign = axisInverted ? -1 : 1 ;
                    if (isBetween(minTickValue, maxTickValue, maxAxisValue + sign * aRadius)
                        && isBetween(minTickValue, maxTickValue, maxAxisValue)) {
                      highFarBorders = new GeneralPath() ;
                      highFarBorders.moveTo((float) (xx - aRadius),
                                            (float) (maxAxisValue + sign * aRadius)) ;
                      highFarBorders.lineTo((float) (xx + aRadius),
                                            (float) (maxAxisValue + sign * aRadius)) ;
                      highFarBorders.lineTo((float) xx, (float) maxAxisValue) ;
                      highFarBorders.lineTo((float) (xx - aRadius),
                                            (float) (maxAxisValue + sign * aRadius)) ;
                      highFarBorders.closePath() ;
                    }
                  }

                  if (outlierListCollection.isLowFarOut()) {
                    double xx = x11 + reallyBarWidth / 2.0 ;
                    int sign = axisInverted ? 1 : -1 ;
                    if (isBetween(minTickValue, maxTickValue, minAxisValue + sign * aRadius)
                        && isBetween(minTickValue, maxTickValue, minAxisValue)) {
                      lowFarBorders = new GeneralPath() ;
                      lowFarBorders.moveTo((float) (xx - aRadius),
                                           (float) (minAxisValue + sign * aRadius)) ;
                      lowFarBorders.lineTo((float) (xx + aRadius),
                                           (float) (minAxisValue + sign * aRadius)) ;
                      lowFarBorders.lineTo((float) (xx), (float) (minAxisValue)) ;
                      lowFarBorders.lineTo((float) (xx - aRadius),
                                           (float) (minAxisValue + sign * aRadius)) ;
                      lowFarBorders.closePath() ;
                    }
                  }
                }

                BoxAndWhiskerChartShape boxShape =
                  new BoxAndWhiskerChartShape(x11, Math.max(yyMax, yyMin),
                                              reallyBarWidth,
                                              Math.abs(yyMax - yyMin),
                                              axisInverted ? Math.abs(yyMin - yyQ1)
                                              : Math.abs(yyQ3 - yyMax),
                                              axisInverted ? Math.abs(yyQ3 - yyMax)
                                              : Math.abs(yyMin - yyQ1), avgEllipse,
                                              yyMedian, this.artifactPaint, null,
                                              (Ellipse2D[]) ellipseList.toArray(new Ellipse2D[0]),
                                              highFarBorders, lowFarBorders) ;
                IChartShape shape = createItemShape(orientation, yAxis, 0, j,
                                                    new Rectangle2D.Double(), fillPaints[0],
                                                    borderPaints[0], borderStrokes[0], alpha,
                                                    boxShape) ;
                renderShapes.addDataItemShape(shape) ;
                continue ;
              }
            }
            renderShapes.addDataItemShape(null) ;
          }
        }
      }
    } else if (orientation.equals(ChartOrientation.HORIZONTAL)) { //ˮƽ
      for (int i = 0 ; i < rows ; i++) {
        int currRowCols = dataset.getColumnCount(i) ;
        for (int j = 0 ; j < currRowCols ; j++) {
          double xxMean = 0.0 ;
          double aRadius = 0 ; // average radius
          double xxOutlier = 0.0 ;

          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof IntervalXBoxAndWhiskerData) {
            IntervalXBoxAndWhiskerData boxData = (IntervalXBoxAndWhiskerData) data ;
            double startXValue = boxData.getStartX() ;
            double endXValue = boxData.getEndX() ;
            double xQ1 = boxData.getQ1() ;
            double xQ3 = boxData.getQ3() ;
            double xMax = boxData.getMaxRegularValue() ;
            double xMin = boxData.getMinRegularValue() ;
            if (!Double.isNaN(xQ1) && !Double.isNaN(xQ3)
                && !Double.isNaN(xMax) && !Double.isNaN(xMin)) {
              double startX = xAxis.valueToJava2D(dataArea, xAxisEdge, startXValue, false) ;
              double endX = xAxis.valueToJava2D(dataArea, xAxisEdge, endXValue, false) ;
              double xxQ1 = yAxis.valueToJava2D(dataArea, yAxisEdge, xQ1, false) ;
              double xxQ3 = yAxis.valueToJava2D(dataArea, yAxisEdge, xQ3, false) ;
              double xxMax = yAxis.valueToJava2D(dataArea, yAxisEdge, xMax, false) ;
              double xxMin = yAxis.valueToJava2D(dataArea, yAxisEdge, xMin, false) ;
              double minxMaxMin = Math.min(xxMin, xxMax) ;
              //ֵ᷶Χֵ
              if (startX == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || endX == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || xxQ1 == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || xxQ3 == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || xxMax == AbstractValueAxis.INVALID_COORDINATE_VALUE
                  || xxMin == AbstractValueAxis.INVALID_COORDINATE_VALUE
                ) {
                renderShapes.addDataItemShape(null) ;
                continue ;
              }

              double y11 = Math.min(startX, endX) ;
              double reallyBarWidth = Math.abs(endX - startX) ;
              double halfReallyBarWidth = reallyBarWidth * 0.5 ;
              double xMean = boxData.getMean() ;
              Ellipse2D avgEllipse = null ;
              if (!Double.isNaN(xMean)) {
                xxMean = yAxis.valueToJava2D(dataArea, yAxisEdge, xMean, false) ;
                aRadius = reallyBarWidth / 4 ;
                if (isBetween(minTickValue, maxTickValue, xxMean - aRadius)
                    && isBetween(minTickValue, maxTickValue, xxMean + aRadius)) {
                  avgEllipse = new Ellipse2D.Double(minxMaxMin + aRadius,
                                                    y11 - Math.abs(xxMean - minxMaxMin) - aRadius,
                                                    aRadius * 2, aRadius * 2) ;
                }
              }

              double xMedian = boxData.getMedian() ;
              double xxMedian = Double.NaN ;
              if (!Double.isNaN(xMedian)) {
                xxMedian = yAxis.valueToJava2D(dataArea, yAxisEdge, xMedian, false) ;
                xxMedian = y11 - (xxMedian - Math.min(xxMax, xxMin)) ;
              }

              // draw yOutliers...
              if (yAxis instanceof BoxAndWhiskerValueAxis) {
                BoxAndWhiskerValueAxis bawAxis = (BoxAndWhiskerValueAxis) yAxis ;
                NumberRange range = bawAxis.getZoomDatasetRange() ;
                if (range == null) {
                  range = bawAxis.getDatasetRange() ;
                }
                double maxAxisValue = yAxis.valueToJava2D(dataArea, yAxisEdge,
                                                          range.getUpperBound(), false) ;
                double minAxisValue = yAxis.valueToJava2D(dataArea, yAxisEdge,
                                                          range.getLowerBound(), false) ;

                // draw outliers
                double oRadius = reallyBarWidth / 3 ; // outlier radius
                List outliers = new ArrayList() ;
                OutlierListCollection outlierListCollection = new OutlierListCollection() ;

                // From outlier array sort out which are outliers and put these into a
                // list If there are any farouts, set the flag on the
                // OutlierListCollection
                List xOutliers = boxData.getOutliers() ;
                List ellipseList = new ArrayList() ;
                GeneralPath highFarBorder = null ;
                GeneralPath lowFarBorder = null ;
                if (xOutliers != null) {
                  for (int m = 0 ; m < xOutliers.size() ; m++) {
                    double outlier = ((Number) xOutliers.get(m)).doubleValue() ;
                    double minOutlier = boxData.getMinOutlier() ;
                    double maxOutlier = boxData.getMaxOutlier() ;
                    double minRegular = boxData.getMinRegularValue() ;
                    double maxRegular = boxData.getMaxRegularValue() ;
                    if (outlier > maxOutlier) {
                      outlierListCollection.setHighFarOut(true) ;
                    } else if (outlier < minOutlier) {
                      outlierListCollection.setLowFarOut(true) ;
                    } else if (outlier > maxRegular) {
                      xxOutlier = yAxis.valueToJava2D(dataArea, yAxisEdge, outlier, false) ;
                      outliers.add(new Outlier(xxOutlier, y11 + halfReallyBarWidth, oRadius)) ;
                    } else if (outlier < minRegular) {
                      xxOutlier = yAxis.valueToJava2D(dataArea, yAxisEdge, outlier, false) ;
                      outliers.add(new Outlier(xxOutlier, y11 + halfReallyBarWidth, oRadius)) ;
                    }
                    Collections.sort(outliers) ;
                  }

                  // Process outliers. Each outlier is either added to the
                  // appropriate outlier list or a new outlier list is made
                  for (Iterator iterator = outliers.iterator() ; iterator.hasNext() ; ) {
                    Outlier outlier = (Outlier) iterator.next() ;
                    outlierListCollection.add(outlier) ;
                  }

                  for (Iterator iterator = outlierListCollection.iterator() ; iterator.hasNext() ; ) {
                    OutlierList list = (OutlierList) iterator.next() ;
                    Outlier outlier = list.getAveragedOutlier() ;
                    Point2D point = outlier.getPoint() ;

                    if (list.isMultiple()) {
                      if (axisInverted) {
                        if (isBetween(minTickValue, maxTickValue,
                                      minxMaxMin + Math.abs(point.getX() - minxMaxMin)
                                      - oRadius * 1.5)
                            && isBetween(minTickValue, maxTickValue,
                                         minxMaxMin + Math.abs(point.getX() - minxMaxMin)
                                         - oRadius * 0.5)) {
                          Ellipse2D dot1 = new Ellipse2D.Double(minxMaxMin + halfReallyBarWidth
                                                                - oRadius * 1.5, y11
                                                                + Math.abs(point.getX()
                                                                           - minxMaxMin)
                                                                - oRadius * 1.5, oRadius, oRadius) ;
                          ellipseList.add(dot1) ;
                          Ellipse2D dot2 = new Ellipse2D.Double(minxMaxMin + halfReallyBarWidth
                                                                + oRadius / 2, y11
                                                                + Math.abs(point.getX()
                                                                           - minxMaxMin)
                                                                - oRadius * 1.5, oRadius, oRadius) ;
                          ellipseList.add(dot2) ;

                        }
                      } else {
                        if (isBetween(minTickValue, maxTickValue,
                                      minxMaxMin + Math.abs(point.getX() - minxMaxMin) - oRadius)
                            && isBetween(minTickValue, maxTickValue,
                                         minxMaxMin + Math.abs(point.getX() - minxMaxMin)
                                         - 2 * oRadius)) {
                          Ellipse2D dot1 = new Ellipse2D.Double(minxMaxMin + halfReallyBarWidth
                                                                - oRadius * 1.5, y11
                                                                - Math.abs(point.getX()
                                                                           - minxMaxMin)
                                                                - oRadius, oRadius, oRadius) ;
                          ellipseList.add(dot1) ;
                          Ellipse2D dot = new Ellipse2D.Double(minxMaxMin + halfReallyBarWidth
                                                               + oRadius / 2, y11
                                                               - Math.abs(point.getX() - minxMaxMin)
                                                               - oRadius, oRadius, oRadius) ;
                          ellipseList.add(dot) ;
                        }
                      }
                    } else {
                      if (axisInverted) {
                        if (isBetween(minTickValue, maxTickValue,
                                      minxMaxMin + Math.abs(point.getX() - minxMaxMin)
                                      - oRadius * 1.5)
                            && isBetween(minTickValue, maxTickValue,
                                         minxMaxMin + Math.abs(point.getX() - minxMaxMin)
                                         - oRadius * 0.5)) {
                          Ellipse2D dot = new Ellipse2D.Double(minxMaxMin + halfReallyBarWidth
                                                               - oRadius / 2, y11
                                                               + Math.abs(point.getX() - minxMaxMin)
                                                               - oRadius * 1.5, oRadius, oRadius) ;
                          ellipseList.add(dot) ;
                        }
                      } else {
                        if (isBetween(minTickValue, maxTickValue,
                                      minxMaxMin + Math.abs(point.getX() - minxMaxMin) - oRadius)
                            && isBetween(minTickValue, maxTickValue,
                                         minxMaxMin + Math.abs(point.getX() - minxMaxMin)
                                         - 2 * oRadius)) {
                          Ellipse2D dot = new Ellipse2D.Double(minxMaxMin + halfReallyBarWidth
                                                               - oRadius / 2, y11
                                                               - Math.abs(point.getX() - minxMaxMin)
                                                               - oRadius, oRadius, oRadius) ;
                          ellipseList.add(dot) ;
                        }
                      }
                    }
                  }

                  // draw farout indicators
                  if (outlierListCollection.isHighFarOut()) {
                    if (axisInverted) {
                      if (isBetween(minTickValue, maxTickValue,
                                    minxMaxMin - (maxAxisValue - minxMaxMin))
                          && isBetween(minTickValue, maxTickValue,
                                       minxMaxMin - (maxAxisValue - minxMaxMin - aRadius))) {
                        highFarBorder = new GeneralPath() ;
                        highFarBorder.moveTo((float) (minxMaxMin + halfReallyBarWidth - aRadius),
                                             (float) (y11 + maxAxisValue - minxMaxMin - aRadius)) ;
                        highFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth + aRadius),
                                             (float) (y11 + maxAxisValue - minxMaxMin - aRadius)) ;
                        highFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth),
                                             (float) (y11 + maxAxisValue - minxMaxMin)) ;
                        highFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth - aRadius),
                                             (float) (y11 + maxAxisValue - minxMaxMin - aRadius)) ;
                        highFarBorder.closePath() ;
                      }
                    } else {
                      double len = Math.abs(xxMax - xxMin) ;
                      if (isBetween(minTickValue, maxTickValue, minxMaxMin + len - aRadius)
                          && isBetween(minTickValue, maxTickValue, minxMaxMin + len)) {
                        highFarBorder = new GeneralPath() ;
                        highFarBorder.moveTo((float) (minxMaxMin + halfReallyBarWidth - aRadius),
                                             (float) (y11 - len + aRadius)) ;
                        highFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth + aRadius),
                                             (float) (y11 - len + aRadius)) ;
                        highFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth),
                                             (float) (y11 - len)) ;
                        highFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth - aRadius),
                                             (float) (y11 - len + aRadius)) ;
                        highFarBorder.closePath() ;
                      }
                    }
                  }

                  if (outlierListCollection.isLowFarOut()) {
                    if (axisInverted) {
                      if (isBetween(minTickValue, maxTickValue,
                                    minxMaxMin + (minAxisValue - minxMaxMin))
                          && isBetween(minTickValue, maxTickValue,
                                       minxMaxMin + (minAxisValue - minxMaxMin - aRadius))) {
                        lowFarBorder = new GeneralPath() ;
                        lowFarBorder.moveTo((float) (minxMaxMin + halfReallyBarWidth - aRadius),
                                            (float) (y11 - (minAxisValue - minxMaxMin - aRadius))) ;
                        lowFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth + aRadius),
                                            (float) (y11 - (minAxisValue - minxMaxMin - aRadius))) ;
                        lowFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth),
                                            (float) (y11 - (minAxisValue - minxMaxMin))) ;
                        lowFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth - aRadius),
                                            (float) (y11 - (minAxisValue - minxMaxMin - aRadius))) ;
                        lowFarBorder.closePath() ;
                      }
                    } else {
                      double len = Math.abs(minTickValue - minxMaxMin) - aRadius ;
                      if (isBetween(minTickValue, maxTickValue, minxMaxMin - (len - aRadius))
                          && isBetween(minTickValue, maxTickValue, minxMaxMin - len)) {
                        lowFarBorder = new GeneralPath() ;
                        lowFarBorder.moveTo((float) (minxMaxMin + halfReallyBarWidth - aRadius),
                                            (float) (y11 + len - aRadius)) ;
                        lowFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth + aRadius),
                                            (float) (y11 + len - aRadius)) ;
                        lowFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth),
                                            (float) (y11 + len)) ;
                        lowFarBorder.lineTo((float) (minxMaxMin + halfReallyBarWidth - aRadius),
                                            (float) (y11 + len - aRadius)) ;
                        lowFarBorder.closePath() ;
                      }
                    }
                  }
                }

                BoxAndWhiskerChartShape boxShape =
                  new BoxAndWhiskerChartShape(minxMaxMin, y11, reallyBarWidth,
                                              Math.abs(xxMax - xxMin),
                                              axisInverted ? Math.abs(xxMin - xxQ1)
                                              : Math.abs(xxQ3 - xxMax),
                                              axisInverted ? Math.abs(xxQ3 - xxMax)
                                              : Math.abs(xxMin - xxQ1), avgEllipse, xxMedian,
                                              this.artifactPaint,
                                              (Ellipse2D[]) ellipseList.toArray(new Ellipse2D[0]),
                                              highFarBorder, lowFarBorder) ;
                boxShape.setRotate(IAngle.ANGLE_DOWN_90) ;
                IChartShape shape = createItemShape(orientation, yAxis, 0, j,
                                                    new Rectangle2D.Double(), fillPaints[0],
                                                    borderPaints[0], borderStrokes[0], alpha,
                                                    boxShape) ;
                renderShapes.addDataItemShape(shape) ;
                continue ;
              }
            }
            renderShapes.addDataItemShape(null) ;
          }
        }
      }

    }

  }

  /**
   * <zh>ͼ.</zh>
   * <en>Return the legend type.</en>
   *
   * @return LegendType <zh>ͼ.</zh>
   * <en>The legend type.</en>
   */
  public LegendType getLegendType() {
    return LegendType.ROW ;
  }

  /**
   * <zh>жvalueǷminValuemaxValue֮.</zh>
   * <en>Returns true if value between minValue and maxValue,otherwise return false.</en>
   * @param minValue double <zh>Сֵ.</zh>
   * <en>The minimum value.</en>
   * @param maxValue double <zh>ֵ.</zh>
   * <en>The maximum value.</en>
   * @param value double <zh>ҪȽϵֵ.</zh>
   * <en>The value.</en>
   *
   * @return boolean
   */
  private static boolean isBetween(double minValue, double maxValue, double value) {
    if (Double.isNaN(minValue) || Double.isNaN(maxValue) || Double.isNaN(value)) {
      return false ;
    }

    return (value >= minValue && value <= maxValue) || value >= maxValue && value <= minValue ;
  }

  /**
   * <zh>ֻͼ,ͼε.</zh>
   * <en>Create a shape with no attribute specified.</en>
   *
   * @param rect Rectangle2D <zh>ͼľ (Ϊnull).</zh>
   * <en>The bar rectangle (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   * @param appendParam Object <zh>ӵĲ,ڴݶĶ (Ϊnull).</zh>
   * <en>The append parameter,used for special propose (null permitted).</en>
   *
   * @return IChartShape <zh>ͼ.</zh>
   * <en>The shape.</en>
   */
  protected IChartShape createItemShapeOnly(Rectangle2D rect, String text,
                                            Object appendParam) {
    if (appendParam == null) {
      throw new IllegalArgumentException("Null 'appendParam' argument.") ;
    }

    return (IChartShape) appendParam ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof XYBoxAndWhiskerRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    XYBoxAndWhiskerRender that = (XYBoxAndWhiskerRender) obj ;
    if (!EqualsUtil.equals(this.artifactPaint, that.artifactPaint)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = super.hashCode() + HashcodeUtil.hashCode(this.artifactPaint) ;

    return result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "[artifactPaint=" + this.artifactPaint + "]" ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writePaint(stream, this.artifactPaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.artifactPaint = SerialUtil.readPaint(stream) ;
  }

}
