package com.ijchart.xychart.render.xy ;

import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.LegendType ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.data.XYZNumberChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.shape.EllipseTextChartShape ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>XYͼͼ.</zh>
 * <en>A render that draws shapes for each data item, and lines between data
 * items (for use with the {@link com.ijchart.xychart.plot.XYPlot} class).</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class XYBubbleRender extends AbstractXYRender implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 5641551178654361705L ;

  /**
   * <zh>domainvalue᷽.</zh>
   * <en>Scaled on domain axis and value axis.</en>
   */
  public static final int SCALE_ON_BOTH_AXES = 0 ;

  /**
   * <zh>domain᷽.</zh>
   * <en>Scaled on domain axis.</en>
   */
  public static final int SCALE_ON_DOMAIN_AXIS = 1 ;

  /**
   * <zh>value᷽.</zh>
   * <en>Scaled on value axis.</en>
   */
  public static final int SCALE_ON_RANGE_AXIS = 2 ;

  /**
   * <zh>.</zh>
   * <en>The scale type.</en>
   */
  private int scaleType ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public XYBubbleRender(IMatrixDataset dataset) {
    this(dataset, SCALE_ON_BOTH_AXES) ;
  }

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   * @param scaleType int <zh>,ֵܵΪ:{@link #SCALE_ON_BOTH_AXES},
   * {@link #SCALE_ON_DOMAIN_AXIS},{@link #SCALE_ON_RANGE_AXIS}.</zh>
   * <en>The scale type (must be one of:{@link #SCALE_ON_BOTH_AXES}, {@link #SCALE_ON_DOMAIN_AXIS},
   * {@link #SCALE_ON_RANGE_AXIS}).</en>
   */
  public XYBubbleRender(IMatrixDataset dataset, int scaleType) {
    super(dataset) ;
    if (scaleType < 0 || scaleType > 2) {
      throw new IllegalArgumentException("Invalid 'scaleType'.") ;
    }

    this.scaleType = scaleType ;
  }

  /**
   * <zh>.</zh>
   * <en>Returns the scale type.</en>
   *
   * @return int <zh>,ֵܵΪ:{@link #SCALE_ON_BOTH_AXES},
   * {@link #SCALE_ON_DOMAIN_AXIS},{@link #SCALE_ON_RANGE_AXIS}.</zh>
   * <en>The type,must be one of:{@link #SCALE_ON_BOTH_AXES}, {@link #SCALE_ON_DOMAIN_AXIS},
   * {@link #SCALE_ON_RANGE_AXIS}.</en>
   */
  public int getScaleType() {
    return this.scaleType ;
  }

  /**
   * <zh>.</zh>
   * <en>The scale type.</en>
   *
   * @param scaleType int <zh>,ֵܵΪ:{@link #SCALE_ON_BOTH_AXES},
   * {@link #SCALE_ON_DOMAIN_AXIS},{@link #SCALE_ON_RANGE_AXIS}.</zh>
   * <en>The type,must be one of:{@link #SCALE_ON_BOTH_AXES}, {@link #SCALE_ON_DOMAIN_AXIS},
   * {@link #SCALE_ON_RANGE_AXIS}.</en>

   */
  public void setScaleType(int scaleType) {
    if (scaleType < 0 || scaleType > 2) {
      throw new IllegalArgumentException("Invalid 'scaleType'.") ;
    }

    if (this.scaleType != scaleType) {
      this.scaleType = scaleType ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param xAxis AbstractValueAxis <zh>xֵ (Ϊnull).</zh>
   * <en>The x value axis (null not permitted).</en>
   * @param xAxisEdge RectangleEdge <zh>xλ(Ϊnull).</zh>
   * <en>The x axis edge.</en>
   * @param yAxis AbstractValueAxis <zh>yֵ (Ϊnull).</zh>
   * <en>The y value axis (null not permitted).</en>
   * @param yAxisEdge RectangleEdge <zh>yλ (Ϊnull).</zh>
   * <en>The y axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(AbstractValueAxis xAxis, RectangleEdge xAxisEdge,
                           AbstractValueAxis yAxis, RectangleEdge yAxisEdge,
                           Rectangle2D dataArea, ChartOrientation orientation,
                           RenderShapeCollection renderShapes, float alpha) {
    if (xAxis == null) {
      throw new IllegalArgumentException("Null 'xAxis' argument.") ;
    }
    if (xAxisEdge == null) {
      throw new IllegalArgumentException("Null 'xAxisEdge' argument.") ;
    }
    if (yAxis == null) {
      throw new IllegalArgumentException("Null 'yAxis' argument.") ;
    }
    if (yAxisEdge == null) {
      throw new IllegalArgumentException("Null 'yAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }

    int rows = dataset.getRowCount() ;
    this.fillPaints = getFillPaints(rows) ;
    this.borderPaints = getBorderPaints(rows) ;
    this.borderStrokes = getBorderStrokes(rows) ;

    for (int i = 0 ; i < rows ; i++) {
      int currRowCols = dataset.getColumnCount(i) ;
      for (int j = 0 ; j < currRowCols ; j++) {
        IChartData data = dataset.getData(i, j) ;
        if (data != null && data instanceof XYZNumberChartData) {
          XYZNumberChartData xyz = (XYZNumberChartData) data ;
          double x = xyz.getX() ;
          double y = xyz.getY() ;
          double z = xyz.getZ() ;
          if (Double.isNaN(x) || Double.isNaN(y) || Double.isNaN(z)) {
            renderShapes.addDataItemShape(null) ;
          } else {
            double transX = xAxis.valueToJava2D(dataArea, xAxisEdge, x, false) ;
            double transY = yAxis.valueToJava2D(dataArea, yAxisEdge, y, false) ;

            double transDomain = 0.0 ;
            double transRange = 0.0 ;
            double zero = 0 ;

            boolean isInvalid = false ;
            switch (this.scaleType) {
            case SCALE_ON_DOMAIN_AXIS:
              zero = xAxis.valueToJava2D(dataArea, xAxisEdge, 0, false) ;
              transDomain = xAxis.valueToJava2D(dataArea, xAxisEdge, z, false) ;

              isInvalid = zero == AbstractValueAxis.INVALID_COORDINATE_VALUE
                          || transDomain == AbstractValueAxis.INVALID_COORDINATE_VALUE ;
              transDomain -= zero ;
              transRange = transDomain ;
              break ;
            case SCALE_ON_RANGE_AXIS:
              zero = yAxis.valueToJava2D(dataArea, yAxisEdge, 0, false) ;
              double zPos = yAxis.valueToJava2D(dataArea, yAxisEdge, z, false) ;
              transRange = zero - zPos ;
              transDomain = transRange ;

              isInvalid = zero == AbstractValueAxis.INVALID_COORDINATE_VALUE
                          || zPos == AbstractValueAxis.INVALID_COORDINATE_VALUE ;
              break ;
            default:
              double zero1 = xAxis.valueToJava2D(dataArea, xAxisEdge, 0, false) ;
              double zero2 = yAxis.valueToJava2D(dataArea, yAxisEdge, 0, false) ;
              transDomain = xAxis.valueToJava2D(dataArea, xAxisEdge, z, false) ;
              double tmpZPos = yAxis.valueToJava2D(dataArea, yAxisEdge, z, false) ;
              transRange = zero2 - tmpZPos ;

              isInvalid = zero1 == AbstractValueAxis.INVALID_COORDINATE_VALUE
                          || zero2 == AbstractValueAxis.INVALID_COORDINATE_VALUE
                          || tmpZPos == AbstractValueAxis.INVALID_COORDINATE_VALUE ;
              transDomain -= zero1 ;
            }

            if (transX == AbstractValueAxis.INVALID_COORDINATE_VALUE
                || transY == AbstractValueAxis.INVALID_COORDINATE_VALUE
                || zero == AbstractValueAxis.INVALID_COORDINATE_VALUE || isInvalid) {
              renderShapes.addDataItemShape(null) ;
              continue ;
            }

            transDomain = Math.abs(transDomain) ;
            transRange = Math.abs(transRange) ;
            Rectangle2D rect = null ;
            if (orientation.equals(ChartOrientation.VERTICAL)) {
              rect = new Rectangle2D.Double(transX - transDomain / 2.0,
                                            transY - transRange / 2.0, transDomain, transRange) ;
            } else if (orientation.equals(ChartOrientation.HORIZONTAL)) {
              rect = new Rectangle2D.Double(transY - transRange / 2.0,
                                            transX - transDomain / 2.0, transRange, transDomain) ;
            }

            EllipseTextChartShape shape = new EllipseTextChartShape(rect, null) ;
            IChartShape chartShape = createItemShape(orientation, yAxis, i, j, rect, fillPaints[i],
                                                     borderPaints[i], borderStrokes[i], alpha,
                                                     shape) ;
            renderShapes.addDataItemShape(chartShape) ;
          }
        } else {
          renderShapes.addDataItemShape(null) ;
        }
      }
    }
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Return the legend type.</en>
   *
   * @return LegendType <zh>ͼ.</zh>
   * <en>The legend type.</en>
   */
  public LegendType getLegendType() {
    return LegendType.ROW ;
  }

  /**
   * <zh>ÿͼͼ.</zh>
   * <en>Returns the shape array of legend items.</en>
   *
   * @param legendCount int <zh>ͼĸ.</zh>
   * <en>The legend items count.</en>
   * @return IChartShape[] <zh>ͼ.</zh>
   * <en>The shapes array.</en>
   */
  protected IChartShape[] getLegendItemShapes(int legendCount) {
    if (legendCount > 0) {
      final int legendWidth = 8, legendHeight = 8 ; //ͼ,߶
      IChartShape[] shapes = new IChartShape[legendCount] ;

      for (int i = 0 ; i < legendCount ; i++) {
        Rectangle2D rect = new Rectangle2D.Double(0, 0, legendWidth, legendHeight) ;
        shapes[i] = new EllipseTextChartShape(rect, null) ;
      }

      return shapes ;
    }

    return null ;
  }

  /**
   * <zh>ֻͼ,ͼε.</zh>
   * <en>Create a shape with no attribute specified.</en>
   *
   * @param rect Rectangle2D <zh>ͼľ (Ϊnull).</zh>
   * <en>The bar rectangle (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   * @param appendParam Object <zh>ӵĲ,ڴݶĶ (Ϊnull).</zh>
   * <en>The append parameter,used for special propose (null permitted).</en>
   *
   * @return IChartShape <zh>ͼ.</zh>
   * <en>The shape.</en>
   */
  protected IChartShape createItemShapeOnly(Rectangle2D rect, String text,
                                            Object appendParam) {
    if (appendParam == null) {
      throw new IllegalArgumentException("Null 'appendParam' argument.") ;
    }

    return (IChartShape) appendParam ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof XYBubbleRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    XYBubbleRender that = (XYBubbleRender) obj ;
    if (this.scaleType != that.scaleType) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int result = super.hashCode() ;
    result = result * 37 + this.scaleType ;

    return result ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "XYBubbleRender" ;
  }

}
