package com.ijchart.xychart.render.xy ;

import java.awt.Shape ;
import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.LegendType ;
import com.ijchart.xychart.data.ErrorChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.shape.ErrorCrossChartShape ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.shape.LineChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.ui.supplier.IDrawingSupplier ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>Errorͼ.</zh>
 * <en>Draw error chart shapes.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class XYErrorRender extends AbstractXYRender implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -3136019251937260767L ;

  /**
   * <zh>ȱʡǷxError.</zh>
   * <en>A default flag that controls whether or not the x-error bars are drawn.</en>
   */
  public static final boolean DEFAULT_DRAW_X_ERROR = true ;

  /**
   * <zh>ȱʡǷyError.</zh>
   * <en>A default flag that controls whether or not the y-error bars are drawn.</en>
   */
  public static final boolean DEFAULT_DRAW_Y_ERROR = true ;

  /**
   * <zh>ȱʡxyߵյĽߵĳ.</zh>
   * <en>The default length of the cap at the end of the error bars.</en>
   */
  public static final double DEFAULT_CAP_LENGTH = 4 ;

  /**
   * <zh>ȱʡǷ.</zh>
   * <en>A default flag that controls whether or not draw join lines.</en>
   */
  public static final boolean DEFAULT_DRAW_JOIN_LINE = false ;

  /**
   * <zh>ǷxError.</zh>
   * <en>A flag that controls whether or not the x-error bars are drawn.</en>
   */
  private boolean drawXError ;

  /**
   * <zh>ǷyError.</zh>
   * <en>A flag that controls whether or not the y-error bars are drawn.</en>
   */
  private boolean drawYError ;

  /**
   * <zh>xyߵյĽߵĳ.</zh>
   * <en>The length of the cap at the end of the error bars.</en>
   */
  private double capLength ;

  /**
   * <zh>Ƿ.</zh>
   * <en>A flag that controls whether or not draw join lines.</en>
   */
  private boolean drawJoinLine ;

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public XYErrorRender(IMatrixDataset dataset) {
    super(dataset) ;

    this.drawXError = DEFAULT_DRAW_X_ERROR ;
    this.drawYError = DEFAULT_DRAW_Y_ERROR ;
    this.capLength = DEFAULT_CAP_LENGTH ;
    this.drawJoinLine = DEFAULT_DRAW_JOIN_LINE ;

    setBaseItemFill(true) ;
    setBaseItemDrawBorder(true) ;
    setAutoSelectBorderPaint(true) ;
  }

  /**
   * <zh>ǷxError.</zh>
   * <en>Returns whether or not the x-error bars are drawn.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDrawXError() {
    return this.drawXError ;
  }

  /**
   * <zh>ǷxError,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets whether or not the x-error bars are drawn,and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to
   *  all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setDrawXError(boolean flag) {
    if (this.drawXError != flag) {
      this.drawXError = flag ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ǷyError,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Returns whether or not the y-error bars are drawn,and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent}
   *  to all registered listeners.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDrawYError() {
    return this.drawYError ;
  }

  /**
   * <zh>ǷyError.</zh>
   * <en>Sets whether or not the y-error bars are drawn.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setDrawYError(boolean flag) {
    if (this.drawYError != flag) {
      this.drawYError = flag ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>xyߵյĽߵĳ.</zh>
   * <en>Returns the length of the cap at the end of the error bars.</en>
   *
   * @return double <zh>.</zh>
   * <en>The length.</en>
   */
  public double getCapLength() {
    return this.capLength ;
  }

  /**
   * <zh>xyߵյĽߵĳ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the length of the cap at the end of the error bars,and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param length double <zh>,>=0.</zh>
   * <en>The length,must be >=0.</en>
   */
  public void setCapLength(double length) {
    if (length < 0) {
      throw new IllegalArgumentException("Argument 'length' must be >=0.") ;
    }

    if (this.capLength != length) {
      this.capLength = length ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Returns a flag that controls whether or not draw join lines.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDrawJoinLine() {
    return this.drawJoinLine ;
  }

  /**
   * <zh>Ƿ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets a flag that controls whether or not draw join lines and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>Ƿ.</zh>
   * <en>A flag that controls whether or not draw join lines.</en>
   */
  public void setDrawJoinLine(boolean flag) {
    if (this.drawJoinLine != flag) {
      this.drawJoinLine = flag ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param xAxis AbstractValueAxis <zh>xֵ (Ϊnull).</zh>
   * <en>The x value axis (null not permitted).</en>
   * @param xAxisEdge RectangleEdge <zh>xλ(Ϊnull).</zh>
   * <en>The x axis edge.</en>
   * @param yAxis AbstractValueAxis <zh>yֵ (Ϊnull).</zh>
   * <en>The y value axis (null not permitted).</en>
   * @param yAxisEdge RectangleEdge <zh>yλ (Ϊnull).</zh>
   * <en>The y axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(AbstractValueAxis xAxis, RectangleEdge xAxisEdge,
                           AbstractValueAxis yAxis, RectangleEdge yAxisEdge,
                           Rectangle2D dataArea, ChartOrientation orientation,
                           RenderShapeCollection renderShapes, float alpha) {
    if (xAxis == null) {
      throw new IllegalArgumentException("Null 'xAxis' argument.") ;
    }
    if (xAxisEdge == null) {
      throw new IllegalArgumentException("Null 'xAxisEdge' argument.") ;
    }
    if (yAxis == null) {
      throw new IllegalArgumentException("Null 'yAxis' argument.") ;
    }
    if (yAxisEdge == null) {
      throw new IllegalArgumentException("Null 'yAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }

    int rows = dataset.getRowCount() ;
    this.fillPaints = getFillPaints(rows) ;
    getBorderPaints(rows) ;
    this.borderPaints = fillPaints ;
    this.borderStrokes = getBorderStrokes(rows) ;

    boolean isVertical = orientation.equals(ChartOrientation.VERTICAL) ; //ֱ
    IDrawingSupplier supplier = getDrawingSupplier() ;
    double lastXPos = Double.NaN, lastYPos = Double.NaN ;

    for (int i = 0 ; i < rows ; i++) {
      Shape shape = supplier.getNextShape() ;
      int currRowCols = dataset.getColumnCount(i) ;
      lastXPos = Double.NaN ;
      lastYPos = Double.NaN ;
      for (int j = 0 ; j < currRowCols ; j++) {
        ErrorChartData data = (ErrorChartData) dataset.getData(i, j) ;
        if (data != null) {
          double x = data.getXValue() ;
          double xLow = data.getXLowValue() ;
          double xHigh = data.getXHighValue() ;

          double y = data.getYValue() ;
          double yLow = data.getYLowValue() ;
          double yHigh = data.getYHighValue() ;

          if (Double.isNaN(x) || Double.isNaN(y) || Double.isNaN(xLow) || Double.isNaN(yLow)
              || Double.isNaN(xHigh) || Double.isNaN(yHigh)) {
            lastXPos = Double.NaN ;
            lastYPos = Double.NaN ;
            renderShapes.addDataItemShape(null) ;
            continue ;
          }

          double xPos = xAxis.valueToJava2D(dataArea, xAxisEdge, x, false) ;
          double xLowPos = xAxis.valueToJava2D(dataArea, xAxisEdge, xLow, false) ;
          double xHighPos = xAxis.valueToJava2D(dataArea, xAxisEdge, xHigh, false) ;
          double yPos = yAxis.valueToJava2D(dataArea, yAxisEdge, y, false) ;
          double yLowPos = yAxis.valueToJava2D(dataArea, yAxisEdge, yLow, false) ;
          double yHighPos = yAxis.valueToJava2D(dataArea, yAxisEdge, yHigh, false) ;
          //ֵ᷶Χֵ
          if (xPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
              || xLowPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
              || xHighPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
              || yPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
              || yLowPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
              || yHighPos == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
            lastXPos = Double.NaN ;
            lastYPos = Double.NaN ;
            renderShapes.addDataItemShape(null) ;
            continue ;
          }

          ErrorCrossChartShape errorShape = null ;
          if (isVertical) {
            errorShape = new ErrorCrossChartShape(Math.min(xLowPos, xHighPos),
                                                  Math.max(xLowPos, xHighPos), xPos, yPos,
                                                  Math.min(yLowPos, yHighPos),
                                                  Math.max(yLowPos, yHighPos), shape) ;
            errorShape.setDrawXError(this.drawXError) ;
            errorShape.setDrawYError(this.drawYError) ;
          } else {
            errorShape = new ErrorCrossChartShape(Math.min(yLowPos, yHighPos),
                                                  Math.max(yLowPos, yHighPos), yPos, xPos,
                                                  Math.min(xLowPos, xHighPos),
                                                  Math.max(xLowPos, xHighPos), shape) ;
            errorShape.setDrawXError(this.drawYError) ;//ȡ
            errorShape.setDrawYError(this.drawXError) ;//ȡ
          }
          errorShape.setCapLength(this.capLength) ;

          IChartShape chartShape = createItemShape(orientation, yAxis, i, j, null,
                                                   fillPaints[i], borderPaints[i],
                                                   borderStrokes[i], alpha, errorShape) ;
          renderShapes.addDataItemShape(chartShape) ;
          if (j > 0 && this.drawJoinLine) {
            LineChartShape lineShape = null ;
            if (isVertical) {
              lineShape = new LineChartShape(lastXPos, lastYPos, xPos, yPos) ;
            } else {
              lineShape = new LineChartShape(lastYPos, lastXPos, yPos, xPos) ;
            }
            IChartShape lineChartShape = createItemShape(orientation, yAxis, i, j, null,
                                                         fillPaints[i], borderPaints[i],
                                                         borderStrokes[i], alpha, lineShape) ;
            renderShapes.addJoinLineShape(lineChartShape) ;
          }

          lastXPos = xPos ;
          lastYPos = yPos ;
        } else {
          lastXPos = Double.NaN ;
          lastYPos = Double.NaN ;
          renderShapes.addDataItemShape(null) ;
        }
      }
    }

  }

  /**
   * <zh>ͼ.</zh>
   * <en>Return the legend type.</en>
   *
   * @return LegendType <zh>ͼ.</zh>
   * <en>The legend type.</en>
   */
  public LegendType getLegendType() {
    return LegendType.ROW ;
  }

  /**
   * <zh>ֻͼ,ͼε.</zh>
   * <en>Create a shape with no attribute specified.</en>
   *
   * @param rect Rectangle2D <zh>ͼľ (Ϊnull).</zh>
   * <en>The bar rectangle (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   * @param appendParam Object <zh>ӵĲ,ڴݶĶ (Ϊnull).</zh>
   * <en>The append parameter,used for special propose (null permitted).</en>
   *
   * @return IChartShape <zh>ͼ.</zh>
   * <en>The shape.</en>
   */
  protected IChartShape createItemShapeOnly(Rectangle2D rect, String text,
                                            Object appendParam) {
    if (appendParam == null) {
      throw new IllegalArgumentException("Null 'appendParam' argument.") ;
    }

    return (IChartShape) appendParam ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates if some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof XYErrorRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    XYErrorRender that = (XYErrorRender) obj ;
    if (this.drawXError != that.drawXError) {
      return false ;
    }
    if (this.drawYError != that.drawYError) {
      return false ;
    }
    if (this.capLength != that.capLength) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double hashcode = super.hashCode() ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.drawXError) ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.drawYError) ;
    hashcode = 37 * hashcode + this.capLength ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.drawJoinLine) ;

    return (int) hashcode ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   *   java.lang.Cloneable
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "XYErrorRender" ;
  }

}
