package com.ijchart.xychart.render.xy ;

import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.HorizontalAlignment ;
import com.ijchart.xychart.constant.LegendType ;
import com.ijchart.xychart.constant.TextPosition ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.data.IXYChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.shape.DualBreakLineChartShape ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>XYͼͼ.</zh>
 * <en>A render that draws shapes for each data item, and lines between data
 * items (for use with the {@link com.ijchart.xychart.plot.XYPlot} class).</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class XYStepLineRender extends AbstractXYRender implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -8591853263454319847L ;

  /**
   * <zh>.</zh>
   * <en>constructor.</en>
   *
   * @param dataset IMatrixDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public XYStepLineRender(IMatrixDataset dataset) {
    super(dataset) ;

    setBaseItemDrawBorder(true) ;
    setAutoSelectBorderPaint(true) ;
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param xAxis AbstractValueAxis <zh>xֵ (Ϊnull).</zh>
   * <en>The x value axis (null not permitted).</en>
   * @param xAxisEdge RectangleEdge <zh>xλ(Ϊnull).</zh>
   * <en>The x axis edge.</en>
   * @param yAxis AbstractValueAxis <zh>yֵ (Ϊnull).</zh>
   * <en>The y value axis (null not permitted).</en>
   * @param yAxisEdge RectangleEdge <zh>yλ (Ϊnull).</zh>
   * <en>The y axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(AbstractValueAxis xAxis, RectangleEdge xAxisEdge,
                           AbstractValueAxis yAxis, RectangleEdge yAxisEdge,
                           Rectangle2D dataArea, ChartOrientation orientation,
                           RenderShapeCollection renderShapes, float alpha) {
    if (xAxis == null) {
      throw new IllegalArgumentException("Null 'xAxis' argument.") ;
    }
    if (xAxisEdge == null) {
      throw new IllegalArgumentException("Null 'xAxisEdge' argument.") ;
    }
    if (yAxis == null) {
      throw new IllegalArgumentException("Null 'yAxis' argument.") ;
    }
    if (yAxisEdge == null) {
      throw new IllegalArgumentException("Null 'yAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }

    int rows = dataset.getRowCount() ;
    this.fillPaints = getFillPaints(rows) ;
    getBorderPaints(rows) ;
    this.borderPaints = fillPaints ;
    this.borderStrokes = getBorderStrokes(rows) ;

    //ߺͼ
    if (orientation.equals(ChartOrientation.VERTICAL)) { //ֱ
      double lastX = Double.NaN ; //һXֵ
      double lastY = Double.NaN ; //һYֵ
      for (int i = 0 ; i < rows ; i++) {
        int lastNotNullValueIndex = getRowDataLastNotNullValue(dataset, i) ;
        int currRowCols = dataset.getColumnCount(i) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof IXYChartData) {
            double xValue = ((IXYChartData) data).getX() ;
            double x = xAxis.valueToJava2D(dataArea, xAxisEdge, xValue, false) ;
            double yValue = ((IXYChartData) data).getY() ;
            double y = yAxis.valueToJava2D(dataArea, yAxisEdge, yValue, false) ;
            //ֵ᷶Χֵ
            if (x == AbstractValueAxis.INVALID_COORDINATE_VALUE
                || y == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
              lastX = Double.NaN ;
              lastY = Double.NaN ;
              renderShapes.addDataItemShape(null) ;
              continue ;
            }

            DualBreakLineChartShape breakLineShape = null ;
            Rectangle2D rect = new Rectangle2D.Double(x, y, 10, 10) ;
            if (j == 0 || (Double.isNaN(lastX) && Double.isNaN(lastY))) {
              setItemLabelHorizontalAlignment(i, j, HorizontalAlignment.LEFT) ;
              breakLineShape = new DualBreakLineChartShape(rect, null, null, x, y,
                                                           x + 1, y, x + 1, y) ;
            } else {
              if (j == lastNotNullValueIndex) {
                setItemLabelHorizontalAlignment(i, j, HorizontalAlignment.RIGHT) ;
              }
              breakLineShape = new DualBreakLineChartShape(rect, null, null, lastX,
                                                           lastY, x, lastY, x, y) ;
            }

            IChartShape chartShape = createItemShape(orientation, yAxis, i, j, rect,
                                                     fillPaints[i], borderPaints[i],
                                                     borderStrokes[i], alpha, breakLineShape) ;
            renderShapes.addDataItemShape(chartShape) ;

            lastX = x ;
            lastY = y ;
          } else {
            lastX = Double.NaN ;
            lastY = Double.NaN ;
            renderShapes.addDataItemShape(null) ;
          }
        }
      }
    } else if (orientation.equals(ChartOrientation.HORIZONTAL)) { //ˮƽ
      double lastX = Double.NaN ; //һXֵ
      double lastY = Double.NaN ; //һYֵ
      for (int i = 0 ; i < rows ; i++) {
        int lastNotNullValueIndex = getRowDataLastNotNullValue(dataset, i) ;
        int currRowCols = dataset.getColumnCount(i) ;

        for (int j = 0 ; j < currRowCols ; j++) {
          IChartData data = dataset.getData(i, j) ;
          if (data != null && data instanceof IXYChartData) {
            double xValue = ((IXYChartData) data).getX() ;
            double y = xAxis.valueToJava2D(dataArea, xAxisEdge, xValue, false) ;
            double yValue = ((IXYChartData) data).getY() ;
            double x = yAxis.valueToJava2D(dataArea, yAxisEdge, yValue, false) ;

            //ֵ᷶Χֵ
            if (x == AbstractValueAxis.INVALID_COORDINATE_VALUE
                || y == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
              lastX = Double.NaN ;
              lastY = Double.NaN ;
              renderShapes.addDataItemShape(null) ;
              continue ;
            }

            DualBreakLineChartShape breakLineShape = null ;
            Rectangle2D rect = new Rectangle2D.Double(x, y, 1, 1) ;
            if (j == 0 || (Double.isNaN(lastX) && Double.isNaN(lastY))) {
              setItemLabelPosition(i, j, TextPosition.OUTSIDE_TOP) ;
              breakLineShape = new DualBreakLineChartShape(rect, null, null, x, y,
                                                           x + 1, y, x + 1, y) ;
            } else {
              if (j == lastNotNullValueIndex) {
                setItemLabelPosition(i, j, TextPosition.OUTSIDE_BOTTOM) ;
              }
              breakLineShape = new DualBreakLineChartShape(rect, null, null, lastX,
                                                           lastY, x, lastY, x, y) ;
            }
            IChartShape chartShape = createItemShape(orientation, yAxis, i, j, rect,
                                                     fillPaints[i], borderPaints[i],
                                                     borderStrokes[i], alpha, breakLineShape) ;
            renderShapes.addDataItemShape(chartShape) ;

            lastX = x ;
            lastY = y ;
          } else {
            lastX = Double.NaN ;
            lastY = Double.NaN ;
            renderShapes.addDataItemShape(null) ;
          }
        }
      }
    }

  }

  /**
   * <zh>ݼһеһΪnullֵ.</zh>
   * <en>Calculate the index of the last not null value.</en>
   *
   * @param dataset IMatrixDataset
   * @param row int
   * @return int
   */
  private static int getRowDataLastNotNullValue(IMatrixDataset dataset, int row) {
    int cols = dataset.getColumnCount(row) ;
    for (int i = cols - 1 ; i >= 0 ; i--) {
      if (dataset.getData(row, i) != null) {
        return i ;
      }
    }

    return 0 ;
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Return the legend type.</en>
   *
   * @return LegendType <zh>ͼ.</zh>
   * <en>The legend type.</en>
   */
  public LegendType getLegendType() {
    return LegendType.ROW ;
  }

  /**
   * <zh>ֻͼ,ͼε.</zh>
   * <en>Create a shape with no attribute specified.</en>
   *
   * @param rect Rectangle2D <zh>ͼľ (Ϊnull).</zh>
   * <en>The bar rectangle (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   * @param appendParam Object <zh>ӵĲ,ڴݶĶ (Ϊnull).</zh>
   * <en>The append parameter,used for special propose (null permitted).</en>
   *
   * @return IChartShape <zh>ͼ.</zh>
   * <en>The shape.</en>
   */
  protected IChartShape createItemShapeOnly(Rectangle2D rect, String text,
                                            Object appendParam) {
    if (appendParam == null) {
      throw new IllegalArgumentException("Null 'appendParam' argument.") ;
    }

    return (IChartShape) appendParam ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this instance for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return A boolean.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof XYStepLineRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    return super.hashCode() ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "XYStepRender" ;
  }

}
