package com.ijchart.xychart.tickunit ;

import java.io.Serializable ;
import java.text.DecimalFormat ;
import java.text.NumberFormat ;

import com.ijchart.xychart.util.MathUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ڼ͵ݵ̶ֵ.</zh>
 * <en>Calculate the tick unit for number data and all values is logarithmic.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class LogTickUnit extends DoubleTickUnit implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 592288758202718238L ;

  /**
   * <zh>ȱʡĶ.</zh>
   * <en>The default logarithm base.</en>
   */
  public static final double DEFAULT_LOGARITHM_BASE = 10.0 ;

  /**
   * <zh>.</zh>
   * <en>The logarithm base.</en>
   */
  private double logarithmBase ;

  /**
   * <zh>logֵ,Math.log(logarithmBase).</zh>
   * <en>The log value for base logarithm,equals Math.log(logarithmBase).</en>
   */
  private double baseLogValue ;

  /**
   * <zh>ָǩ.</zh>
   * <en>The exponent labels.</en>
   */
  private String[] exponentLabels ;

  /**
   * <zh>ȱʡ.</zh>
   * <en>A default constructor.</en>
   */
  public LogTickUnit() {
    this(DEFAULT_LOGARITHM_BASE, Double.NaN) ;
  }

  /**
   * <zh>.</zh>
   * <en>A constructor.</en>
   *
   * @param logarithmBase double <zh>,1.</zh>
   * <en>The logarithm base,must be greater than 1.</en>
   * @param notUseFlag <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public LogTickUnit(double logarithmBase, boolean notUseFlag) {
    this(logarithmBase, Double.NaN) ;
  }

  /**
   * <zh>.</zh>
   * <en>A constructor.</en>
   *
   * @param notUseFlag <zh>ֵ.</zh>
   * <en>A boolean.</en>
   * @param step double <zh>.</zh>
   * <en>The step.</en>
   */
  public LogTickUnit(boolean notUseFlag, double step) {
    this(DEFAULT_LOGARITHM_BASE, step) ;
  }

  /**
   * <zh>.</zh>
   * <en>A constructor.</en>
   *
   * @param logarithmBase double <zh>,1.</zh>
   * <en>The logarithm base,must be greater than 1.</en>
   * @param step double <zh>.</zh>
   * <en>The step.</en>
   */
  public LogTickUnit(double logarithmBase, double step) {
    super(step) ;
    if (logarithmBase <= 1.0) {
      throw new IllegalArgumentException("Requires 'logarithmBase' > 1.0.") ;
    }

    this.logarithmBase = logarithmBase ;
    this.baseLogValue = Math.log(this.logarithmBase) ;
    this.exponentLabels = null ;
  }

  /**
   * <zh>ض.</zh>
   * <en>Returns the logarithm base.</en>
   *
   * @return double <zh>.</zh>
   * <en>The logarithm base.</en>
   */
  public double getLogarithmBase() {
    return this.logarithmBase ;
  }

  /**
   * <zh>ػlogֵ.</zh>
   * <en>Returns the log value for base logarithm.</en>
   *
   * @return double <zh>logֵ.</zh>
   * <en>The log value for base logarithm.</en>
   */
  public double getBaseLogValue() {
    return this.baseLogValue ;
  }

  /**
   * <zh>ָǩ.</zh>
   * <en>Returns the exponent labels.</en>
   *
   * @return String[] <zh>ָǩ (Ϊnull).</zh>
   * <en>The exponent labels (maybe null).</en>
   */
  public String[] getExponentLabels() {
    return this.exponentLabels ;
  }

  /**
   * <zh>СֵֵݵȷϵĿ̶ֵ,̶ֵdouble͵ĸ.</zh>
   * <en>According to the minimum and maximum data and the largest number of coordinates
   * to determine the scale of axis,all values are double.</en>
   *
   * @param minValue double <zh>Сֵ.</zh>
   * <en>The minimum value.</en>
   * @param maxValue double <zh>ֵ.</zh>
   * <en>The max value.</en>
   * @param maxCount int <zh>̶ֵ,0.</zh>
   * <en>The maximum count of tick units (must great than 0).</en>
   * @param includeZero boolean <zh>ֵ,Ƿ0ֵ.</zh>
   * <en>A boolean flag indicates whether or not include zero value.</en>
   *
   * @return String[] <zh>̶ֵǿ,ֵС˳,ÿԪض,(Ϊnull).</zh>
   * <en>An array of tick units,the values are sorted ascending.
   * All elements in array is number (never null).</en>
   */
  public String[] getTickUnits(double minValue, double maxValue, int maxCount, boolean includeZero) {
    if (minValue > maxValue) {
      throw new IllegalArgumentException("Must be 'minValue'<='maxValue'.") ;
    }
    if (maxCount < 1) {
      throw new IllegalArgumentException("Must be 'maxCount'>0.") ;
    }

    double start = MathUtil.sign(minValue) * Math.log(Math.abs(minValue)) / this.baseLogValue ;
    double end = MathUtil.sign(maxValue) * Math.log(Math.abs(maxValue)) / this.baseLogValue ;
    if (Double.isNaN(start) || Double.isNaN(end)) {
      return new String[] {String.valueOf(minValue)} ;
    }

    String[] ticks = null ;
    if (minValue == maxValue) {
      ticks = new String[] {String.valueOf(start)} ;
    } else {
      ticks = super.getTickUnits(Math.min(start, end), Math.max(start, end), maxCount, includeZero) ;
      this.exponentLabels = new String[ticks.length] ;
      System.arraycopy(ticks, 0, this.exponentLabels, 0, ticks.length) ;
    }

    NumberFormat baseFormat = new DecimalFormat("#.#") ;
    String formatBase = baseFormat.format(this.logarithmBase) ;

    for (int i = 0 ; i < ticks.length ; i++) {
      double value = Double.parseDouble(ticks[i]) ;
      if (value == 0.0) {
        ticks[i] = "0" ;
      } else if (value < 0.0) {
        ticks[i] = "-" + formatBase + "^" + ticks[i].substring(1) ;
      } else if (value > 0.0) {
        ticks[i] = formatBase + "^" + ticks[i] ;
      }
    }

    return ticks ;
  }

  /**
   * <zh>ԱǷһ.</zh>
   * <en>Tests the axis for equality with an arbitrary object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return true or false.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }

    if (!(obj instanceof LogTickUnit)) {
      return false ;
    }

    LogTickUnit that = (LogTickUnit) obj ;
    if (this.logarithmBase != that.logarithmBase) {
      return false ;
    }

    return true ;
  }

  /**
   * Returns a clone of the object.
   *
   * @return A clone.
   *
   * @throws CloneNotSupportedException if some component of the axis does
   *         not support cloning.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>hashCode.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double result = super.hashCode() ;
    result = result * 37 + this.logarithmBase ;

    return (int) result ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return A string representation of the object.
   */
  public String toString() {
    return "LogTickUnit" ;
  }

}
